# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

"""
This module provides scapy based classes that represent the match parameters to use when building
the match, mask and last buffers in HwsData. this should be always in sync with the item_data_box
struct to insure size and offset compatability.
"""

import unittest

try:
    import logging
    logging.getLogger("scapy.runtime").setLevel(logging.ERROR)
    from scapy.packet import Packet
    from scapy.fields import BitField, ByteField, IPField, ShortField, PacketField, MACField, \
        IntField
    from scapy.layers.inet6 import IP6Field
except ImportError:
    raise unittest.SkipTest('scapy package is needed in order to run tests')
import pyverbs.providers.mlx5.mlx5hws_enums as e

class Layer24(Packet):
    fields_desc = [
        BitField('l3_type', 0, 2),
        BitField('l4_type', 0, 4),
        BitField('reserved_0', 0, 2),
        ShortField('eth_type', 0),
        BitField('reserved_1', 0, 8), # Align SMAC
        MACField('smac', '00:00:00:00:00:00'),
        BitField('reserved_2', 0, 16), # Align DMAC
        MACField('dmac', '00:00:00:00:00:00'),
        BitField('reserved_3', 0, 16), # Align VLAN
        BitField('first_prio', 0, size=3),
        BitField('first_cfi', 0, size=1),
        BitField('first_vid', 0, size=12),
        BitField('first_cvlan', 0, size=1),
        BitField('first_svlan', 0, size=1),
        BitField('reserved_4', 0, 14), # Align VLAN
        BitField('second_prio', 0, size=3),
        BitField('second_cfi', 0, size=1),
        BitField('second_vid', 0, size=12),
        BitField('second_cvlan', 0, size=1),
        BitField('second_svlan', 0, size=1),
        BitField('ip_version', 0, 4),
        BitField('ip_frag', 0, 1),
        BitField('reserved_5', 0, 9), # class Packet has to be aligned to 8
        ShortField('ip_sport', 0),
        ShortField('ip_dport', 0),
        ShortField('ip_tcp_sport', 0),
        ShortField('ip_tcp_dport', 0),
        ByteField('ip_tcp_flags', 0),
        ShortField('ip_udp_sport', 0),
        ShortField('ip_udp_dport', 0),
        ShortField('ipv4_len', 0),
        IPField('ipv4_src', '0.0.0.0'),
        IPField('ipv4_dst', '0.0.0.0'),
        IP6Field("ipv6_src", "::0"),
        IP6Field("ipv6_dst", "::0"),
    ]


class Misc(Packet):
    fields_desc = [
        IntField('metadata_reg_c_0', 0),
        IntField('metadata_reg_c_1', 0),
        IntField('metadata_reg_c_2', 0),
        IntField('metadata_reg_c_3', 0),
        IntField('metadata_reg_c_4', 0),
        IntField('metadata_reg_c_5', 0),
        IntField('metadata_reg_c_6', 0),
        IntField('metadata_reg_c_7', 0),
        IntField('metadata_reg_c_8', 0),
        IntField('metadata_reg_c_9', 0),
        IntField('metadata_reg_c_10', 0),
        IntField('metadata_reg_c_11', 0),
        IntField('metadata_reg_a', 0),
        IntField('metadata_reg_b', 0),
        IntField('flex_parser_0', 0),
        IntField('flex_parser_1', 0),
        IntField('flex_parser_2', 0),
        IntField('flex_parser_3', 0),
        IntField('flex_parser_4', 0),
        IntField('flex_parser_5', 0),
        IntField('flex_parser_6', 0),
        IntField('flex_parser_7', 0),
        BitField('flex_parser_0_ok', 0, 1),
        BitField('flex_parser_1_ok', 0, 1),
        BitField('flex_parser_2_ok', 0, 1),
        BitField('flex_parser_3_ok', 0, 1),
        BitField('flex_parser_4_ok', 0, 1),
        BitField('flex_parser_5_ok', 0, 1),
        BitField('flex_parser_6_ok', 0, 1),
        BitField('flex_parser_7_ok', 0, 1),
        BitField('src_qp', 0, 24),
        BitField('reserved_0', 0, 8),
        BitField('random_num', 0, 16)
    ]


class MatchParams(Packet):
    fields_desc = [
        PacketField('outer_headers', Layer24(), Layer24),
        PacketField('inner_headers', Layer24(), Layer24),
        PacketField('misc_parameters', Misc(), Misc),
    ]

    def __init__(self):
        super().__init__()
        self.outer_headers_offset = {}
        self.inner_headers_offset = {}
        self.misc_parameters_offset = {}
        self.init_headers_offset()
        self.build_field_to_offset()

    @property
    def size(self):
        return (len(self.outer_headers) +
                len(self.inner_headers) +
                len(self.misc_parameters))

    def init_headers_offset(self):
        """
        Initialize header name to offset dictionary.
        """
        header_name_dict_tup = [(self.outer_headers, self.outer_headers_offset),
                                (self.inner_headers, self.inner_headers_offset),
                                (self.misc_parameters, self.misc_parameters_offset)]
        offset = 0
        for packet_field, headers_offset in header_name_dict_tup:
            for field_name in packet_field.default_fields:
                field = packet_field.get_field(field_name)
                headers_offset[field_name] = offset
                offset += self.field_to_size(field)

    def field_to_size(self, field):
        """
        Not all scapy.fields has a size attribute, this function returns
        a predefined size for such types.
        :param field: A field from scapy.fields.
        """
        SHORT_SIZE = 16
        IP6_SIZE = 128
        BYTE_SIZE = 8
        MAC_SIZE = 48
        INT_SIZE = 32
        IP_SIZE = 32
        if isinstance(field, ShortField):
            return SHORT_SIZE
        if isinstance(field, IntField):
            return INT_SIZE
        if isinstance(field, MACField):
            return MAC_SIZE
        if isinstance(field, ByteField):
            return BYTE_SIZE
        if isinstance(field, IPField):
            return IP_SIZE
        if isinstance(field, IP6Field):
            return IP6_SIZE

        return field.size

    def build_field_to_offset(self):
        """
        Builds the offset for a given field name within the item data object.
        This function builds the mapping between field names to their corresponding offsets in
        outer, inner or misc fields representation in the item data object.
        It handles cases where a single field may be split across multiple field name (FNAME)
        objects, particularly for larger fields like IPv6 addresses.
        """
        DW_SIZE = 32
        def generate_offsets(header_offset):
            return {
                e.MLX5DV_HWS_FNAME_L3_TYPE: header_offset["l3_type"],
                e.MLX5DV_HWS_FNAME_L4_TYPE: header_offset["l4_type"],
                e.MLX5DV_HWS_FNAME_ETH_TYPE: header_offset["eth_type"],
                e.MLX5DV_HWS_FNAME_SMAC_47_16: header_offset["smac"],
                e.MLX5DV_HWS_FNAME_SMAC_15_0: header_offset["smac"] + DW_SIZE,
                e.MLX5DV_HWS_FNAME_DMAC_47_16: header_offset["dmac"],
                e.MLX5DV_HWS_FNAME_DMAC_15_0: header_offset["dmac"] + DW_SIZE,
                e.MLX5DV_HWS_FNAME_FIRST_PRIO: header_offset["first_prio"],
                e.MLX5DV_HWS_FNAME_FIRST_CFI: header_offset["first_cfi"],
                e.MLX5DV_HWS_FNAME_FIRST_VID: header_offset["first_vid"],
                e.MLX5DV_HWS_FNAME_FIRST_CVLAN: header_offset["first_cvlan"],
                e.MLX5DV_HWS_FNAME_FIRST_SVLAN: header_offset["first_svlan"],
                e.MLX5DV_HWS_FNAME_SECOND_PRIO: header_offset["second_prio"],
                e.MLX5DV_HWS_FNAME_SECOND_CFI: header_offset["second_cfi"],
                e.MLX5DV_HWS_FNAME_SECOND_VID: header_offset["second_vid"],
                e.MLX5DV_HWS_FNAME_SECOND_CVLAN: header_offset["second_cvlan"],
                e.MLX5DV_HWS_FNAME_SECOND_SVLAN: header_offset["second_svlan"],
                e.MLX5DV_HWS_FNAME_IP_VERSION: header_offset["ip_version"],
                e.MLX5DV_HWS_FNAME_IP_FRAG: header_offset["ip_frag"],
                e.MLX5DV_HWS_FNAME_IP_SPORT: header_offset["ip_sport"],
                e.MLX5DV_HWS_FNAME_IP_DPORT: header_offset["ip_dport"],
                e.MLX5DV_HWS_FNAME_IP_TCP_SPORT: header_offset["ip_tcp_sport"],
                e.MLX5DV_HWS_FNAME_IP_TCP_DPORT: header_offset["ip_tcp_dport"],
                e.MLX5DV_HWS_FNAME_IP_TCP_FLAGS: header_offset["ip_tcp_flags"],
                e.MLX5DV_HWS_FNAME_IP_UDP_SPORT: header_offset["ip_udp_sport"],
                e.MLX5DV_HWS_FNAME_IP_UDP_DPORT: header_offset["ip_udp_dport"],
                e.MLX5DV_HWS_FNAME_IPV4_LEN: header_offset["ipv4_len"],
                e.MLX5DV_HWS_FNAME_IPV4_SRC: header_offset["ipv4_src"],
                e.MLX5DV_HWS_FNAME_IPV4_DST: header_offset["ipv4_dst"],
                e.MLX5DV_HWS_FNAME_IPV6_DST_31_0: header_offset["ipv6_dst"],
                e.MLX5DV_HWS_FNAME_IPV6_DST_63_32: header_offset["ipv6_dst"] + DW_SIZE,
                e.MLX5DV_HWS_FNAME_IPV6_DST_95_64: header_offset["ipv6_dst"] + 2 * DW_SIZE,
                e.MLX5DV_HWS_FNAME_IPV6_DST_127_96: header_offset["ipv6_dst"] + 3 * DW_SIZE,
                e.MLX5DV_HWS_FNAME_IPV6_SRC_31_0: header_offset["ipv6_src"],
                e.MLX5DV_HWS_FNAME_IPV6_SRC_63_32: header_offset["ipv6_src"] + DW_SIZE,
                e.MLX5DV_HWS_FNAME_IPV6_SRC_95_64: header_offset["ipv6_src"] + 2 * DW_SIZE,
                e.MLX5DV_HWS_FNAME_IPV6_SRC_127_96: header_offset["ipv6_src"] + 3 * DW_SIZE
            }

        def generate_misc_offsets(header_offset):
            return {
                e.MLX5DV_HWS_FNAME_REG_C0: header_offset["metadata_reg_c_0"],
                e.MLX5DV_HWS_FNAME_REG_C1: header_offset["metadata_reg_c_1"],
                e.MLX5DV_HWS_FNAME_REG_C2: header_offset["metadata_reg_c_2"],
                e.MLX5DV_HWS_FNAME_REG_C3: header_offset["metadata_reg_c_3"],
                e.MLX5DV_HWS_FNAME_REG_C4: header_offset["metadata_reg_c_4"],
                e.MLX5DV_HWS_FNAME_REG_C5: header_offset["metadata_reg_c_5"],
                e.MLX5DV_HWS_FNAME_REG_C6: header_offset["metadata_reg_c_6"],
                e.MLX5DV_HWS_FNAME_REG_C7: header_offset["metadata_reg_c_7"],
                e.MLX5DV_HWS_FNAME_REG_C8: header_offset["metadata_reg_c_8"],
                e.MLX5DV_HWS_FNAME_REG_C9: header_offset["metadata_reg_c_9"],
                e.MLX5DV_HWS_FNAME_REG_C10: header_offset["metadata_reg_c_10"],
                e.MLX5DV_HWS_FNAME_REG_C11: header_offset["metadata_reg_c_11"],
                e.MLX5DV_HWS_FNAME_REG_A: header_offset["metadata_reg_a"],
                e.MLX5DV_HWS_FNAME_REG_B: header_offset["metadata_reg_b"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_0: header_offset["flex_parser_0"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_1: header_offset["flex_parser_1"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_2: header_offset["flex_parser_2"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_3: header_offset["flex_parser_3"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_4: header_offset["flex_parser_4"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_5: header_offset["flex_parser_5"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_6: header_offset["flex_parser_6"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_7: header_offset["flex_parser_7"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_0_OK: header_offset["flex_parser_0_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_1_OK: header_offset["flex_parser_1_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_2_OK: header_offset["flex_parser_2_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_3_OK: header_offset["flex_parser_3_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_4_OK: header_offset["flex_parser_4_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_5_OK: header_offset["flex_parser_5_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_6_OK: header_offset["flex_parser_6_ok"],
                e.MLX5DV_HWS_FNAME_FLEX_PARSER_7_OK: header_offset["flex_parser_7_ok"],
                e.MLX5DV_HWS_FNAME_SOURCE_QP: header_offset["src_qp"],
                e.MLX5DV_HWS_FNAME_RANDOM_NUM: header_offset["random_num"]
            }

        self.field_to_offset = {
            e.MLX5DV_HWS_HEADER_OUTER: generate_offsets(self.outer_headers_offset),
            e.MLX5DV_HWS_HEADER_INNER: generate_offsets(self.inner_headers_offset),
            e.MLX5DV_HWS_HEADER_NONE: generate_misc_offsets(self.misc_parameters_offset)
        }

    def get_field_offset(self, fname, fheader):
        """
        Retrieves the offset for a given field name within the item data object.
        """
        return self.field_to_offset[fheader].get(fname, -1)
