# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

import subprocess
import ipaddress
import unittest
import socket
import struct
import errno
import time
import os

from pyverbs.providers.mlx5.hws_match_template import HwsField, HwsItem, HwsMatchTemplate, \
    HwsMatchTemplateAttr
from pyverbs.providers.mlx5.hws_resource import HwsResourceAttr, HwsResourceEnqueueAsoAttr, \
    HwsResource
from pyverbs.providers.mlx5.hws_matcher import HwsMatcherAttr, HwsMatcher, HwsMatcherMetric, \
    HwsMetricMatcherTemplateAttr
from tests.mlx5_prm_structs import AsoWqeDataSeg, AsoCtrl, AsoData, FlowMeterParams, ConnTrackAso
from pyverbs.providers.mlx5.hws_rule import HwsRuleAttr, HwsRule, get_rule_handle_size
from pyverbs.providers.mlx5.mlx5dv import Mlx5Context, Mlx5DVContextAttr, Wqe
from pyverbs.providers.mlx5.hws_context import HwsContextAttr, HwsContext
from pyverbs.providers.mlx5.hws_table import HwsTableAttr, HwsTable
from tests.test_mlx5_dr import Mlx5DrTirResources, Mlx5DrResources
from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
from pyverbs.providers.mlx5.hws_item_data import HwsData
import pyverbs.providers.mlx5.mlx5hws_enums as e
from tests.hws_match_params import MatchParams
import pyverbs.providers.mlx5.hws_actions as a
import pyverbs.providers.mlx5.hws_parser as p
from tests.mlx5_base import Mlx5RDMATestCase
import pyverbs.enums as pe
from pyverbs.mr import MR
import tests.utils as u

NUM_OF_QUEUES = 5
DEFAULT_QUEUE_SIZE = 1024
MIN_COUNTER_BULK_LOG_SIZE = 2
MAX_COUNTER_BULK_LOG_SIZE = 23
MELLANOX_VENDOR_ID = 0x02c9
FLOW_METER_BUCKET_OVERFLOW = 0x1
CT_ASO = 0x31
CT_VALID_SYND = 0x1
CT_INVALID_SYND = 0x80

# MATCHER STE formats
MATCH_8DW = 4
MATCH_11DW = 5
MATCH_RANGE = 7
MATCH_4DW_RANGE = 8

SET_ACTION = 0x1
ADD_ACTION = 0x2
COPY_ACTION = 0x3


class ModifyFields:
    """
    Supported SW steering modify fields.
    """
    OUT_SMAC_47_16 = 0x1
    OUT_SMAC_15_0 = 0x2
    OUT_DMAC_47_16 = 0x4
    OUT_DMAC_15_0 = 0x5
    OUT_TCP_SPORT = 0x8
    OUT_TCP_DPORT = 0x9
    OUT_UDP_SPORT = 0xb
    OUT_UDP_DPORT = 0xc
    OUT_SIPV4 = 0x15
    OUT_DIPV4 = 0x16
    META_DATA_REG_C_0 = 0x51
    META_DATA_REG_C_1 = 0x52
    META_DATA_REG_C_2 = 0x53
    META_DATA_REG_C_3 = 0x54
    META_DATA_REG_C_4 = 0x55
    META_DATA_REG_C_5 = 0x56
    META_DATA_REG_C_6 = 0x57
    META_DATA_REG_C_7 = 0x58
    META_DATA_REG_C_8 = 0x8F
    META_DATA_REG_C_9 = 0x90
    META_DATA_REG_C_10 = 0x91
    META_DATA_REG_C_11 = 0x92


class ModifyFieldsLen:
    """
    Supported SW steering modify fields length.
    """
    MAC_47_16 = 32
    MAC_15_0 = 16
    PORT = 16
    IPV4 = 32
    META_DATA_REG_C = 32

def mac_to_int(mac):
    """
    Convert MAC address string to an integer.
    :param mac: MAC address in a string format.
    """
    return int(mac.replace(":", ""), 16)

def int_to_mac(value):
    """
    Convert an integer to MAC address string.
    :param value: Integer value of a MAC address.
    """
    mac = f'{value:012x}'  # Convert to hex, zero-padded to 12 characters
    return ':'.join(mac[i : i + 2] for i in range(0, 12, 2))

def prepare_data(fields_list, data_list, match, mask):
    """
    Prepare match and mask buffers of type MatchParams and create HwsData out of them.
    :param fields_list: List of field name and header type tuples.
    :param data_list: List of match value and mask tuples.
    """
    for ((field, header), index) in zip(fields_list, range(len(fields_list))):
        if field in ['smac', 'dmac']:
            # Convert MAC addresses to integers.
            mac_value_int = mac_to_int(data_list[index][0])
            mac_mask_int = mac_to_int(data_list[index][1])
            # Applying mask to the value.
            mac_masked_int = mac_value_int & mac_mask_int
            # Convert result back to MAC address string format.
            mac_masked = int_to_mac(mac_masked_int)
            if header == e.MLX5DV_HWS_HEADER_INNER:
                setattr(match.inner_headers, field, mac_masked)
                setattr(mask.inner_headers, field, data_list[index][1])
            elif header == e.MLX5DV_HWS_HEADER_OUTER:
                setattr(match.outer_headers, field, mac_masked)
                setattr(mask.outer_headers, field, data_list[index][1])
        elif field in ['ipv4_src','ipv4_dst']:
            # Convert IP addresses to integers.
            ip_value_int = int(ipaddress.ip_address(data_list[index][0]))
            ip_mask_int = int(ipaddress.ip_address(data_list[index][1]))
            # Applying mask to the value.
            ip_masked = ip_value_int & ip_mask_int
            if header == e.MLX5DV_HWS_HEADER_INNER:
                setattr(match.inner_headers, field, ip_masked)
                setattr(mask.inner_headers, field, data_list[index][1])
            elif header == e.MLX5DV_HWS_HEADER_OUTER:
                setattr(match.outer_headers, field, ip_masked)
                setattr(mask.outer_headers, field, data_list[index][1])
        else:
            masked_value = data_list[index][0] & data_list[index][1]
            if header == e.MLX5DV_HWS_HEADER_INNER:
                setattr(match.inner_headers, field, masked_value)
                setattr(mask.inner_headers, field, data_list[index][1])
            elif header == e.MLX5DV_HWS_HEADER_OUTER:
                setattr(match.outer_headers, field, masked_value)
                setattr(mask.outer_headers, field, data_list[index][1])
            elif header == e.MLX5DV_HWS_HEADER_NONE:
                setattr(match.misc_parameters, field, masked_value)
                setattr(mask.misc_parameters, field, data_list[index][1])
    return HwsData(match, mask)

def field_name_to_enum_list(field):
    field2enum = {
        'l3_type': [e.MLX5DV_HWS_FNAME_L3_TYPE],
        'l4_type': [e.MLX5DV_HWS_FNAME_L4_TYPE],
        'eth_type': [e.MLX5DV_HWS_FNAME_ETH_TYPE],
        'smac': [e.MLX5DV_HWS_FNAME_SMAC_47_16, e.MLX5DV_HWS_FNAME_SMAC_15_0],
        'dmac': [e.MLX5DV_HWS_FNAME_DMAC_47_16, e.MLX5DV_HWS_FNAME_DMAC_15_0],
        'first_prio': [e.MLX5DV_HWS_FNAME_FIRST_PRIO],
        'first_cfi': [e.MLX5DV_HWS_FNAME_FIRST_CFI],
        'first_vid': [e.MLX5DV_HWS_FNAME_FIRST_VID],
        'first_cvlan': [e.MLX5DV_HWS_FNAME_FIRST_CVLAN],
        'first_svlan': [e.MLX5DV_HWS_FNAME_FIRST_SVLAN],
        'second_prio': [e.MLX5DV_HWS_FNAME_SECOND_PRIO],
        'second_cfi': [e.MLX5DV_HWS_FNAME_SECOND_CFI],
        'second_vid': [e.MLX5DV_HWS_FNAME_SECOND_VID],
        'second_cvlan': [e.MLX5DV_HWS_FNAME_SECOND_CVLAN],
        'second_svlan': [e.MLX5DV_HWS_FNAME_SECOND_SVLAN],
        'ip_version': [e.MLX5DV_HWS_FNAME_IP_VERSION],
        'ip_frag': [e.MLX5DV_HWS_FNAME_IP_FRAG],
        'ip_sport': [e.MLX5DV_HWS_FNAME_IP_SPORT],
        'ip_dport': [e.MLX5DV_HWS_FNAME_IP_DPORT],
        'ip_tcp_sport': [e.MLX5DV_HWS_FNAME_IP_TCP_SPORT],
        'ip_tcp_dport': [e.MLX5DV_HWS_FNAME_IP_TCP_DPORT],
        'ip_tcp_flags': [e.MLX5DV_HWS_FNAME_IP_TCP_FLAGS],
        'ip_udp_sport': [e.MLX5DV_HWS_FNAME_IP_UDP_SPORT],
        'ip_udp_dport': [e.MLX5DV_HWS_FNAME_IP_UDP_DPORT],
        'ipv4_len': [e.MLX5DV_HWS_FNAME_IPV4_LEN],
        'ipv4_src': [e.MLX5DV_HWS_FNAME_IPV4_SRC],
        'ipv4_dst': [e.MLX5DV_HWS_FNAME_IPV4_DST],
        'ipv6_dst': [e.MLX5DV_HWS_FNAME_IPV6_DST_127_96, e.MLX5DV_HWS_FNAME_IPV6_DST_95_64,
                     e.MLX5DV_HWS_FNAME_IPV6_DST_63_32, e.MLX5DV_HWS_FNAME_IPV6_DST_31_0],
        'ipv6_src': [e.MLX5DV_HWS_FNAME_IPV6_SRC_127_96, e.MLX5DV_HWS_FNAME_IPV6_SRC_95_64,
                     e.MLX5DV_HWS_FNAME_IPV6_SRC_63_32, e.MLX5DV_HWS_FNAME_IPV6_SRC_31_0],
        'metadata_reg_c_0': [e.MLX5DV_HWS_FNAME_REG_C0],
        'metadata_reg_c_1': [e.MLX5DV_HWS_FNAME_REG_C1],
        'metadata_reg_c_2': [e.MLX5DV_HWS_FNAME_REG_C2],
        'metadata_reg_c_3': [e.MLX5DV_HWS_FNAME_REG_C3],
        'metadata_reg_c_4': [e.MLX5DV_HWS_FNAME_REG_C4],
        'metadata_reg_c_5': [e.MLX5DV_HWS_FNAME_REG_C5],
        'metadata_reg_c_6': [e.MLX5DV_HWS_FNAME_REG_C6],
        'metadata_reg_c_7': [e.MLX5DV_HWS_FNAME_REG_C7],
        'metadata_reg_c_8': [e.MLX5DV_HWS_FNAME_REG_C8],
        'metadata_reg_c_9': [e.MLX5DV_HWS_FNAME_REG_C9],
        'metadata_reg_c_10': [e.MLX5DV_HWS_FNAME_REG_C10],
        'metadata_reg_c_11': [e.MLX5DV_HWS_FNAME_REG_C11],
        'metadata_reg_a': [e.MLX5DV_HWS_FNAME_REG_A],
        'metadata_reg_b': [e.MLX5DV_HWS_FNAME_REG_B],
        'src_qp': [e.MLX5DV_HWS_FNAME_SOURCE_QP],
        'random_num': [e.MLX5DV_HWS_FNAME_RANDOM_NUM],
    }
    return field2enum[field]

def prepare_item(fields_list, data_list):
    """
    Prepare HwsData and HwsField list and Create HwsItem out of them.
    :param fields_list: List of field name and header type tuples.
    :param data_list: List of match value and mask tuples.
    """
    match = MatchParams()
    mask = MatchParams()
    data = prepare_data(fields_list, data_list, match, mask)
    fields = []
    for (field, header) in fields_list:
        fields_list = field_name_to_enum_list(field)
        for f in fields_list:
            fields.append(HwsField(f, header, e.MLX5DV_HWS_FIELD_MATCH_OP_EQ,
                                   e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                                   match.get_field_offset(f, header)))
    return HwsItem(fields, len(fields), data)

def create_eth_type_item(eth_type=u.PacketConsts.ETHER_TYPE_IPV4):
    match = MatchParams()
    mask = MatchParams()
    mask.outer_headers.eth_type = 0xffff
    match.outer_headers.eth_type = eth_type
    data = HwsData(match, mask)
    field_eth = HwsField(e.MLX5DV_HWS_FNAME_ETH_TYPE, e.MLX5DV_HWS_HEADER_OUTER,
                         e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                         match.get_field_offset(e.MLX5DV_HWS_FNAME_ETH_TYPE,
                                                e.MLX5DV_HWS_HEADER_OUTER))
    item = HwsItem([field_eth], 1, data)
    return item

def create_sipv4_item(ip_val=u.PacketConsts.SRC_IP, ip_mask="255.255.255.255",
                      header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on src IPv4.
    """
    match = MatchParams()
    mask = MatchParams()
    int_ip_val = int(ipaddress.ip_address(ip_val))
    int_ip_mask = int(ipaddress.ip_address(ip_mask))
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ipv4_src = int_ip_val & int_ip_mask
        mask.outer_headers.ipv4_src = ip_mask
    else:
        match.inner_headers.ipv4_src = int_ip_val & int_ip_mask
        mask.inner_headers.ipv4_src = ip_mask
    data = HwsData(match, mask)
    field_sip = HwsField(e.MLX5DV_HWS_FNAME_IPV4_SRC, header_type, e.MLX5DV_HWS_FIELD_MATCH_OP_EQ,
                         e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                         match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV4_SRC, header_type))
    item = HwsItem([field_sip], len([field_sip]), data)
    return item

def skip_if_hws_not_supported(ctx):
    hca_caps = u.query_hca_caps(ctx)
    wqe_support = hca_caps.wqe_based_flow_table_update_cap
    if not wqe_support:
        raise unittest.SkipTest('HWS WQE based insertion cap is not supported')

def create_dipv4_item(ip_val=u.PacketConsts.DST_IP, ip_mask="255.255.255.255",
                      header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on dst IPv4.
    """
    match = MatchParams()
    mask = MatchParams()
    int_ip_val = int(ipaddress.ip_address(ip_val))
    int_ip_mask = int(ipaddress.ip_address(ip_mask))
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ipv4_dst = int_ip_val & int_ip_mask
        mask.outer_headers.ipv4_dst = ip_mask
    else:
        match.inner_headers.ipv4_dst = int_ip_val & int_ip_mask
        mask.inner_headers.ipv4_dst = ip_mask
    data = HwsData(match, mask)
    field_dip = HwsField(e.MLX5DV_HWS_FNAME_IPV4_DST, header_type, e.MLX5DV_HWS_FIELD_MATCH_OP_EQ,
                         e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                         match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV4_DST, header_type))
    item = HwsItem([field_dip], len([field_dip]), data)
    return item

def create_sipv6_item(ip_val=u.PacketConsts.SRC_IP6, ip_mask=("ffff:" * 8)[:-1],
                      header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on src IPv6.
    """
    match = MatchParams()
    mask = MatchParams()
    int_ip_val = int(ipaddress.ip_address(ip_val))
    int_ip_mask = int(ipaddress.ip_address(ip_mask))
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ipv6_src = str(ipaddress.IPv6Address(int_ip_val & int_ip_mask))
        mask.outer_headers.ipv6_src = ip_mask
    else:
        match.inner_headers.ipv6_src = str(ipaddress.IPv6Address(int_ip_val & int_ip_mask))
        mask.inner_headers.ipv6_src = ip_mask

    data = HwsData(match, mask)
    fields = []
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_SRC_127_96, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_SRC_127_96, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_SRC_95_64, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_SRC_95_64, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_SRC_63_32, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_SRC_63_32, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_SRC_31_0, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_SRC_31_0, header_type))
    item = HwsItem(fields, len(fields), data)
    return item

def create_dipv6_item(ip_val=u.PacketConsts.DST_IP6, ip_mask=("ffff:" * 8)[:-1],
                      header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on dst IPv6.
    """
    match = MatchParams()
    mask = MatchParams()
    int_ip_val = int(ipaddress.ip_address(ip_val))
    int_ip_mask = int(ipaddress.ip_address(ip_mask))
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ipv6_dst = str(ipaddress.IPv6Address(int_ip_val & int_ip_mask))
        mask.outer_headers.ipv6_dst = ip_mask
    else:
        match.inner_headers.ipv6_dst = str(ipaddress.IPv6Address(int_ip_val & int_ip_mask))
        mask.inner_headers.ipv6_dst = ip_mask
    data = HwsData(match, mask)
    fields = []
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_DST_127_96, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_DST_127_96, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_DST_95_64, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_DST_95_64, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_DST_63_32, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_DST_63_32, header_type))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IPV6_DST_31_0, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE),
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV6_DST_31_0, header_type))
    item = HwsItem(fields, len(fields), data)
    return item

def create_l3_type_item(l3_type=u.PacketConsts.IP_V4, l3_mask=0b11,
                        header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on L3 type.
    """
    match = MatchParams()
    mask = MatchParams()
    # L3 Type 0=None, 1=IPv4, 2=IPv6, 3=reserved - 2bit
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.l3_type = 1 if l3_type == u.PacketConsts.IP_V4 else 2
        mask.outer_headers.l3_type = l3_mask
    else:
        match.inner_headers.l3_type = 1 if l3_type == u.PacketConsts.IP_V4 else 2
        mask.inner_headers.l3_type = l3_mask
    data = HwsData(match, mask)
    field_l3_type = HwsField(e.MLX5DV_HWS_FNAME_L3_TYPE, header_type,
                             e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                             match.get_field_offset(e.MLX5DV_HWS_FNAME_L3_TYPE, header_type))
    item = HwsItem([field_l3_type], len([field_l3_type]), data)
    return item

def create_l4_type_item(l4_type=socket.IPPROTO_UDP, l4_mask=0xf,
                        header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on L4 type.
    """
    match = MatchParams()
    mask = MatchParams()
    # L4 type 0=None, 1=TCP, 2=UDP, 3=ICMP, 4-15=reserved - 4bits
    if l4_type == socket.IPPROTO_TCP:
        l4_val = 1
    elif l4_type == socket.IPPROTO_UDP:
        l4_val = 2
    elif l4_type == socket.IPPROTO_ICMP:
        l4_val = 3
    else:
        raise TypeError(f'Unsupported l4 type: {l4_type}.')
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.l4_type = l4_val & l4_mask
        mask.outer_headers.l4_type = l4_mask
    else:
        match.inner_headers.l4_type = l4_val & l4_mask
        mask.inner_headers.l4_type = l4_mask
    data = HwsData(match, mask)
    field_l4_type = HwsField(e.MLX5DV_HWS_FNAME_L4_TYPE, header_type,
                             e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                             match.get_field_offset(e.MLX5DV_HWS_FNAME_L4_TYPE, header_type))
    item = HwsItem([field_l4_type], len([field_l4_type]), data)
    return item

def create_tcp_port_items(sport=u.PacketConsts.SRC_PORT, dport=u.PacketConsts.DST_PORT,
                          sport_mask=0xffff, dport_mask=0xffff,
                          header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on TCP src and dst ports.
    """
    match = MatchParams()
    mask = MatchParams()
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ip_tcp_sport = sport & sport_mask
        mask.outer_headers.ip_tcp_sport = sport_mask
        match.outer_headers.ip_tcp_dport = dport & dport_mask
        mask.outer_headers.ip_tcp_dport = dport_mask
    else:
        match.inner_headers.ip_tcp_sport = sport & sport_mask
        mask.inner_headers.ip_tcp_sport = sport_mask
        match.inner_headers.ip_tcp_dport = dport & dport_mask
        mask.inner_headers.ip_tcp_dport = dport_mask
    data = HwsData(match, mask)
    fields = []
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_TCP_SPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_TCP_SPORT, header_type)))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_TCP_DPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_TCP_DPORT, header_type)))
    item = HwsItem(fields, len(fields), data)
    return item

def create_outer_tcp_sport_range_item(sport_value=u.PacketConsts.SRC_PORT, sport_mask=0xffff,
                                      sport_last=0xffff):
    match = MatchParams()
    mask = MatchParams()
    last = MatchParams()
    match.outer_headers.ip_tcp_sport = sport_value
    mask.outer_headers.ip_tcp_sport = sport_mask
    last.outer_headers.ip_tcp_sport = sport_last
    data = HwsData(match, mask, last)
    field = HwsField(e.MLX5DV_HWS_FNAME_IP_TCP_SPORT, e.MLX5DV_HWS_HEADER_OUTER,
                     e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                     match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_TCP_SPORT,
                                            e.MLX5DV_HWS_HEADER_OUTER))
    return HwsItem([field], len([field]), data)

def create_outer_tcp_dport_range_item(dport_value=u.PacketConsts.DST_PORT, dport_mask=0xffff,
                                      dport_last=0xffff):
    match = MatchParams()
    mask = MatchParams()
    last = MatchParams()
    match.outer_headers.ip_tcp_dport = dport_value
    mask.outer_headers.ip_tcp_dport = dport_mask
    last.outer_headers.ip_tcp_dport = dport_last
    data = HwsData(match, mask, last)
    field = HwsField(e.MLX5DV_HWS_FNAME_IP_TCP_DPORT, e.MLX5DV_HWS_HEADER_OUTER,
                     e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                     match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_TCP_DPORT,
                                            e.MLX5DV_HWS_HEADER_OUTER))
    return HwsItem([field], len([field]), data)

def create_port_items(sport=u.PacketConsts.SRC_PORT, dport=u.PacketConsts.DST_PORT,
                      sport_mask=0xffff, dport_mask=0xffff,
                      header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on generic src and dst ports.
    """
    match = MatchParams()
    mask = MatchParams()
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ip_sport = sport & sport_mask
        mask.outer_headers.ip_sport = sport_mask
        match.outer_headers.ip_dport = dport & dport_mask
        mask.outer_headers.ip_dport = dport_mask
    else:
        match.inner_headers.ip_sport = sport & sport_mask
        mask.inner_headers.ip_sport = sport_mask
        match.inner_headers.ip_dport = dport & dport_mask
        mask.inner_headers.ip_dport = dport_mask
    data = HwsData(match, mask)
    fields = []
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_SPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_SPORT, header_type)))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_DPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_DPORT, header_type)))
    item = HwsItem(fields, len(fields), data)
    return item

def create_udp_port_items(sport=u.PacketConsts.SRC_PORT, dport=u.PacketConsts.DST_PORT,
                          sport_mask=0xffff, dport_mask=0xffff,
                          header_type=e.MLX5DV_HWS_HEADER_OUTER):
    """
    Creates HwsItem to match on UDP src and dst ports.
    """
    match = MatchParams()
    mask = MatchParams()
    if header_type == e.MLX5DV_HWS_HEADER_OUTER:
        match.outer_headers.ip_udp_sport = sport &sport_mask
        mask.outer_headers.ip_udp_sport = sport_mask
        match.outer_headers.ip_udp_dport = dport & dport_mask
        mask.outer_headers.ip_udp_dport = dport_mask
    else:
        match.inner_headers.ip_udp_sport = sport &sport_mask
        mask.inner_headers.ip_udp_sport = sport_mask
        match.inner_headers.ip_udp_dport = dport & dport_mask
        mask.inner_headers.ip_udp_dport = dport_mask
    data = HwsData(match, mask)
    fields = []
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_UDP_SPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_UDP_SPORT, header_type)))
    fields.append(HwsField(e.MLX5DV_HWS_FNAME_IP_UDP_DPORT, header_type,
                           e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                           match.get_field_offset(e.MLX5DV_HWS_FNAME_IP_UDP_DPORT, header_type)))
    item = HwsItem(fields, len(fields), data)
    return item

def requires_jumbo_support(func):
    def func_wrapper(instance):
        hca_caps2 = u.get_hca_caps_layout_2(instance)
        jumbo_support = hca_caps2.format_select_dw_8_6_ext
        if not jumbo_support:
            raise unittest.SkipTest('Device doesn\'t support jumbo matching')
        return func(instance)
    return func_wrapper

def create_vlan_hdr(tpi=u.PacketConsts.VLAN_TPID, prio=u.PacketConsts.VLAN_PRIO,
                    cfi=u.PacketConsts.VLAN_CFI, vid=u.PacketConsts.VLAN_ID):
    # |   TPID  |  PRI   |  CFI  |   VID   |
    # | 2 bytes | 3 bits | 1 bit | 12 bits |
    return (tpi << 16) + (prio << 13) + (cfi << 12) + vid

def is_bit_set(num, pos):
    return (num >> pos) & 1

def check_wqe_based_flow_table_update_support(instance):
    hca_caps = u.query_hca_caps(Mlx5Context(Mlx5DVContextAttr(), instance.dev_name))
    if not hca_caps.wqe_based_flow_table_update_cap:
        raise unittest.SkipTest('Device doesn\'t support wqe based flow table update.')

def requires_compare_support(func):
    def func_wrapper(instance):
        check_wqe_based_flow_table_update_support(instance)
        wqe_based_flow_table_caps = u.query_wqe_based_flow_table_caps(instance)
        if not is_bit_set(wqe_based_flow_table_caps.ste_format_gen_wqe, MATCH_4DW_RANGE):
            raise unittest.SkipTest('Device doesn\'t support gen WQE Compare match format.')
        return func(instance)
    return func_wrapper

def requires_range_support(func):
    def func_wrapper(instance):
        check_wqe_based_flow_table_update_support(instance)
        wqe_based_flow_table_caps = u.query_wqe_based_flow_table_caps(instance)
        if not is_bit_set(wqe_based_flow_table_caps.ste_format_gen_wqe, MATCH_RANGE):
            raise unittest.SkipTest('Device doesn\'t support gen WQE Range match format.')
        return func(instance)
    return func_wrapper

def requires_rtc_hash_split_table_support(func):
    def func_wrapper(instance):
        wqe_based_flow_table_caps = u.query_wqe_based_flow_table_caps(instance)
        if not wqe_based_flow_table_caps.rtc_hash_split_table:
            raise unittest.SkipTest('FW doesn\'t support hash split tables')
        return func(instance)
    return func_wrapper

def requires_rtc_linear_lookup_tables_support(func):
    def func_wrapper(instance):
        wqe_based_flow_table_caps = u.query_wqe_based_flow_table_caps(instance)
        linear_lookup = wqe_based_flow_table_caps.rtc_linear_lookup_table
        access_idx_mode = wqe_based_flow_table_caps.access_index_mode
        if not linear_lookup or not access_idx_mode:
            raise unittest.SkipTest('FW doesn\'t support linear lookup tables')
        return func(instance)
    return func_wrapper

def requires_adv_rdma_support(func):
    def func_wrapper(instance):
        hca_caps = u.query_hca_caps(Mlx5Context(Mlx5DVContextAttr(), instance.dev_name))
        if not hca_caps.adv_rdma_cap:
            raise unittest.SkipTest('Advanced RDMA cap is not supported')
        else:
            adv_rdma_caps = u.query_adv_rdma_caps(instance)
            rx_support = adv_rdma_caps.rdma_transport_rx_flow_table_properties.ft_support
            tx_support = adv_rdma_caps.rdma_transport_tx_flow_table_properties.ft_support
            if not rx_support or not tx_support:
                raise unittest.SkipTest('FW doesn\'t support advanced RDMA tables')
        return func(instance)
    return func_wrapper

def get_nvconfig(test, param):
    try:
        cmd_out = subprocess.check_output(['mlxconfig', '-d', f'{test.pci_dev}', 'q'])
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"command '{e.cmd}' return with error (code {e.returncode}): {e.output}")
    words = str(cmd_out).split()
    param_idx = words.index(param)
    if "True" in words[param_idx + 1]:
        return 1
    elif "False" in words[param_idx + 1]:
        return 0
    else:
        return int(words[param_idx + 1])

def parametrized(dec):
    """
    Decorator for creating decorators with parameters
    """
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer

@parametrized
def requires_nvconfig(func, param, vals):
    if not isinstance(vals, list):
        vals = [vals]
    def inner(instance):
        try:
            if get_nvconfig(instance, param) not in vals:
                raise unittest.SkipTest(f'Supported for {param} = {vals}.')
        except Exception as ex:
            raise unittest.SkipTest(f'Failed query nvconfig: {ex}.')
        return func(instance)
    return inner

def requires_geneve_tlv_sample(func):
    def func_wrapper(instance):
        hca_caps = u.query_hca_caps(Mlx5Context(Mlx5DVContextAttr(), instance.dev_name))
        if not hca_caps.geneve_tlv_sample:
            raise unittest.SkipTest('FW doesn\'t support geneve tlv sample')
        return func(instance)
    return func_wrapper

def requires_geneve_tlv_option_offset(func):
    def func_wrapper(instance):
        hca_caps = u.query_hca_caps(Mlx5Context(Mlx5DVContextAttr(), instance.dev_name))
        if not hca_caps.geneve_tlv_option_offset:
            raise unittest.SkipTest('FW doesn\'t support geneve tlv option offset')
        return func(instance)
    return func_wrapper

def create_generic_aso_wqe(agr_obj, data=None):
    """
    Create and return aso WQE with always-hit conditions.
    The ASO Access WQE allows to access the ASO Context: Read and Conditional Update (both can
        be performed by a single WQE).
    ASO Data segmant - 64 Bytes of data.
    ASO Ctrl segment includes-
        For Read, the memory location where the ASO context will be written.
        The data modified and its location (offset) in the ASO context.
        The update condition:
            - 2 sub-conditions. (each has operand, offset, data and mask)
            - The logic operand between the sub-conditions.
    :param data: 64 bytes of data to put in ASO Data segmant or None if in read only mode.
    """
    aso_data = AsoData(buffer=data)
    # va_r - Address within l_key to scatter the ASO Context ,bit 0 is the read_enable flag.
    # lkey - Local Memory Key where the ASO Context will be scattered.
    # data_mask_mode - BYTEWISE_64BYTE.
    # data_mask - Mask defining which bits/bytes will be updated,
    #             each bit in the mask represents a Byte in the data input.
    # condition_x_operand - 1 for ALWAYS_TRUE or 0 for ALWAYS_FALSE.
    # condition_operand - 1 for LOGICAL_OR.
    cond_op = data is not None
    aso_ctrl = AsoCtrl(va_r=(agr_obj.mr.buf | 1), l_key=agr_obj.mr.lkey,
                       data_mask=0xffffffffffffffff, data_mask_mode=1,
                       condition_0_operand=cond_op, condition_1_operand=cond_op,
                       condition_operand=1)
    seg = AsoWqeDataSeg(ctrl=aso_ctrl, data=aso_data)
    return Wqe([seg]), len(seg)


class Mlx5DvHwsResources(Mlx5DrTirResources):
    def __init__(self, dev_name, ib_port, gid_index=0, wc_flags=0, msg_size=1024,
                 qp_count=1, server=False):
        self.server = server
        self.ib_port = ib_port
        super().__init__(dev_name=dev_name, ib_port=ib_port, gid_index=gid_index,
                         wc_flags=wc_flags, msg_size=msg_size, qp_count=qp_count, server=server)
        self.attr = self.ctx.query_device()
        self.create_hws_context()

    def create_cq(self):
        """
        Mlx5DrResources creating extended CQ (CQEX), while Mlx5DrTirResources is
        creating a regular CQ.
        """
        Mlx5DrResources.create_cq(self=self)

    def create_mr(self):
        """
        Initializes self.mr with an MR of length msg_size * MAX_DIFF_PACKETS
        """
        self.mr = MR(self.pd, self.msg_size * u.MAX_DIFF_PACKETS, pe.IBV_ACCESS_LOCAL_WRITE)

    def create_hws_context(self, queues=NUM_OF_QUEUES, queue_size=DEFAULT_QUEUE_SIZE):
        ctx_attr = HwsContextAttr(pd=self.pd, queues=queues, queue_size=queue_size)
        self.hws_ctx = HwsContext(self.ctx, ctx_attr)

    def create_table(self, level=1, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX, ib_port=None):
        attr = HwsTableAttr(table_type, level, ib_port)
        return HwsTable(self.hws_ctx, attr)

    def create_matcher(self, table, matcher_templates, action_templates=None, prio=1,
                       mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE):
        action_templates = action_templates if action_templates else [a.HwsActionTemplate([], 0 )]
        matcher_attr = HwsMatcherAttr(priority=prio, resource_mode=mode)
        return HwsMatcher(table, matcher_templates, len(matcher_templates),
                          action_templates, len(action_templates),
                          matcher_attr)

    def create_root_fwd_rule(self, items, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX, level=1):
        match_templates = [HwsMatchTemplate(self.hws_ctx, items, len(items))]
        action_templates = [a.HwsActionTemplate([], 0)]
        matcher_attr = HwsMatcherAttr(priority=1,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        root_matcher = HwsMatcher(self.root_table, match_templates, len(match_templates),
                                  action_templates, len(action_templates), matcher_attr)
        rule_attr = HwsRuleAttr(user_data=bytes(8))
        table = self.create_table(level=level, table_type=table_type)
        action_flags = self.table_type_to_action_type(table_type, root=True)
        fwd_tbl_a = a.HwsActionDestTable(self.hws_ctx, table, action_flags)
        rule = HwsRule(root_matcher, 0, items, 0, [], fwd_tbl_a, rule_attr,
                       ctx=self.hws_ctx)
        return root_matcher, rule, table

    @staticmethod
    def table_type_to_action_type(table_type, root=False):
        if root:
            table2action = {
            e.MLX5DV_HWS_TABLE_TYPE_NIC_RX : e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
            e.MLX5DV_HWS_TABLE_TYPE_NIC_TX : e.MLX5DV_HWS_ACTION_FLAG_NIC_TX,
            e.MLX5DV_HWS_TABLE_TYPE_FDB_RX : e.MLX5DV_HWS_ACTION_FLAG_FDB,
            e.MLX5DV_HWS_TABLE_TYPE_FDB_TX : e.MLX5DV_HWS_ACTION_FLAG_FDB,
            e.MLX5DV_HWS_TABLE_TYPE_FDB : e.MLX5DV_HWS_ACTION_FLAG_FDB
        }
            return table2action[table_type] | e.MLX5DV_HWS_ACTION_FLAG_ROOT
        else:
            table2action = {
            e.MLX5DV_HWS_TABLE_TYPE_NIC_RX : e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
            e.MLX5DV_HWS_TABLE_TYPE_NIC_TX : e.MLX5DV_HWS_ACTION_FLAG_NIC_TX,
            e.MLX5DV_HWS_TABLE_TYPE_FDB_RX : e.MLX5DV_HWS_ACTION_FLAG_FDB_RX,
            e.MLX5DV_HWS_TABLE_TYPE_FDB_TX : e.MLX5DV_HWS_ACTION_FLAG_FDB_TX,
            e.MLX5DV_HWS_TABLE_TYPE_FDB : e.MLX5DV_HWS_ACTION_FLAG_FDB
        }
            return table2action[table_type]

    def vport_to_ib_port(self, vport_num):
        for i in range(1, self.attr.phys_port_cnt + 1):
            try:
                dv_port_attr = Mlx5Context.query_mlx5_port(self.ctx, i)
            except PyverbsRDMAError as ex:
                if ex.error_code in [errno.EOPNOTSUPP, errno.EPROTONOSUPPORT]:
                    continue  # Port is inactive
                raise ex
            if vport_num == dv_port_attr.vport:
                return i
        raise PyverbsError( f'Failed to find IB port for vport {vport_num}.')

    def vf_to_ib_port(self, vf_name):
        """
        Converts VF name to IB port number
        """
        for i in range(2):
            dev = os.listdir(f'/sys/class/infiniband/{self.dev_name}/device/virtfn{i}/infiniband/')
            if dev[0].strip() == vf_name:
                return self.vport_to_ib_port(i + 1)

    def init_steering_resources(self, items=None, root_items=None, at=None,
                                table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX,
                                matcher_attr=None):
        """
        Init the basic hws resources. Basic test layout root_tbl -> tbl (1 matcher)
        :param items: The items to use in the matchers. If not set, use sipv4 item.
        :param root_items: items to use in the root matcher. If not set,
                           use items for both root and non-root matchers
        :param at: List of ActionTemplates
        :param table_type: The tables type.
        :param mather_attr: The attribute with which to create the matcher.
        """
        root_items = root_items if root_items is not None else []
        items = items if (items != None) else [create_sipv4_item()]
        self.action_templates = at if at else [a.HwsActionTemplate([], 0)]
        if table_type in [e.MLX5DV_HWS_TABLE_TYPE_FDB_RX, e.MLX5DV_HWS_TABLE_TYPE_FDB_TX]:
            root_table_type = e.MLX5DV_HWS_TABLE_TYPE_FDB
        else:
            root_table_type = table_type
        self.root_table = self.create_table(level=0, table_type=root_table_type)
        # Create root rule.
        self.root_matcher, self.root_dest_tbl_rule, self.table = \
            self.create_root_fwd_rule(root_items, table_type=table_type)
        # Create table 1 matcher.
        self.matcher_templates = [HwsMatchTemplate(self.hws_ctx, items, len(items))]
        matcher_attr = matcher_attr if matcher_attr else HwsMatcherAttr(priority=0)
        self.matcher = HwsMatcher(self.table, self.matcher_templates, len(self.matcher_templates),
                                  self.action_templates, len(self.action_templates), matcher_attr)

    def create_counter(self, action_flag=e.MLX5DV_HWS_ACTION_FLAG_NIC_RX, bulk_log_size=2):
        """
        Creates a HWS resource with type counter, and a counter action.
        :param action_flag: Action flags enum
        :param bulk_log_size: log size of the resource bulk
        """
        attr = HwsResourceAttr(resource_type=e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                               bulk_log_size=bulk_log_size)
        resource_counter = HwsResource(self.hws_ctx, attr)
        resource_counter.get_devx_obj()
        counter_a = a.HwsActionCounter(self.hws_ctx, [resource_counter], 1, action_flag)
        return counter_a, resource_counter

    def query_counter(self, resource_counter, flow_counter_id, counter_offset=0):
        """
        Perform query of packets and bytes of a counter.
        :param resource_counter: HwsResource object
        :param flow_counter_id: devx_obj_id of the resource_counter
        :param counter_offset: resource offset from the HwsResource bulk
        """
        from tests.mlx5_prm_structs import QueryFlowCounterIn, TrafficCounter, QueryFlowCounterOut
        BULK_COUNTER_SIZE = 512
        num_counters = BULK_COUNTER_SIZE if counter_offset != 0 else 0
        query_in = QueryFlowCounterIn(num_of_counters=num_counters, flow_counter_id=flow_counter_id)
        tc_len = len(TrafficCounter())
        outlen = len(QueryFlowCounterOut()) + ((num_counters - 1) * tc_len \
            if num_counters > 1 else 0)
        counter_out = QueryFlowCounterOut(resource_counter.query(query_in, outlen))
        # Get the start index of the needed TC in the output mailbox
        start_idx = len(QueryFlowCounterOut()) + tc_len * (counter_offset - 1)
        stats = TrafficCounter(bytes(counter_out)[start_idx:start_idx + tc_len])
        return stats.packets, stats.octets


class Mlx5HwsTestCase(Mlx5RDMATestCase):
    """
    Test various functionalities of the mlx5 HW steering rules class.
    """
    def setUp(self):
        super().setUp()
        skip_if_hws_not_supported(Mlx5Context(Mlx5DVContextAttr(), self.dev_name))
        self.server = Mlx5DvHwsResources(**self.dev_info, server=True,
                                         wc_flags=pe.IBV_WC_EX_WITH_FLOW_TAG)
        self.client = Mlx5DvHwsResources(**self.dev_info, server=False,
                                         wc_flags=pe.IBV_WC_EX_WITH_FLOW_TAG)
        self.attr = self.server.ctx.query_device()

    def tearDown(self):
        if self.server:
            self.server.hws_ctx.close()
            self.server.ctx.close()
        if self.client:
            self.client.hws_ctx.close()
            self.client.ctx.close()

    def is_cx6lx(self):
        return self.attr.vendor_id == MELLANOX_VENDOR_ID and \
            self.attr.vendor_part_id == u.VendorPartID.CX6LX

    def is_cx6dx(self):
        return self.attr.vendor_id == MELLANOX_VENDOR_ID and \
            self.attr.vendor_part_id == u.VendorPartID.CX6DX

    def is_bf2(self):
        return self.attr.vendor_id == MELLANOX_VENDOR_ID and \
            self.attr.vendor_part_id == u.VendorPartID.BF2

    def is_greater_than_cx7(self):
        return self.attr.vendor_id == MELLANOX_VENDOR_ID and \
            self.attr.vendor_part_id >= u.VendorPartID.CX7

    def get_flow_meter_reg_id(self):
        """
        Queries hca caps for supported reg C indexes for flow meter.
        :param dv_ctx: Devx context
        :return: List of supported reg C indexes
        """
        bit_regs = 0
        if self.is_greater_than_cx7():
            caps= u.get_hca_caps_layout_2(self)
            bit_regs = caps.return_reg_id
        if bit_regs == 0:
            caps = u.get_qos_caps(Mlx5Context(Mlx5DVContextAttr(), self.dev_name))
            bit_regs = caps.flow_meter_reg_id
        result = []
        for i in range(0, bit_regs.bit_length()):
            if bit_regs & (1 << i) != 0:
                result.append(i)
        return result


class HwsBasicAPITest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws basic API functionality class.
    """
    def get_max_queue_size(self):
        hca_caps = u.query_hca_caps(self.server.ctx)
        log_max_qp_sz = hca_caps.log_max_qp_sz
        SEND_MAX_WQES_PER_RULE = 32
        max_queue_size = (2 ** log_max_qp_sz) / SEND_MAX_WQES_PER_RULE
        return max_queue_size

    def test_mlx5dv_hws_context_api(self):
        """
        Open contexts with different configurations.
        """
        # The common cases is to open a queue per thread.
        queues_arr = [0, 4, 16, 32]
        max_queue_size = self.get_max_queue_size()
        queue_size_arr = [1, max_queue_size/2, max_queue_size]
        # Create contexts
        ctx_arr = []
        for queues, queue_size in zip(queues_arr, queue_size_arr):
            ctx_attr = HwsContextAttr(pd=self.server.pd, queues=queues, queue_size=queue_size)
            ctx_arr.append(HwsContext(self.server.ctx, ctx_attr))
        for ctx in ctx_arr:
            ctx.close()

    def test_mlx5dv_hws_context_api_bad_queue_size(self):
        """
        Bad flow - Open contexts with sq_size > MAX_QP_SIZE.
        Expected syndrome - BAD_PARAM  | 0x27A8E4 |  create_sq: sq_size > MAX_QP_SIZE.
        """
        queues = 4
        max_queue_size = self.get_max_queue_size()
        ctx_attr = HwsContextAttr(pd=self.server.pd, queues=queues, queue_size=(max_queue_size + 1))
        with self.assertRaises(PyverbsError) as ex:
            HwsContext(self.server.ctx, ctx_attr)

    @requires_adv_rdma_support
    def test_mlx5dv_hws_table_api(self):
        """
        Creates tables of all types. Root not supporting FDB TX and FDB RX.
        """
        tables = []
        table_types = range(7)
        non_root_table_levels = range(1, 7)
        for t_type in [e.MLX5DV_HWS_TABLE_TYPE_NIC_RX,
                       e.MLX5DV_HWS_TABLE_TYPE_NIC_TX,
                       e.MLX5DV_HWS_TABLE_TYPE_FDB]:
            tables.append(self.server.create_table(0, t_type))
        for t_type in [e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX,
                       e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX]:
            tables.append(self.server.create_table(0, t_type, self.ib_port))
        # Non root
        self.server.create_table(1, e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX, self.ib_port)
        for t_level in non_root_table_levels:
            for t_type in table_types:
                if t_type in [e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX,
                              e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX]:
                    tables.append(self.server.create_table(t_level, t_type, self.ib_port))
                else:
                    tables.append(self.server.create_table(t_level, t_type))

    def test_mlx5dv_hws_mt(self):
        """
        Create empty match template.
        Create match template with 1 item (sip).
        Create match template with 10 items (10 sips).
        """
        mt_attr = HwsMatchTemplateAttr()
        # empty matcher
        mt1 = HwsMatchTemplate(self.server.hws_ctx, [], 0, mt_attr)
        items = [create_sipv4_item()]
        # 1 item
        mt2 = HwsMatchTemplate(self.server.hws_ctx, items, len(items), mt_attr)
        # 10 items
        mt3 = HwsMatchTemplate(self.server.hws_ctx, items*10, len(items*10), mt_attr)

    def test_mlx5dv_hws_bad_comp_mask_mt(self):
        """
        Bad flow. Try creating MT with bad comp mask. Expect exception.
        """
        bad_mt_attr = HwsMatchTemplateAttr(1)
        with self.assertRaises(PyverbsError):
            HwsMatchTemplate(self.server.hws_ctx, [], 0, bad_mt_attr)

    def test_mlx5dv_hws_bad_comp_mask_at(self):
        """
        Bad flow. Try creating AT with bad comp mask. Expect exception.
        """
        bad_at_attr = a.HwsActionTemplateAttr(1)
        with self.assertRaises(PyverbsError):
            a.HwsActionTemplate([], 0, bad_at_attr)

    def test_mlx5dv_hws_at(self):
        """
        Create empty action template.
        Create action template with 1 action.
        Create action template with 30 actions.
        """
        at_attr = a.HwsActionTemplateAttr()
        # empty matcher
        at1 = a.HwsActionTemplate([], 0, at_attr)
        # 1 action
        tag_a = a.HwsActionTag(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at2 = a.HwsActionTemplate([tag_a], 1, at_attr)
        # 30 1-slot actions is maximum (10 STEs each one has 3 slots for actions)
        max_1slot_actions = 30
        at3 = a.HwsActionTemplate([tag_a] * max_1slot_actions, max_1slot_actions, at_attr)

    def test_mlx5dv_hws_matcher_attr_root(self):
        """
        To create a root matcher there is only one valid configuration: resource mode RULE,
        insert by hash and distribute by hash.
        """
        self.server.init_steering_resources()
        matcher_attr_root = HwsMatcherAttr(
            priority=0,
            insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_HASH,
            distribute_mode=e.MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH,
            resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
            rule_log=2
        )
        matcher_root = HwsMatcher(self.server.table, self.server.matcher_templates,
                                  len(self.server.matcher_templates), self.server.action_templates,
                                  len(self.server.action_templates), matcher_attr_root)

    def test_mlx5dv_hws_matcher_attr_common_case(self):
        """
        Insert by hash, Distribute by hash (resource mode both RULE and HTABLE).
        """
        self.server.init_steering_resources()
        matcher_attrs = [HwsMatcherAttr(priority=0,
                                        resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                        rule_log=2),
                         HwsMatcherAttr(priority=0,
                                        resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE,
                                        row_log=2, col_log=2)]
        for matcher_attr in matcher_attrs:
            matcher = HwsMatcher(self.server.table, self.server.matcher_templates,
                                 len(self.server.matcher_templates), self.server.action_templates,
                                 len(self.server.action_templates), matcher_attr)

    @requires_rtc_hash_split_table_support
    def test_mlx5dv_hws_matcher_attr_index_and_hash(self):
        """
        Insert by index, Distribute by hash (if supported: caps - rtc_hash_split_table).
        """
        self.server.init_steering_resources(items=[])
        matcher_attr = HwsMatcherAttr(priority=0, insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
                                      distribute_mode=e.MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_HASH,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                      rule_log=2)
        matcher = HwsMatcher(self.server.table, self.server.matcher_templates,
                             len(self.server.matcher_templates), self.server.action_templates,
                             len(self.server.action_templates), matcher_attr)

    @requires_rtc_linear_lookup_tables_support
    def test_mlx5dv_hws_matcher_attr_index_and_linear(self):
        """
        Insert by index, Distribute by linear (if supported: caps -  rtc_linear_lookup_table
                                               and access_index_mode).
        """
        self.server.init_steering_resources(items=[])
        matcher_attr = HwsMatcherAttr(priority=0, insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
                                      distribute_mode=e.MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                      rule_log=2)
        matcher = HwsMatcher(self.server.table, self.server.matcher_templates,
                             len(self.server.matcher_templates), self.server.action_templates,
                             len(self.server.action_templates), matcher_attr)

    def test_mlx5dv_hws_rule_size(self):
        """
        Check that get rule handle function returns positive number.
        """
        self.assertGreater(get_rule_handle_size(), 0)

    def test_mlx5dv_hws_queue_pending(self):
        """
        Verify queue pending basic functionality.
        Verify queues are empty before usage.
        Add 3 rules  - verify pending operation number is increasing accordingly.
        Drain queues - verify they are empty again.
        """
        for i in range(NUM_OF_QUEUES):
            self.assertEqual(self.server.hws_ctx.queue_pending(i), 0, f'Queue {i} should be empty.')
        items = [create_sipv4_item()]
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                  e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rules = []
        num_of_rules = 3
        for i in range(NUM_OF_QUEUES):
            for j in range(num_of_rules):
                rules.append(HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                                     HwsRuleAttr(queue_id=i, user_data=bytes(8), burst=1)))
                self.assertEqual(self.server.hws_ctx.queue_pending(i), j + 1,
                                 f'Queue {i} pending operations number is wrong')
            self.server.hws_ctx.queue_execute_op(i, e.MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC)
            for _ in range(num_of_rules):
                self.server.hws_ctx.check_op_completion(i)
            self.assertEqual(self.server.hws_ctx.queue_pending(i), 0,
                            f'Queue {i} pending operations number is wrong')

    def test_mlx5dv_hws_rule_query(self):
        """
        Verify the basic functionality of rule query function.
        Check the match tag size. Check that the tag match includes src IP and dst IP values.
        """
        sip = '1.2.3.4'
        dip = '5.6.7.8'
        items = [create_sipv4_item(ip_val=sip), create_dipv4_item(ip_val=dip)]
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                       HwsRuleAttr(user_data=bytes(8)), self.server.hws_ctx)
        rule_query = rule.rule_query()
        rule_match_tag_size = 32 # for all non jumbo matches
        self.assertEqual(rule_query.match_tag_sz, rule_match_tag_size,
                         'Rule match tag size is wrong.')
        # Convert to string
        match_tag = rule_query.match_tag.decode('utf-8')
        self.assertTrue(socket.inet_aton(sip).decode('utf-8') in match_tag,
                        'Source IP is not found in rule match tag')
        self.assertTrue(socket.inet_aton(dip).decode('utf-8') in match_tag,
                        'Destination IP is not found in rule match tag')

    @requires_rtc_hash_split_table_support
    def test_mlx5dv_rule_hash_calculate(self):
        """
        Create hash split table 1X4 matching on UDP src port.
        Create 4 rules on that matcher with 4 counters.
        Note: No need to provide values to match on, table index is a hash
        calculated from packet value of UDP src port.
        Send packets in a loop with different src UDP ports and predicting
        by hash calculation, what counter will the packet hit and comparing
        the queried counter value with the predicted counter value.
        """
        udp_fields_list = [('ip_udp_sport', e.MLX5DV_HWS_HEADER_OUTER)]
        udp_data_list = [(0, 0xffff)]
        item_udp_sport = [prepare_item(udp_fields_list, udp_data_list)]
        matcher_attr = HwsMatcherAttr(priority=0,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                      insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
                                      rule_log=2, always_hit=True)
        counter_a, resource_counter = self.server.create_counter()
        at = [a.HwsActionTemplate([counter_a], 1)]
        self.server.init_steering_resources(item_udp_sport, root_items=[create_sipv4_item()],
                                            at=at, matcher_attr=matcher_attr)
        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)

        num_actions = 4
        rules = []
        for i in range(num_actions):
            # Empty list of items
            rules.append(HwsRule(self.server.matcher, 0, [], 0,
                                 [a.HwsActionData(resource_offset=i)], dest_def_miss,
                                 HwsRuleAttr(user_data=bytes(8), rule_idx=i),
                                 ctx=self.server.hws_ctx))
        predicted_counters = [0, 0, 0, 0]
        for sport in range(16):
            udp_data_list = [(sport, 0xffff)]
            item_data = prepare_data(udp_fields_list, udp_data_list, MatchParams(), MatchParams())
            index = self.server.matcher.rule_hash_calculate([item_data], mt_idx=0)
            predicted_counters[index] += 1
            packet = u.gen_packet(self.client.msg_size, src_port=sport)
            u.hws_send_packets(self.client, [packet])
            counters_value = [self.server.query_counter(resource_counter,
                                                        resource_counter.devx_obj_id,
                                                        counter_offset=i) for i in range(4)]
            packets_counters = [value[0] for value in counters_value]
            self.assertEqual(packets_counters, predicted_counters,
                             'SW hash calculation failed to predict counter index')


class HwsActionTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws actions functionality class.
    """
    def test_mlx5dv_hws_action_tir(self):
        """
        Create TIR and recv packets using TIR action.
        """
        items = [create_sipv4_item()]
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)
        # TODO verify_rule_removal - waiting for counter

    def test_mlx5dv_hws_action_drop(self):
        """
        Create Drop actions on RX and TX and TIR action, recv packets using TIR WQ.
        Create 3 rules:
        TX rule: if src IP is 2.2.2.2 action Drop.
        RX rule: if src IP is 3.3.3.3. action Drop.
        RX rule: if eth type is IPv4 go to TIR, expecting only 1.1.1.1
        Send three packets (1.1.1.1, 2.2.2.2, 3.3.3.3).
        Expected only the third + validate data of 1.1.1.1.
        """
        self.hws_rules = []
        list_ip = []
        for i in [1, 2, 3]:
            list_ip.append('.'.join(str(i) * 4))
        tx_items = [create_sipv4_item(list_ip[1])]
        rx_items = [create_sipv4_item(list_ip[2])]
        eth_items = [create_eth_type_item()]
        self.server.init_steering_resources(root_items=eth_items)
        self.client.init_steering_resources(root_items=eth_items,
                                            table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_TX)

        rx_a = a.HwsActionDestDrop(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tx_a = a.HwsActionDestDrop(self.client.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_TX)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)

        tir_mt = [HwsMatchTemplate(self.server.hws_ctx, [], 0)]
        tir_at = [a.HwsActionTemplate([], 0 )]
        tir_matcher_attr = HwsMatcherAttr(priority=2,
                                          resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        tir_matcher = HwsMatcher(self.server.table, tir_mt, len(tir_mt), tir_at, len(tir_at),
                                 tir_matcher_attr)
        self.hws_rules.append(HwsRule(tir_matcher, 0, eth_items, 0, [], tir_a,
                                      HwsRuleAttr(user_data=bytes(8)), self.server.hws_ctx))
        self.hws_rules.append(HwsRule(self.server.matcher, 0, rx_items, 0, [], rx_a,
                                      HwsRuleAttr(user_data=bytes(8)), self.server.hws_ctx))
        self.hws_rules.append(HwsRule(self.client.matcher, 0, tx_items, 0, [], tx_a,
                                      HwsRuleAttr(user_data=bytes(8)), self.client.hws_ctx))

        packets = []
        for ip in list_ip:
            packets.append(u.gen_packet(self.server.msg_size, src_ipv4=ip))

        with self.assertRaises(PyverbsError) as ex:
            u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packets[2],
                          timeout=1)
            u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packets[1],
                          timeout=1)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packets[0])

    def test_mlx5dv_hws_action_tag(self):
        """
        Use Tag action, send packets and verify that the packet is tagged.
        """
        items = [create_sipv4_item()]
        tag_a = a.HwsActionTag(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at = [a.HwsActionTemplate([tag_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tag_value = 0x1234
        action_data = a.HwsActionData(value=tag_value)
        rule = HwsRule(self.server.matcher, 0, items, 0, [action_data], tir_a,
                       HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet],
                          tag_value=tag_value)

    def verify_counter(self, agr_ob, resource_counter, expected_num_packets, exp_msg_size=None,
                       counter_offset=0):
        rec_packets, rec_bytes = agr_ob.query_counter(resource_counter,
                                                      resource_counter.devx_obj_id,
                                                      counter_offset=counter_offset)
        self.assertEqual(rec_packets, expected_num_packets,
                         f'Incorrect counter received packet {expected_num_packets} != ' \
                         f'{rec_packets}')
        if exp_msg_size:
            # RX hws counters include FCS\VCRC in the byte count on cx6dx/lx/bf2.
            # Add extra 4 bytes to each packet.
            if agr_ob.server:
                fix_factor = 4 if self.is_cx6dx() or self.is_cx6lx() or self.is_bf2() else 0
                exp_msg_size += fix_factor
            expectd_bytes = exp_msg_size * expected_num_packets
            self.assertEqual(rec_bytes, expectd_bytes,
                             f'Incorrect counter received bytes {expectd_bytes} != {rec_bytes}')

    def test_mlx5dv_hws_action_counter(self):
        """
        Create counter action with default miss action on the client (TX),
        and a counter action with tir action on the server (RX).
        Send packets between them and verify the counter value on both sides.
        """
        items = [create_sipv4_item()]
        action_tx_flag = e.MLX5DV_HWS_ACTION_FLAG_NIC_TX
        counter_a_tx, resource_counter_tx = self.client.create_counter(action_tx_flag)
        counter_a_rx, resource_counter_rx = self.server.create_counter()

        dest_def_miss = a.HwsActionDestDefaultMiss(self.client.hws_ctx, action_tx_flag)
        tir = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at_tx = [a.HwsActionTemplate([counter_a_tx], 1)]
        at_rx = [a.HwsActionTemplate([counter_a_rx], 1)]
        self.client.init_steering_resources(items=items, at=at_tx,
                                            table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_TX)
        self.server.init_steering_resources(items=items, at=at_rx,
                                            table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        rule_tx = HwsRule(self.client.matcher, 0, items, 0, [a.HwsActionData(resource_offset=0)],
                          dest_def_miss, HwsRuleAttr(user_data=bytes(8)), ctx=self.client.hws_ctx)
        rule_rx = HwsRule(self.server.matcher, 0, items, 0,[a.HwsActionData(resource_offset=0)],
                          tir, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)
        self.verify_counter(self.client, resource_counter_tx, self.server.num_msgs, len(packet))
        self.verify_counter(self.server, resource_counter_rx, self.server.num_msgs, len(packet))

    def test_mlx5dv_hws_action_counter_root(self):
        """
        Create counter action on a root table with default miss.
        Send packets and verify the counter value.
        """
        root_table = self.server.create_table(level=0, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)

        root_items = [create_sipv4_item()]
        mt_root = [HwsMatchTemplate(self.server.hws_ctx, root_items, len(root_items))]

        action_root_flag = e.MLX5DV_HWS_ACTION_FLAG_NIC_RX | e.MLX5DV_HWS_ACTION_FLAG_ROOT
        counter_a_root, resource_counter_root = self.server.create_counter(action_root_flag)

        at_root = [a.HwsActionTemplate([counter_a_root], 1)]
        m_attr = HwsMatcherAttr(priority=0, resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        root_matcher = HwsMatcher(root_table, mt_root, len(mt_root), at_root, len(at_root), m_attr)

        # Create root rule.
        def_miss_a = a.HwsActionDestDefaultMiss(self.server.hws_ctx, action_root_flag)
        rule = HwsRule(root_matcher, 0, root_items, 0, [a.HwsActionData(resource_offset=0)],
                       def_miss_a, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)

        packet = u.gen_packet(self.server.msg_size)
        u.hws_send_packets(self.client, [packet], iters=self.server.num_msgs)
        self.verify_counter(self.server, resource_counter_root, self.server.num_msgs,
                            len(packet))

    def test_mlx5dv_hws_action_table(self):
        """
        Create a chain of tables connected using HwsActionDestTable actions on NIC RX.
        root_tbl -> tbl1 -> tbl2 -> tbl3 (TIR)
        Send packets, verify the connection using action TIR.
        """
        items = [create_sipv4_item()]
        self.server.init_steering_resources()
        tbl2 = self.server.create_table(level=2)
        tbl3 = self.server.create_table(level=3)
        mt = [HwsMatchTemplate(self.server.hws_ctx, items, len(items))]
        at = [a.HwsActionTemplate([], 0 )]
        tbl2_matcher = HwsMatcher(tbl2, mt, len(mt), at, len(at), HwsMatcherAttr(priority=0))
        tbl3_matcher = HwsMatcher(tbl3, mt, len(mt), at, len(at), HwsMatcherAttr(priority=0))
        fwd_tbl_12 = a.HwsActionDestTable(self.server.hws_ctx, tbl2,
                                          e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        fwd_tbl_23 = a.HwsActionDestTable(self.server.hws_ctx, tbl3,
                                          e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                               e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rule_12 = HwsRule(self.server.matcher, 0, items, 0, [], fwd_tbl_12,
                          HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        rule_23 = HwsRule(tbl2_matcher, 0, items, 0, [], fwd_tbl_23,
                          HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        rule_tir = HwsRule(tbl3_matcher, 0, items, 0, [],
                           tir, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)

    def test_mlx5dv_hws_action_pop_push_vlan(self):
        """
        Create rule with two push and one pop VLAN actions.
        Generate a packet 4 bytes smaller than expected, send it, and verify
        VLAN header addition by using the TIR action.
        """
        items = [create_sipv4_item()]
        push_a = a.HwsActionPushVlan(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        pop_a = a.HwsActionPopVlan(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        actions = [push_a, push_a, pop_a]
        at = [a.HwsActionTemplate(actions, len(actions))]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_data = a.HwsActionData(vlan_hdr=create_vlan_hdr())
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [push_data, push_data], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.client.msg_size - u.PacketConsts.VLAN_HEADER_SIZE,
                              with_vlan=False)
        exp_packet = u.gen_packet(self.server.msg_size, with_vlan=True)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def hws_action_modify(self, using_equeue_arg_write=False):
        """
        Create modify action with two set actions to set src mac to 88:88:88:88:88:88.
        Create rule on NIC RX with modify and TIR actions.
        Send packet and verify with TIR that src mac was modified correctly.
        param: using_equeue_arg_write: Provide the modify value using arg wrtie API if True.
                                       Otherwise, set using action data during rule creation.
        """
        from tests.mlx5_prm_structs import SetActionIn
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=0)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        action1 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                              data=0x88888888, length=ModifyFieldsLen.MAC_47_16)
        action2 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_15_0,
                              data=0x8888, length=ModifyFieldsLen.MAC_15_0)
        actions = [action1, action2]
        mdf_a = a.HwsActionModify(self.server.hws_ctx, actions, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
                                  [resource], 1)
        at = [a.HwsActionTemplate([mdf_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        if using_equeue_arg_write:
            attr = a.HwsActionEnqueueArgWriteAttr(comp_mask=0, user_data=bytes(8),
                                                  resource_offset=0, data_size=8 * len(actions),
                                                  arg_data=actions, burst=False)
            qid = 0
            mdf_a.enqueue_arg_write(resource, qid, attr)
            self.server.hws_ctx.check_op_completion(qid)
            tir_rule = HwsRule(self.server.matcher, 0, items, 0,
                               [a.HwsActionData(resource_offset=0)], tir_a,
                               HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)

        else:
            mdf_data = a.HwsActionData(modify_data=actions, resource_offset=0)
            tir_rule = HwsRule(self.server.matcher, 0, items, 0, [mdf_data], tir_a,
                               HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        exp_src_mac = bytes.fromhex("88:88:88:88:88:88".replace(':', ''))
        exp_packet = u.gen_packet(self.server.msg_size, src_mac=exp_src_mac)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def test_mlx5dv_hws_action_modify(self):
        """
        Create modify action with two set actions to set src mac to 88:88:88:88:88:88.
        Provide the values using action data for the rule.
        Create rule on NIC RX with modify and TIR actions.
        Send packet and verify with TIR that src mac was modified correctly.
        """
        self.hws_action_modify()

    def test_mlx5dv_hws_action_arg_write(self):
        """
        Create modify action with two set actions to set src mac to 88:88:88:88:88:88.
        Provide the values using enqueue arg write API.
        Create rule on NIC RX with modify and TIR actions.
        Send packet and verify with TIR that src mac was modified correctly.
        """
        self.hws_action_modify(using_equeue_arg_write=True)

    def test_mlx5dv_hws_action_decap_l2(self):
        """
        Create decapsulation action: tunnel L2 to L2 type.
        Create a rule on NIC RX with decap and TIR action.
        Send encapsulated packet and verify decapsulation.
        """
        items = [create_sipv4_item()]
        decap_a = a.HwsActionReformatTnlL2ToL2(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at = [a.HwsActionTemplate([decap_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN)
        exp_packet = u.gen_packet(self.server.msg_size - len(outer))
        packet = outer + exp_packet
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def test_mlx5dv_hws_action_encap_l2(self):
        """
        Create encapsulation action: L2 to tunnel L2 type.
        Create a rule on NIC RX with encap and TIR action.
        Send untunneled packet and verify encapsulation.
        """
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B, 0)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN)
        encap_a = a.HwsActionReformatL2ToTnlL2(self.server.hws_ctx, outer,
                                               e.MLX5DV_HWS_ACTION_FLAG_NIC_RX, [resource], 1)
        at = [a.HwsActionTemplate([encap_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        encap_data = a.HwsActionData(reformat_l2_to_tnl_l2_data=outer, resource_offset=0)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [encap_data], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size - len(outer))
        exp_packet = outer + packet
        # Skip IPv4 header checksum and UDP src port due to entropy
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet, skip_idxs=[24, 25, 34, 35])

    def create_dest_root_table(test, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX):
        """
        On root table:
            - Matcher1 with priority 1 matches on smac. Add rule with action goto
            FT1
            - Matcher2 with priority 5 matches on UDP src port. Add rule with
            action goto FT2
        On FT1:
            - Rule1 with action counter1 + goto dest root table with priority 5
        On FT2:
            - Rule2 with action counter2
        Send a packet to hit matcher1 and matcher2 (smac and sport)
        Send a packet to hit matcher1 but not matcher2 (smac and sport+1)
        Send a packet to hit matcher2 but not matcher1 (new smac and sport)
        Verify counter values.
        """
        action_flag = test.server.table_type_to_action_type(table_type, root=False)
        action_flag_root = test.server.table_type_to_action_type(table_type, root=True)
        smac_fields_list = [('smac', e.MLX5DV_HWS_HEADER_OUTER)]
        smac_data_list = [(u.PacketConsts.SRC_MAC, u.PacketConsts.MAC_MASK)]
        items_smac = [prepare_item(smac_fields_list, smac_data_list)]
        udp_fields_list = [('ip_udp_sport', e.MLX5DV_HWS_HEADER_OUTER)]
        udp_data_list = [(u.PacketConsts.SRC_PORT, 0xffff)]
        items_udp_sport = [prepare_item(udp_fields_list, udp_data_list)]
        counter_a1, resource_counter1 = test.server.create_counter(action_flag)
        counter_a2, resource_counter2 = test.server.create_counter(action_flag)
        dest_def_miss = a.HwsActionDestDefaultMiss(test.server.hws_ctx, action_flag)
        # Matcher1 with priority 1 matches on smac. Add rule with action goto FT1
        at_ctr1 = [a.HwsActionTemplate([counter_a1], 1)]
        test.server.init_steering_resources(root_items=items_smac, items=items_smac,
                                            table_type=table_type, at=at_ctr1)
        # Matcher2 with priority 5 matches on UDP src port. Add rule with action goto FT2
        mt_udp_sport = [HwsMatchTemplate(test.server.hws_ctx, items_udp_sport,
                                         len(items_udp_sport))]
        test.root_matcher2 = test.server.create_matcher(test.server.root_table,
                                                        matcher_templates=mt_udp_sport,
                                                        prio=5)
        action_dest_root = a.HwsActionDestRoot(test.server.hws_ctx, priority=5, flags=action_flag)
        test.server.table2 = test.server.create_table(table_type=table_type, level=2)
        fwd_tbl_r2 = a.HwsActionDestTable(test.server.hws_ctx, test.server.table2, action_flag_root)
        test.server.dest_tbl2_rule = HwsRule(test.root_matcher2, 0, items_udp_sport, 0, [],
                                             fwd_tbl_r2, HwsRuleAttr(user_data=bytes(8)),
                                             ctx=test.server.hws_ctx)
        # On FT1: Rule1 with action counter1 + goto dest root table with priority 5
        test.rule1 = HwsRule(test.server.matcher, 0, items_smac, 0,
                             [a.HwsActionData(resource_offset=0)], action_dest_root,
                             HwsRuleAttr(user_data=bytes(8)), ctx=test.server.hws_ctx)
        # On FT2: Rule2 with action counter2
        at_ctr2 = [a.HwsActionTemplate([counter_a2], 1)]
        test.server.matcher2 = test.server.create_matcher(test.server.table2,
                                                          matcher_templates=mt_udp_sport,
                                                          action_templates=at_ctr2,
                                                          prio=1)
        test.rule2 = HwsRule(test.server.matcher2, 0, items_udp_sport, 0,
                             [a.HwsActionData(resource_offset=0)], dest_def_miss,
                             HwsRuleAttr(user_data=bytes(8)), ctx=test.server.hws_ctx)
        packets = []
        smac_vals = [u.PacketConsts.SRC_MAC, u.PacketConsts.SRC_MAC,  '88:88:88:88:88:88']
        for smac, sport_inc in zip(smac_vals, [0, 1, 0]):
                packets.append(u.gen_packet(test.server.msg_size,
                                            src_mac=bytes.fromhex(smac.replace(':', '')),
                                            src_port=u.PacketConsts.SRC_PORT + sport_inc))
        for packet in packets:
            u.sampler_traffic(test.client, test.server, test.server.num_msgs, packet_to_send=packet)
        test.server.num_msgs = test.server.num_msgs * 2
        test.verify_counter(test.client, resource_counter1, test.server.num_msgs)
        test.verify_counter(test.client, resource_counter2, test.server.num_msgs)

    def test_mlx5dv_hws_dest_root_table_rx(self):
        """
        Verify dest root table action on RX
        """
        self.create_dest_root_table()

    def test_mlx5dv_hws_dest_root_table_tx(self):
        """
        Verify dest root table action on TX
        """
        self.create_dest_root_table(table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_TX)

    def test_mlx5dv_hws_action_decap_l3(self):
        """
        Create decapsulation action: tunnel L3 to L2 type.
        Create a rule on NIC RX with decap and TIR action.
        Send encapsulated packet and verify decapsulation.
        """
        items = [create_sipv4_item()]
        l2_hdr = u.get_l2_header()
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B, 0)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        decap_a = a.HwsActionReformatTnlL3ToL2(self.server.hws_ctx, l2_hdr,
                                               e.MLX5DV_HWS_ACTION_FLAG_NIC_RX, [resource], 1)
        at = [a.HwsActionTemplate([decap_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        decap_data = a.HwsActionData(reformat_tnl_l3_to_l2_data=l2_hdr, resource_offset=0)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [decap_data], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.GRE)
        inner = u.gen_packet(self.server.msg_size - len(outer) - len(l2_hdr), l2=False)
        packet = outer + inner
        exp_packet = l2_hdr + inner
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def test_mlx5dv_hws_action_encap_l3(self):
        """
        Create encapsulation action: L2 to tunnel L3 type.
        Create a rule on NIC RX with encap and TIR action.
        Send untunneled packet and verify encapsulation.
        """
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B, 0)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.GRE)
        encap_a = a.HwsActionReformatL2ToTnlL3(self.server.hws_ctx, outer,
                                             e.MLX5DV_HWS_ACTION_FLAG_NIC_RX, [resource], 1)
        at = [a.HwsActionTemplate([encap_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        encap_data = a.HwsActionData(reformat_l2_to_tnl_l3_data=outer, resource_offset=0)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [encap_data], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size - len(outer) + u.PacketConsts.ETHER_HEADER_SIZE)
        exp_packet = outer + u.gen_packet(self.server.msg_size - len(outer), l2=False)
        # Skip IPv4 header checksum and UDP src port due to entropy
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet, skip_idxs=[24, 25, 34, 35])

    def test_mlx5dv_hws_action_insert_header(self):
        """
        Verify insert header action using list of two insert headers:
        1. Representing encap
        2. Represeting push VLAN
        Verify both with traffic.
        """
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=1)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        encap_data = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN)
        vlan_hdr = struct.pack('!I', create_vlan_hdr())
        add_encap_hdr_attr = a.HwsActionInsertHeaderAttr(data=encap_data,
                                                         anchor=u.ActionReformatAnchor.PACKET_START,
                                                         encap=True)
        add_encap_hdr_a = a.HwsActionInsertHeader(self.server.hws_ctx, [resource], 1,
                                                  add_encap_hdr_attr,
                                                  e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        vlan_start_idx = 12
        add_vlan_hdr_attr = a.HwsActionInsertHeaderAttr(data=vlan_hdr,
                                                        anchor=u.ActionReformatAnchor.PACKET_START,
                                                        offset=vlan_start_idx)
        add_vlan_hdr_a = a.HwsActionInsertHeader(self.server.hws_ctx, [resource], 1,
                                                 add_vlan_hdr_attr,
                                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at = [a.HwsActionTemplate([add_vlan_hdr_a], 1), a.HwsActionTemplate([add_encap_hdr_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        encap_action_data = a.HwsActionData(insert_header=encap_data, resource_offset=0)

        encap_rule = HwsRule(self.server.matcher, 0, items, 1, [encap_action_data],
                             tir_a, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size - len(encap_data))
        exp_packet = encap_data + packet
        # skip_idxs to avoid UDP hash entropy (34, 35) and checksum (24, 25)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet, skip_idxs=[24, 25, 34, 35,])
        encap_rule.close()

        vlan_action_data = a.HwsActionData(insert_header=vlan_hdr, resource_offset=1)
        vlan_rule = HwsRule(self.server.matcher, 0, items, 0, [vlan_action_data],
                            tir_a, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size - u.PacketConsts.VLAN_HEADER_SIZE)
        exp_packet = u.gen_packet(self.server.msg_size, with_vlan=True)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def test_mlx5dv_hws_action_remove_header(self):
        """
        Verify remove header action representing decap.
        """
        items = [create_sipv4_item()]
        attr = a.HwsActionRemoveHeaderAttr(e.MLX5DV_HWS_ACTION_REMOVE_HEADER_TYPE_BY_HEADER,
                                           start_anchor=u.ActionReformatAnchor.PACKET_START,
                                           end_anchor=u.ActionReformatAnchor.INNER_MAC,
                                           decap=True)
        rm_hdr_a = a.HwsActionRemoveHeader(self.server.hws_ctx, attr,
                                           e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at = [a.HwsActionTemplate([rm_hdr_a], 1)]
        self.server.init_steering_resources(items=items, at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        decap_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                             HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN)
        inner_len = self.server.msg_size - len(outer)
        inner = u.gen_packet(inner_len)
        exp_packet = inner
        packet = outer + inner
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet,
                      expected_packet=exp_packet)

    def create_aso_meter_rules(self, agr_obj, table_type, reg_id):
        """
        Create and configure the ASO meter.
        Create a rule with ASO meter and goto the next table actions.
        On the next table create two rules with counter actions matching on
        green and red metadata values.
        Return counters.
        """
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_METER, 0)
        meter_resource = HwsResource(agr_obj.hws_ctx, resource_attr)
        action_flag = agr_obj.table_type_to_action_type(table_type)
        meter_a = a.HwsActionAsoMeter(agr_obj.hws_ctx, [meter_resource], 1, reg_id, action_flag)
        agr_obj.table2 = agr_obj.create_table(level=2, table_type=table_type)
        fwd_tbl_a = a.HwsActionDestTable(agr_obj.hws_ctx, agr_obj.table2, action_flag)
        at = [a.HwsActionTemplate([meter_a], 1)]
        agr_obj.init_steering_resources(items=items, at=at, table_type=table_type)
        # Configure meter
        meter_data = a.HwsActionData(init_color=e.MLX5DV_HWS_ACTION_ASO_METER_COLOR_GREEN,
                                     resource_offset=0)
        attr = HwsResourceEnqueueAsoAttr(burst=False, user_data=bytes(8), offset=0)
        aso_data = FlowMeterParams(valid=0x1, bucket_overflow=FLOW_METER_BUCKET_OVERFLOW,
                                   start_color=e.MLX5DV_HWS_ACTION_ASO_METER_COLOR_UNDEFINED,
                                   cir_mantissa=1,  # 8 * (10^9) * 1 * 2^(-6) bps
                                   cir_exponent=6)
        wqe, wqe_len = create_generic_aso_wqe(agr_obj, aso_data)
        meter_resource.enqueue_aso_wqe(1, wqe, wqe_len, attr)
        self.rules.append(HwsRule(agr_obj.matcher, 0, items, 0, [meter_data], fwd_tbl_a,
                          HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx))
        # Match on reg_c
        fields_list = [(f'metadata_reg_c_{reg_id}', e.MLX5DV_HWS_HEADER_NONE)]
        green_data_list = [(e.MLX5DV_HWS_ACTION_ASO_METER_COLOR_GREEN, 0xffffffff)]
        green_regc_items = [prepare_item(fields_list, green_data_list)]
        red_data_list = [(e.MLX5DV_HWS_ACTION_ASO_METER_COLOR_RED, 0xffffffff)]
        red_regc_items = [prepare_item(fields_list, red_data_list)]
        reg_c_mt = [HwsMatchTemplate(agr_obj.hws_ctx, green_regc_items, len(green_regc_items))]
        # Counters rules
        green_ctr_a, green_ctr_res = agr_obj.create_counter(action_flag=action_flag)
        red_ctr_a, red_ctr_res = agr_obj.create_counter(action_flag=action_flag)
        green_counter_at = [a.HwsActionTemplate([green_ctr_a], 1)]
        red_counter_at = [a.HwsActionTemplate([red_ctr_a], 1)]
        green_matcher = agr_obj.create_matcher(agr_obj.table2, reg_c_mt, green_counter_at)
        red_matcher = agr_obj.create_matcher(agr_obj.table2, reg_c_mt, red_counter_at)
        dest_def_miss = a.HwsActionDestDefaultMiss(agr_obj.hws_ctx, action_flag)
        self.rules.append(HwsRule(green_matcher, 0, green_regc_items, 0,
                                  [a.HwsActionData(resource_offset=0)], dest_def_miss,
                                  HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx))
        self.rules.append(HwsRule(red_matcher, 0, red_regc_items, 0,
                                  [a.HwsActionData(resource_offset=0)], dest_def_miss,
                                  HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx))
        return green_ctr_res, red_ctr_res

    def test_mlx5dv_hws_action_aso_meter(self):
        """
        Create rules with ASO flow meter configured to 15MBps both on RX and TX non root tables.
        Create rules with counters to count Red and Green packets. Send traffic with rate above
        30Mbps and check green and red packets counters are above 0 on TX nad RX.
        """
        regs = self.get_flow_meter_reg_id()
        self.rules = []
        green_rx_ctr_res, red_rx_ctr_res = \
            self.create_aso_meter_rules(self.server, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX,
                                        reg_id=regs[0])
        green_tx_ctr_res, red_tx_ctr_res = \
            self.create_aso_meter_rules(self.client, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_TX,
                                        reg_id=regs[1])
        # We want to send at least at 30MBps speed
        rate_limit = 30
        u.high_rate_send(self.client, u.gen_packet(self.client.msg_size), rate_limit)
        ctrs = [green_rx_ctr_res, red_rx_ctr_res, green_tx_ctr_res, red_tx_ctr_res]
        for ctr, agr_obj in zip(ctrs, [self.server] * 2 + [self.client] * 2):
            rec_packets, _ = agr_obj.query_counter(ctr, ctr.devx_obj_id)
            self.assertGreater(rec_packets, 0,
                               f'Number of packets should be greater then 0: {rec_packets}')

    def create_ct_aso_rules(self, agr_obj, table_type, reg_id, synd=CT_VALID_SYND):
        """
        Create CT ASO resource and configure it (in case of valid syndrome).
        Create a rule with CT aso and forward to the next table.
        Create counter resource and rule with counter action on the next table
        matching on the syndrome.
        Return counter.
        """
        items = [create_sipv4_item()]
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK, 0)
        ct_resource = HwsResource(agr_obj.hws_ctx, resource_attr)
        action_flag = agr_obj.table_type_to_action_type(table_type)
        ct_aso_a = a.HwsActionAsoCT(agr_obj.hws_ctx, [ct_resource], 1, reg_id, action_flag)
        at = [a.HwsActionTemplate([ct_aso_a], 1)]
        agr_obj.init_steering_resources(items=items, at=at, table_type=table_type)
        agr_obj.table2 = agr_obj.create_table(level=2, table_type=table_type)
        fwd_tbl_a = a.HwsActionDestTable(agr_obj.hws_ctx, agr_obj.table2, action_flag)
        # Configure CT ASO
        if table_type == e.MLX5DV_HWS_TABLE_TYPE_NIC_TX:
            direction = e.MLX5DV_HWS_ACTION_ASO_CT_DIRECTION_INITIATOR
        else:
            direction = e.MLX5DV_HWS_ACTION_ASO_CT_DIRECTION_RESPONDER
        ct_aso_data = a.HwsActionData(direction=direction, resource_offset=0, resource_index=0)
        if synd == CT_VALID_SYND:
            attr = HwsResourceEnqueueAsoAttr(user_data=bytes(8), offset=0)
            aso_data = ConnTrackAso(valid=1, state=1)
            wqe, wqe_len = create_generic_aso_wqe(agr_obj, aso_data)
            ct_resource.enqueue_aso_wqe(1, wqe, wqe_len, attr)
        self.rules.append(HwsRule(agr_obj.matcher, 0, items, 0, [ct_aso_data], fwd_tbl_a,
                          HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx))
        # Counter rule
        fields_list = [(f'metadata_reg_c_{reg_id}', e.MLX5DV_HWS_HEADER_NONE)]
        data_list = [(synd, 0xffffffff)]
        regc_items = [prepare_item(fields_list, data_list)]
        ctr_a, ctr_res = agr_obj.create_counter(action_flag=action_flag)
        counter_at = [a.HwsActionTemplate([ctr_a], 1)]
        reg_c_mt = [HwsMatchTemplate(agr_obj.hws_ctx, regc_items, len(regc_items))]
        matcher = agr_obj.create_matcher(agr_obj.table2, reg_c_mt, counter_at)
        dest_def_miss = a.HwsActionDestDefaultMiss(agr_obj.hws_ctx, action_flag)
        self.rules.append(HwsRule(matcher, 0, regc_items, 0,
                                  [a.HwsActionData(resource_offset=0)], dest_def_miss,
                                  HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx))
        return ctr_res


    def hws_ct_aso(self, synd=CT_VALID_SYND):
        """
        Validate CT ASO action on RX and TX sides:
        Create TX root table and forward all traffic to non root, on non
        root match on sip with actions CT ASO and forward to FT, on the next
        table match on syndrome which is stored in reg C with counter action.
        Validate counter with expected octets and number of packets.
        :param synd: Syndrome to match on reg C.
        """
        regs = self.get_flow_meter_reg_id()
        self.rules = []
        rx_ctr = self.create_ct_aso_rules(self.server, e.MLX5DV_HWS_TABLE_TYPE_NIC_RX, regs[0],
                                          synd)
        tx_ctr = self.create_ct_aso_rules(self.client, e.MLX5DV_HWS_TABLE_TYPE_NIC_TX, regs[-1],
                                          synd)

        packet = u.gen_packet(self.server.msg_size)
        u.sampler_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)
        for ctr, agr_obj in zip([rx_ctr, tx_ctr], [self.server, self.client]):
            rec_packets, _ = agr_obj.query_counter(ctr, ctr.devx_obj_id)
            self.assertEqual(rec_packets, self.server.num_msgs,
                             f'Incorrect counter received packet {self.server.num_msgs} != \
                             {rec_packets}')

    def test_mlx5dv_hws_ct_aso_valid(self):
        """
        CT ASO action on TX and RX with valid syndrome matching on reg C
        """
        self.hws_ct_aso()

    def test_mlx5dv_hws_ct_aso_invalid(self):
        """
        CT ASO action on TX and RX with invalid syndrome matching on reg C
        """
        self.hws_ct_aso(synd=CT_INVALID_SYND)

    def dest_array_rules_ft_tir(self, items, tir_a, at, mt, def_miss_a, offset):
        """
        Function creates a table, matcher, fwd action for the new table,
        a rule on the new matcher with default miss destination action
        and action data with provided offset.
        Creates a destination array action with action TIR and FT and a rule with it.
        """
        rules = []
        table = self.server.create_table(level=2, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        fwd_action = a.HwsActionDestTable(self.server.hws_ctx, table,
                                          e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        matcher = self.server.create_matcher(table, mt, at)
        rules.append(HwsRule(matcher, 0, items, 0, [a.HwsActionData(resource_offset=offset)],
                             def_miss_a, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx))
        attrs = [a.HwsActionDestArrayAttr(dest_action=tir_a),
                 a.HwsActionDestArrayAttr(dest_action=fwd_action)]
        dst_array_action = a.HwsActionDestArray(self.server.hws_ctx, 2, attrs,
                                                e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rules.append(HwsRule(self.server.matcher, 0, items, 0, [], dst_array_action,
                             HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx))
        return rules

    def test_mlx5dv_hws_dest_array_action_tir_ft(self):
        """
        Create multi destination action on NIC RX with following actions: [TIR][FT]
        On the FT create a rule with counter.
        Send packets and verify counter and verify packets on TIR.
        """
        items = [create_sipv4_item()]
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        counter_a, resource_counter = self.server.create_counter()
        at = [a.HwsActionTemplate([counter_a], 1)]
        mt = [HwsMatchTemplate(self.server.hws_ctx, items, len(items))]
        def_miss_a = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rules = self.dest_array_rules_ft_tir(items, tir_a, at, mt, def_miss_a, 0)
        packet = u.gen_packet(self.server.msg_size)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)
        self.verify_counter(self.server, resource_counter, self.client.num_msgs)

    def test_mlx5dv_hws_two_dest_array_actions_tir_ft(self):
        """
        Create 2 multi destination action on NIC RX with following actions: [TIR][FT]
        On the FTs create a rules with counters.
        Send 2 differnt packets, verify counters and verify packets on TIR.
        """
        ip2 = '1.2.3.4'
        items = [[create_sipv4_item()], [create_sipv4_item(ip_val=ip2)]]
        self.server.init_steering_resources(items=items[0])
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        counter_a, resource_counter = self.server.create_counter()
        at = [a.HwsActionTemplate([counter_a], 1)]
        mt = [HwsMatchTemplate(self.server.hws_ctx, items[0], len(items[0]))]
        def_miss_a = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rules = []
        NUM_OF_DEST_ARRAY_ACTIONS = 2
        for i in range(NUM_OF_DEST_ARRAY_ACTIONS):
            rules += self.dest_array_rules_ft_tir(items[i], tir_a, at, mt, def_miss_a, i)
        for packet in [u.gen_packet(self.server.msg_size),
                       u.gen_packet(self.server.msg_size, src_ipv4=ip2)]:
            u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)
        for offest in range(NUM_OF_DEST_ARRAY_ACTIONS):
            self.verify_counter(self.server, resource_counter, self.client.num_msgs,
                                counter_offset=offest)


class HwsMatchTemplateTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws match template functionality class.
    """
    def test_mlx5dv_hws_5tuple_ipv4_udp(self):
        """
        Verify matching on 5 tuple (IPv4 and UDP) using action TIR.
        """
        _5tuple = [create_sipv4_item(), create_dipv4_item()]
        _5tuple.append(create_l4_type_item())
        _5tuple.append(create_udp_port_items())
        self.server.init_steering_resources(items=_5tuple)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, _5tuple, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)

    @requires_jumbo_support
    def test_mlx5dv_hws_5tuple_ipv6_tcp(self):
        """
        Verify matching on 5 tuple (IPv6 and TCP) using action TIR.
        """
        _5tuple = [create_sipv6_item(), create_dipv6_item()]
        _5tuple.append(create_l3_type_item(l3_type=u.PacketConsts.IP_V6))
        _5tuple.append(create_tcp_port_items())
        self.server.init_steering_resources(items=_5tuple)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, _5tuple, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.IP_V6,
                              l4=u.PacketConsts.TCP_PROTO)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)

    @requires_jumbo_support
    def test_mlx5dv_hws_10tuple(self):
        """
        Verify matching on 5 tuple outer (IPv4 sip and dip and UDP) and 5 tuple inner
        (IPv6 sip and TCP) using action TIR. Using generic IP ports.
        """
        _10tuple = [create_sipv4_item(), create_dipv4_item()]
        _10tuple.append(create_l3_type_item())
        _10tuple.append(create_port_items(dport=u.PacketConsts.VXLAN_PORT))
        _10tuple.append(create_sipv6_item(header_type=e.MLX5DV_HWS_HEADER_INNER))
        _10tuple.append(create_l3_type_item(l3_type=u.PacketConsts.IP_V6,
                                            header_type=e.MLX5DV_HWS_HEADER_INNER))
        _10tuple.append(create_l4_type_item(l4_type=socket.IPPROTO_TCP,
                                            header_type=e.MLX5DV_HWS_HEADER_INNER))
        _10tuple.append(create_port_items(header_type=e.MLX5DV_HWS_HEADER_INNER))
        self.server.init_steering_resources(items=_10tuple)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, _10tuple, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN)
        packet = outer + u.gen_packet(self.server.msg_size - len(outer), l3=u.PacketConsts.IP_V6,
                                      l4=u.PacketConsts.TCP_PROTO)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)

    def test_mlx5dv_hws_vlan(self):
        """
        Prepare an item that contains a full first VLAN fields with SVLAN bit on,
        and a full second VLAN fields with CVLAN bit on.
        Verify matching on first and second VLANs' fields using action TIR.
        """
        fields_list = [('first_prio', e.MLX5DV_HWS_HEADER_OUTER),
                       ('first_cfi', e.MLX5DV_HWS_HEADER_OUTER),
                       ('first_vid', e.MLX5DV_HWS_HEADER_OUTER),
                       ('first_cvlan', e.MLX5DV_HWS_HEADER_OUTER),
                       ('first_svlan', e.MLX5DV_HWS_HEADER_OUTER),
                       ('second_prio', e.MLX5DV_HWS_HEADER_OUTER),
                       ('second_cfi', e.MLX5DV_HWS_HEADER_OUTER),
                       ('second_vid', e.MLX5DV_HWS_HEADER_OUTER),
                       ('second_cvlan', e.MLX5DV_HWS_HEADER_OUTER),
                       ('second_svlan', e.MLX5DV_HWS_HEADER_OUTER)]
        data_list = [(u.PacketConsts.VLAN_PRIO, 0x7), (u.PacketConsts.VLAN_CFI, 0b1),
                     (u.PacketConsts.VLAN_ID, 0xfff), (0, 0b1), (1, 0b1), # first_svlan = 1
                     (u.PacketConsts.VLAN_PRIO, 0x7), (u.PacketConsts.VLAN_CFI, 0b1),
                     (u.PacketConsts.VLAN_ID, 0xfff), (1, 0b1), (0, 0b1)] # second_cvlan = 1
        items = [prepare_item(fields_list, data_list)]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        packet = u.gen_packet(self.server.msg_size, with_vlan=True, svlan=True)
        unexp_no_vlan = u.gen_packet(self.server.msg_size)
        unexp_no_svlan = u.gen_packet(self.server.msg_size, with_vlan=True)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet, unexp_no_vlan,
                                                                           unexp_no_svlan])

    def test_mlx5dv_hws_mac_ipv4_partial_mask(self):
        """
        Verify matching on MAC and IPv4 with full and partial masks using action TIR.
        """
        # Non-root
        fields_list = [('smac', e.MLX5DV_HWS_HEADER_OUTER), ('dmac', e.MLX5DV_HWS_HEADER_OUTER),
                       ('ipv4_src',e.MLX5DV_HWS_HEADER_OUTER),
                       ('ipv4_dst', e.MLX5DV_HWS_HEADER_OUTER)]
        data_list = [(u.PacketConsts.SRC_MAC, '00:00:00:ff:ff:ff'),
                     (u.PacketConsts.DST_MAC, u.PacketConsts.MAC_MASK),
                     (u.PacketConsts.SRC_IP, '0.0.255.255'),
                     (u.PacketConsts.DST_IP, '255.255.255.255')]
        items = [prepare_item(fields_list, data_list)]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        packet_smac = "ff:ff:ff:a5:28:c8"
        packet = u.gen_packet(self.server.msg_size, src_ipv4="2.2.1.1",
                              src_mac=bytes.fromhex(packet_smac.replace(':', '')))
        unexp_packet = u.gen_packet(self.server.msg_size, src_ipv4="2.2.2.1",
                                    src_mac=bytes.fromhex(packet_smac.replace(':', '')))
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet, unexp_packet])

    def test_mlx5dv_hws_udp_sport_dport_partial_mask_root(self):
        """
        Match on UDP sport and dport with partial masks on root table.
        """
        fields_list = [('ip_udp_sport', e.MLX5DV_HWS_HEADER_OUTER),
                       ('ip_udp_dport', e.MLX5DV_HWS_HEADER_OUTER)]
        data_list = [(u.PacketConsts.SRC_PORT, 0xff),
                     (u.PacketConsts.DST_PORT, 0xfff)]
        items = [prepare_item(fields_list, data_list)]
        self.server.init_steering_resources(items=items, root_items=items)

        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # SRC_PORT is 1234 (or 0x04D2) and DST_PORT is 5678 (or 0x162E).
        src_port = 0xffD2
        dst_port = 0xf62E
        packet = u.gen_packet(self.server.msg_size, src_port=src_port, dst_port=dst_port)
        unexp_src_packet = u.gen_packet(self.server.msg_size, src_port=src_port + 1,
                                        dst_port=dst_port)
        unexp_dst_packet = u.gen_packet(self.server.msg_size, src_port=src_port,
                                        dst_port=dst_port + 1)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs,[packet, unexp_src_packet,
                                                                          unexp_dst_packet])

    def test_mlx5dv_hws_sipv4_inner_outer(self):
        """
        Verify partial matching on src IPv4 address, inner and outer using action TIR.
        """
        # Non-root
        fields_list = [('ipv4_src', e.MLX5DV_HWS_HEADER_OUTER),
                       ('ipv4_src', e.MLX5DV_HWS_HEADER_INNER)]
        data_list = [(u.PacketConsts.SRC_IP, '255.255.255.0'),
                     (u.PacketConsts.SRC_IP, '255.255.255.0')]
        items = [prepare_item(fields_list, data_list)]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN,
                                    src_ip="1.1.1.2")
        packet = outer + u.gen_packet(self.server.msg_size - len(outer), src_ip="1.1.1.3")
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet)

    def compare_sip_value(self, match_op, expected_packets, unexpected_packets):
        """
        Create HwsItem with outer src ipv4 HwsField and provided matching operation,
        and HwsData with src_ipv4 value = '2.2.2.2'.
        Verify matching only on the expected packets using action TIR.
        :params: match_op: HwsField matching operation.
        :params: expected_packets: List of packets that supposed to be matched.
        :params: unexpected_packets: List of packets that are not supposed to be matched.
        """
        match = MatchParams()
        mask = MatchParams()
        match.outer_headers.ipv4_src = '2.2.2.2'
        # Compare supports only full masks.
        mask.outer_headers.ipv4_src = '255.255.255.255'
        data = HwsData(match, mask)
        header_type = e.MLX5DV_HWS_HEADER_OUTER
        field_sip = HwsField(e.MLX5DV_HWS_FNAME_IPV4_SRC, header_type, match_op,
                             e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                             match.get_field_offset(e.MLX5DV_HWS_FNAME_IPV4_SRC, header_type))
        items = [HwsItem([field_sip], len([field_sip]), data)]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        for packet in expected_packets:
            packets_to_send = [packet] + unexpected_packets
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, packets_to_send,
                              expected_packet=packet)

    @requires_compare_support
    def test_mlx5dv_hws_compare_all_operations(self):
        """
        Create HwsItem with outer src ipv4 HwsField and different matching operations,
        and HwsData with value = '2.2.2.2'.
        Send three packets with ipv4 addresses of '2.2.2.1', '2.2.2.2' and '2.2.2.3',
        Verify matching using action TIR for each operation.
        """
        packet_lt = u.gen_packet(self.server.msg_size, src_ipv4='2.2.2.1')
        packet_eq = u.gen_packet(self.server.msg_size, src_ipv4='2.2.2.2')
        packet_gt = u.gen_packet(self.server.msg_size, src_ipv4='2.2.2.3')

        test_cases = [(e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, [packet_eq], [packet_lt, packet_gt]),
                      (e.MLX5DV_HWS_FIELD_MATCH_OP_NE, [packet_lt, packet_gt], [packet_eq]),
                      (e.MLX5DV_HWS_FIELD_MATCH_OP_LT, [packet_lt], [packet_eq, packet_gt]),
                      (e.MLX5DV_HWS_FIELD_MATCH_OP_LE, [packet_lt, packet_eq], [packet_gt]),
                      (e.MLX5DV_HWS_FIELD_MATCH_OP_GT, [packet_gt], [packet_lt, packet_eq]),
                      (e.MLX5DV_HWS_FIELD_MATCH_OP_GE, [packet_gt, packet_eq], [packet_lt])]

        for match_op, exp_packets, unexp_packets in test_cases:
            self.compare_sip_value(match_op, exp_packets, unexp_packets)

    @requires_compare_support
    def test_mlx5dv_hws_compare_sip_inner_outer_field(self):
        """
        Create HwsItem with outer src ipv4 HwsField and matching type-to-field
        with the inner src ipv4 with equal operation.
        Send 1st packet with sip(inner) = 1.1.1.1, sip(outer) = 1.1.1.1,
        Send 2nd packet with sip(inner) = 2.2.2.2, sip(outer) = 1.1.1.1,
        Verify matching only on the first packet using action TIR.
        """
        field_sip = HwsField(e.MLX5DV_HWS_FNAME_IPV4_SRC, e.MLX5DV_HWS_HEADER_OUTER,
                             e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_FIELD,
                             fname_b=e.MLX5DV_HWS_FNAME_IPV4_SRC,
                             header_b=e.MLX5DV_HWS_HEADER_INNER)
        items = [HwsItem([field_sip], len([field_sip]))]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        outer = u.gen_outer_headers(self.server.msg_size, tunnel=u.TunnelType.VXLAN,
                                    src_ip="1.1.1.1")
        packet = outer + u.gen_packet(self.server.msg_size - len(outer), src_ipv4="1.1.1.1")
        unexp_packet = outer + u.gen_packet(self.server.msg_size - len(outer), src_ipv4="2.2.2.2")
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet, unexp_packet])

    def calc_hit_ratio_on_counter(self, compare_random_item):
        """
        Insert RX rule with the provided compare_random_item and counter action.
        Send 100k packets and calculate hit ratio on this rule.
        :param compare_random_item: random_num compare item to match.
        :return: Hit ratio.
        """
        counter_a, resource_counter = self.server.create_counter()
        action_templates = [a.HwsActionTemplate([counter_a], 1)]
        self.server.init_steering_resources(items=[compare_random_item], at=action_templates)

        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rule = HwsRule(self.server.matcher, 0, [compare_random_item], 0,
                       [a.HwsActionData(resource_offset=0)], dest_def_miss,
                       HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)

        NUM_PACKETS_100K = 100000
        packet = u.gen_packet(self.server.msg_size)
        for _ in range(NUM_PACKETS_100K):
            u.hws_send_packets(self.client, [packet])
        rec_packets, _ = self.server.query_counter(resource_counter, resource_counter.devx_obj_id)
        return (rec_packets / NUM_PACKETS_100K)

    @requires_compare_support
    def test_mlx5dv_hws_compare_random(self):
        """
        Compare random to value with LT operation to catch 35% of traffic.
        Use counter action to verify.
        Note - although compare supports only fields of size 32 bits, with full mask, random
               field of size 16 bits is an exception.
        """
        TRAFFIC_RATIO = 0.35
        match = MatchParams()
        mask = MatchParams()
        match.misc_parameters.random_num = round(TRAFFIC_RATIO * 0xffff)
        mask.misc_parameters.random_num = 0xffff
        data = HwsData(match, mask)

        header_type = e.MLX5DV_HWS_HEADER_NONE
        field = HwsField(e.MLX5DV_HWS_FNAME_RANDOM_NUM, e.MLX5DV_HWS_HEADER_NONE,
                         e.MLX5DV_HWS_FIELD_MATCH_OP_LT, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                         match.get_field_offset(e.MLX5DV_HWS_FNAME_RANDOM_NUM,
                                                e.MLX5DV_HWS_HEADER_NONE))

        compare_random_item = HwsItem([field], len([field]), data)
        packet_ratio = self.calc_hit_ratio_on_counter(compare_random_item)
        self.assertAlmostEqual(packet_ratio, TRAFFIC_RATIO, None,
                               'Counted number of packets is not close to expected.', 0.005)

    @requires_range_support
    def test_mlx5dv_hws_range_basic(self):
        """
        Create a range matcher on TCP source port from 50 till 100.
        Send packets in and out the range, verify with TIR.
        """
        items = [create_outer_tcp_sport_range_item(50, 0xffff, 100)]
        self.server.init_steering_resources(items=items)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        # Traffic
        src_ports = [50, 75, 100, 49, 101]
        packets = []
        for src_port in src_ports:
            packets.append(u.gen_packet(self.server.msg_size, src_port=src_port))

        for i in range(3):
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs,
                              [packets[i], packets[3], packets[4]])

    def two_counter_rules_validate(self, items1, items2, packets, pkt_mul=1):
        """
        Creates two rules with counter actions using the provided items.
        Sends the provided packets and verifyies the counters.
        :param items1: Items for the 1st MT.
        :param items2: Items for the 2nd MT.
        :param packets: Packets to send for verification of the rules.
        :param pkt_mul: Indecates how many packets should be counted.
        """
        # Root
        self.server.root_table = self.server.create_table(level=0,
                                                          table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        self.server.root_matcher, self.server.root_dest_tbl_rule, self.server.table = \
            self.server.create_root_fwd_rule([], table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Non-root
        mt1 = HwsMatchTemplate(self.server.hws_ctx, items1, len(items1))
        mt2 = HwsMatchTemplate(self.server.hws_ctx, items2, len(items2))
        matcher_templates = [mt1, mt2]

        counter_a, resource_counter = self.server.create_counter()
        action_templates = [a.HwsActionTemplate([counter_a], 1)]

        matcher_attr = HwsMatcherAttr(priority=0, row_log=1, col_log=1)
        matcher = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                        action_templates, len(action_templates), matcher_attr)

        action_rx_flag = e.MLX5DV_HWS_ACTION_FLAG_NIC_RX
        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx, action_rx_flag)
        rule1 = HwsRule(matcher, 0, items1, 0, [a.HwsActionData(resource_offset=0)],
                        dest_def_miss, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        rule2 = HwsRule(matcher, 1, items2, 0, [a.HwsActionData(resource_offset=1)],
                        dest_def_miss, HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)

        for _ in range(self.client.num_msgs):
            u.hws_send_packets(self.client, packets)
        HwsActionTest.verify_counter(self, self.server, resource_counter,
                                     self.client.num_msgs * pkt_mul, counter_offset=0)
        HwsActionTest.verify_counter(self, self.server, resource_counter,
                                     self.client.num_msgs * pkt_mul, counter_offset=1)

    @requires_range_support
    def test_mlx5dv_hws_range_subrange(self):
        """
        Create two rules with range matching and action counter.
            first rule: one field subrange with action counter1.
            second rule: one field range with action counter2.
        Verify counters with one packet each.
        """
        items1 = [create_outer_tcp_sport_range_item(u.PacketConsts.SRC_PORT, 0xffff,
                                                    u.PacketConsts.SRC_PORT + 10)]
        items2 = [create_outer_tcp_sport_range_item(u.PacketConsts.SRC_PORT, 0xffff,
                                                    u.PacketConsts.SRC_PORT + 20)]
        packets = []
        for diff in [-1, 5, 15, 21]:
            packets.append(u.gen_packet(self.server.msg_size,
                                        src_port=u.PacketConsts.SRC_PORT + diff))
        self.two_counter_rules_validate(items1, items2, packets)

    @requires_range_support
    def test_mlx5dv_hws_range_singular_part_mask(self):
        """
        Create two rules with action counter.
            first rule: src IPv4 '1.1.1.1' mask 255.255.255.255, src port 0-0xffff.
            second rule: src IPv4 '1.1.1.1' mask 255.0.0.0, src port 0-0xffff.
        Verify counters with one packet each.
        """
        items1 = [create_sipv4_item(ip_mask="255.255.255.255"), create_outer_tcp_sport_range_item()]
        items2 = [create_sipv4_item(ip_mask="255.0.0.0"), create_outer_tcp_sport_range_item()]
        packets = []
        exp_sip = '1.2.2.2'
        unexp_sip = '2.1.1.1'
        for ip in [u.PacketConsts.SRC_IP, exp_sip, unexp_sip]:
            packets.append(u.gen_packet(self.server.msg_size, src_ipv4=ip))
        self.two_counter_rules_validate(items1, items2, packets)

    def range_subrange_singular(self, ip1, ip2):
        """
        Create two rules with action counter.
            Subrange: src_ip = ip1 and TCP src port 0-0x1000.
            Range: src_ip = ip2 and TCP src port 0-0xffff.
        Verify counters with 3 packets with TCP src port values from range and subrange.
        """
        items1 = [create_sipv4_item(ip1), create_outer_tcp_sport_range_item(0, 0xffff, 0x1000)]
        items2 = [create_sipv4_item(ip2), create_outer_tcp_sport_range_item(0, 0xffff, 0xffff)]
        packets = []
        for src_ip, src_port in zip([ip1] * 3 + [ip2] * 3,
                                    [0, 0x0100, 0x1000, 0x1001, 0x1f00, 0xffff]):
            packets.append(u.gen_packet(self.server.msg_size, l4=u.PacketConsts.TCP_PROTO,
                                        src_ipv4=src_ip, src_port=src_port))
        self.two_counter_rules_validate(items1, items2, packets, pkt_mul=3)

    @requires_range_support
    def test_mlx5dv_hws_range_subrange_singular_same_sip(self):
        """
        Create two rules
            Subrange: src_ip = 1.1.1.1 and TCP src port 0-0x1000 with action CTR1.
            Range: src_ip = 1.1.1.1 and TCP src port 0-0xffff with action CTR2.
        Verify counters with 3 packets with TCP src port values from range and subrange.
        """
        self.range_subrange_singular(u.PacketConsts.SRC_IP, u.PacketConsts.SRC_IP)

    @requires_range_support
    def test_mlx5dv_hws_range_subrange_singular_diff_sip(self):
        """
        Create two rules
            Subrange: src_ip = 1.1.1.1 and TCP src port 0-0x1000 with action CTR1.
            Range: src_ip = 2.2.2.2 and TCP src port 0-0xffff with action CTR2.
        Verify counters with 3 packets with TCP src port values from range and subrange.
        """
        self.range_subrange_singular(u.PacketConsts.SRC_IP, u.PacketConsts.DST_IP)

    @requires_range_support
    def test_mlx5dv_hws_range_complex(self):
        """
        Create two rules with action counter.
            first rule: src_ip, TCP proto, range s_port 0-0xffff.
            second rule: dst_ip, TCP proto, range d_port 0-0xffff.
        Verify counters with 3 packets.
        """
        items1 = [create_sipv4_item(), create_l4_type_item(l4_type=socket.IPPROTO_TCP),
                  create_outer_tcp_sport_range_item(sport_value=0)]
        items2 = [create_dipv4_item(), create_l4_type_item(l4_type=socket.IPPROTO_TCP),
                  create_outer_tcp_dport_range_item(dport_value=0)]
        packets = []
        unexp_sip = '1.1.1.2'
        for sip, sport in zip([u.PacketConsts.SRC_IP, unexp_sip] * 3,
                              [0, u.PacketConsts.SRC_PORT, 0xffff] * 2):
            packets.append(u.gen_packet(self.server.msg_size, l4=u.PacketConsts.TCP_PROTO,
                                        src_ipv4=sip, src_port=sport))
        self.two_counter_rules_validate(items1, items2, packets, pkt_mul=3)


class HwsMatcherTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws matching functionality class.
    """
    def create_small_matcher_verify_rules(self,
                                          resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE):
        """
        Creates a small matcher for 2 rules using rule mode or 4 rules using table
        mode and tries to add 5 rules. Two rules must fit, 4 may fit, 5 - at least
        one must return completion with error.
        """
        # Root
        self.server.root_table = self.server.create_table(level=0,
                                                          table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Create root rule.
        self.server.root_matcher, self.server.root_dest_tbl_rule, self.server.table = \
            self.server.create_root_fwd_rule([], table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)

        items = []
        for i in range(5):
            items += [prepare_item([('ipv4_src', e.MLX5DV_HWS_HEADER_OUTER)],
                                   [('.'.join(str(i+1) * 4), '255.255.255.255')])]
        matcher_templates = [HwsMatchTemplate(self.server.hws_ctx, [items[0]], 1)]
        action_templates = [a.HwsActionTemplate([], 0)]
        if resource_mode == e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE:
            matcher_attr = HwsMatcherAttr(priority=0,
                                          resource_mode=resource_mode,
                                          rule_log=1)
        else:
            matcher_attr = HwsMatcherAttr(priority=0,
                                          resource_mode=resource_mode,
                                          row_log=1, col_log=1)
        matcher = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                             action_templates, len(action_templates), matcher_attr)

        num_rules = 5
        rules = []
        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        for i in range(num_rules):
            # Creating HwsRule without hws_ctx will not poll for completions.
            rules.append(HwsRule(matcher, 0, [items[i]], 0, [], dest_def_miss,
                                 HwsRuleAttr(user_data=bytes(8))))
            try:
                self.server.hws_ctx.check_op_completion(0)
            except PyverbsError:
                # Two rules have to fit in 2x2 table or a 2-rules table
                if i < 2:
                    raise PyverbsError(f'Failed to add {i+1} rules to table that'
                                       ' should fit them.')
                # Adding more rules than the table can fit is expected to fail
                return
        raise PyverbsError(f'Sucessfully sent {num_rules} rules in table '
                           'that should not fit them.')

    def test_mlx5dv_hws_matchers_rule_based_bad_flow(self):
        """
        Bad flow - Create non root matcher using rule based mode for 2 rules.
        Try creating 5 rules. Expect at least one completion with error.
        """
        self.create_small_matcher_verify_rules()

    def test_mlx5dv_hws_matchers_table_based_bad_flow(self):
        """
        Bad flow - Create non root matcher 2x2 using table based mode.
        Try creating 5 rules. Expect at least one completion with error.
        """
        self.create_small_matcher_verify_rules(e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_HTABLE)

    @requires_rtc_hash_split_table_support
    def test_mlx5dv_hws_hash_split_table(self):
        """
        Create hash split table 1X4 matching on UDP src port.
        Create 4 rules on that matcher with 4 counters.
        Send packets in a loop with different src UDP ports until all the
        counters are greater than zero.
        """
        items = [create_udp_port_items()]
        matcher_attr = HwsMatcherAttr(priority=0,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                      insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
                                      rule_log=2, always_hit=True)
        counter_a, resource_counter = self.server.create_counter()
        at = [a.HwsActionTemplate([counter_a], 1)]
        self.server.init_steering_resources(items, at=at, matcher_attr=matcher_attr)
        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        num_actions = 4
        rules = []
        for i in range(num_actions):
            rules.append(HwsRule(self.server.matcher, 0, items, 0,
                                 [a.HwsActionData(resource_offset=i)], dest_def_miss,
                                 HwsRuleAttr(user_data=bytes(8), rule_idx=i),
                                 ctx=self.server.hws_ctx))
        for sport in range(0xffff):
            packet = u.gen_packet(self.client.msg_size, src_port=sport)
            u.hws_send_packets(self.client, [packet], iters=1)
            values = [self.server.query_counter(resource_counter,
                                                resource_counter.devx_obj_id,
                                                counter_offset=i)
                     for i in range(num_actions)]
            if all(val[0] != 0 for val in values):
                return
        raise PyverbsError('Failed to hit all the rules in hash split table')

    @requires_rtc_linear_lookup_tables_support
    def test_mlx5dv_hws_linear_lookup_table(self):
        """
        Linear lookup table feature is when we insert rules by index and
        distribute by linear (using the rec C value).
        Table root:
            Rule - Go to table 1.
        Table 1:
            Rule - udp sport=u.PacketConsts.SRC_PORT, modify reg C to 0x17 and
            go to table 2.
            Rule - udp sport=u.PacketConsts.SRC_PORT + 1, modify reg C to
            0xffff and go to table 2.
        Table 2:
            Create linear lookup matcher of size 1 X 0xffff, create 2 rules
            with counter1 and counter2 to indexes 0x17 and 0xffff.
        Send 2 packets to hit 2 rules.
        Send 1 miss packet.
        Verify counters.
        """
        from tests.mlx5_prm_structs import SetActionIn
        items = [create_udp_port_items()]
        mod_pat = SetActionIn(action_type=SET_ACTION,
                              field=ModifyFields.META_DATA_REG_C_3,
                              length=16, offset=16)
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=1)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        modify_action = a.HwsActionModify(self.server.hws_ctx, [mod_pat],
                                          e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
                                          [resource], 1)
        at = [a.HwsActionTemplate([modify_action], 1)]
        matcher_attr = HwsMatcherAttr(priority=0, row_log=2, col_log=2)
        self.server.init_steering_resources(items=items, at=at,
                                            matcher_attr=matcher_attr)
        self.server.table2 = self.server.create_table(level=2)
        table_a = a.HwsActionDestTable(self.server.hws_ctx, self.server.table2,
                                       e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        regc_values = [0x17, 0xffff]
        mod_data = [SetActionIn(data=val) for val in regc_values]
        rules = []
        num_rules = 2
        for i in range(num_rules):
            rule_items = [create_udp_port_items(sport=u.PacketConsts.SRC_PORT + i)]
            mdf_data = a.HwsActionData(modify_data=[mod_data[i]], resource_offset=i)
            rules.append(HwsRule(self.server.matcher, 0, rule_items, 0, [mdf_data], table_a,
                                 HwsRuleAttr(user_data=bytes(8)),
                                 self.server.hws_ctx))
        # Table 2, the linear lookup table.
        mt = HwsMatchTemplate(self.server.hws_ctx, [], 0)
        counter_a, resource_counter = self.server.create_counter()
        at = [a.HwsActionTemplate([counter_a], 1)]
        matcher_attr2 = HwsMatcherAttr(priority=0,
                                       resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                       insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX,
                                       distribute_mode=e.MLX5DV_HWS_MATCHER_DISTRIBUTE_BY_LINEAR,
                                       rule_log=16)
        linear_matcher = HwsMatcher(self.server.table2, [mt], 1, at, 1, matcher_attr2)
        dest_def_miss = a.HwsActionDestDefaultMiss(self.server.hws_ctx,
                                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        for i, rule_idx in enumerate(regc_values):
            rules.append(HwsRule(linear_matcher, 0, [], 0,
                                 [a.HwsActionData(resource_offset=i)], dest_def_miss,
                                 HwsRuleAttr(user_data=bytes(8), rule_idx=rule_idx),
                                 ctx=self.server.hws_ctx))
        # Send two packets to hit each counter, and one that hits nothing.
        packets = [u.gen_packet(self.client.msg_size, src_port=u.PacketConsts.SRC_PORT + i)
                   for i in range(num_rules + 1)]
        u.hws_send_packets(self.client, packets, iters=1)
        for i in range(num_rules):
            num_packets, _ = self.server.query_counter(resource_counter,
                                                       resource_counter.devx_obj_id,
                                                       counter_offset=i)
            self.assertEqual(num_packets, 1,
                             f'Linear lookup failed with counter {num_packets} instead of 1')

    def test_mlx5dv_hws_matcher_attach_at_basic(self):
        """
        Verify attaching action templates to a matcher.
        On non root table create matcher with 1 mt and 1 empty at,
        maximum number of *additional* action templates = 2.
        Attach to the matcher new at with tag action, Verify with traffic.
        Attach to the matcher new at with counter action, Verify counter.
        Note- FW extended matcher doesn't support additional at, so we must use only 1 mt!
        """
        items = [create_sipv4_item()]
        self.server.init_steering_resources(items=items)
        # AT
        action_templates = [a.HwsActionTemplate([], 0)]
        # Matcher
        matcher_attr = HwsMatcherAttr(priority=0, max_num_of_at_attach=2)
        self.server.matcher = HwsMatcher(self.server.table, self.server.matcher_templates,
                                         len(self.server.matcher_templates), action_templates,
                                         len(action_templates), matcher_attr)
        # TIR
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        # Attach 1st AT
        tag_a = a.HwsActionTag(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        at = a.HwsActionTemplate([tag_a], 1)
        self.server.matcher.attach_at(at)
        new_at_idx = 1
        # Rule
        tag_value = 0x1234
        rule = HwsRule(self.server.matcher, 0, items, new_at_idx,
                       [a.HwsActionData(value=tag_value)], tir_a, HwsRuleAttr(user_data=bytes(8)),
                       ctx=self.server.hws_ctx)
        # Traffic
        packet = u.gen_packet(self.server.msg_size)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet],
                          tag_value=tag_value)
        rule.close()
        # Attach 2nd AT
        counter, resource_counter = self.server.create_counter()
        at = a.HwsActionTemplate([counter], 1)
        self.server.matcher.attach_at(at)
        new_at_idx = 2
        # Rule
        rule = HwsRule(self.server.matcher, 0, items, new_at_idx,
                       [a.HwsActionData(resource_offset=0)], tir_a, HwsRuleAttr(user_data=bytes(8)),
                       ctx=self.server.hws_ctx)
        # Traffic
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet])
        HwsActionTest.verify_counter(self, self.server, resource_counter, self.server.num_msgs)

    def resize_matcher(self, action_templates, actions_data, packets, exp_packets, tags):
        """
        m{i} - matcher of size i
        r{i} - rule: if ip = i.i.i.i do actions (given as params) and goto TIR.
        Create m2. Add r1, r2 to m2
        Create m4. Set resize m2 -> m4. Move r1 and r2 to m4. Delete m2. Add r3 to m4.
        Create m8. Set resize m4 -> m8. Move r1 and r3 to m8. Delete r2 and m4. Add r4 to m8.
        Create m16. Set resize m8 -> m16. Move r1 and r4 to m16. Delete m8 (including r3 inside).
        Check r1 and r4 with traffic.
        :param action_templates: List of HwsActionTemplate containing the actions to perform.
        :param actions_data: List of HwsActionData corresponding to the actions in the templates.
        :param packets: List of lists of packets to send.
        :param exp_packets: List of expected packet, one for each list in packets.
        :param tags: List of expected tags to verify on TIR.
        """
        # Root
        self.server.root_table = self.server.create_table(level=0,
                                                          table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Create root rule.
        self.server.root_matcher, self.server.root_dest_tbl_rule, self.server.table = \
            self.server.create_root_fwd_rule([], table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Non-root
        items = []
        for i in range(4):
            items += [prepare_item([('ipv4_src', e.MLX5DV_HWS_HEADER_OUTER)],
                                   [('.'.join(str(i+1) * 4), '255.255.255.255')])]
        matcher_templates = [HwsMatchTemplate(self.server.hws_ctx, [items[0]], 1)]

        # Matcher size 2
        m2_attr = HwsMatcherAttr(priority=0, rule_log=1, resizable=True,
                                 resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        m2 = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                        action_templates, len(action_templates), m2_attr)

        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        r1 = HwsRule(m2, 0, [items[0]], 0, actions_data[0], tir_a, HwsRuleAttr(user_data=bytes(8)),
                     ctx=self.server.hws_ctx)
        r2 = HwsRule(m2, 0, [items[1]], 0, actions_data[1], tir_a, HwsRuleAttr(user_data=bytes(8)),
                     ctx=self.server.hws_ctx)

        # Matcher size 4
        m4_attr = HwsMatcherAttr(priority=0, rule_log=2, resizable=True,
                                 resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        m4 = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                        action_templates, len(action_templates), m4_attr)
        # Resize
        m2.resize_set_target(m4)
        queue_id=0
        r1.matcher_resize_rule_move(m2, m4, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        r2.matcher_resize_rule_move(m2, m4, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        m2.close()
        r3 = HwsRule(m4, 0, [items[2]], 0, actions_data[2], tir_a, HwsRuleAttr(user_data=bytes(8)),
                     ctx=self.server.hws_ctx)
        # Matcher size 8
        m8_attr = HwsMatcherAttr(priority=0, rule_log=3, resizable=True,
                                 resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        m8 = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                        action_templates, len(action_templates), m8_attr)
        # Resize
        m4.resize_set_target(m8)
        r1.matcher_resize_rule_move(m4, m8, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        r2.close()
        r3.matcher_resize_rule_move(m4, m8, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        m4.close()
        r4 = HwsRule(m8, 0, [items[3]], 0, actions_data[3], tir_a, HwsRuleAttr(user_data=bytes(8)),
                     ctx=self.server.hws_ctx)
        # Matcher size 16
        m16_attr = HwsMatcherAttr(priority=0, rule_log=4, resizable=True,
                                  resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        m16 = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                        action_templates, len(action_templates), m16_attr)
        # Resize
        m8.resize_set_target(m16)
        r1.matcher_resize_rule_move(m8, m16, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        r4.matcher_resize_rule_move(m8, m16, HwsRuleAttr(user_data=bytes(8)))
        self.server.hws_ctx.check_op_completion(queue_id)
        m8.close() # r3 should be closed.
        # Verify rules with traffic.
        for packet, exp_packet, tag in zip(packets, exp_packets, tags):
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, packets=packet,
                              expected_packet=exp_packet, tag_value=tag)

    def test_mlx5dv_hws_simple_matcher_resize_chain(self):
        """
        Create a small matcher(size 2) inserts 2 simple (1 STE) rules with action tag.
        Resize 2->4->8->16.
        On the way move rules, remove old rules, and add new rules.
        Remove unused matchers.
        Verify rules with traffic.
        """
        # AT
        tag_a = a.HwsActionTag(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        action_templates = [a.HwsActionTemplate([tag_a], 1)]
        actions_data = []
        tag_values = [0x1111, 0x2222, 0x3333, 0x4444]
        for tag in tag_values:
            actions_data.append([a.HwsActionData(value=tag)])

        ip_addresses = ['1.1.1.1', '2.2.2.2', '3.3.3.3', '4.4.4.4']
        packets = []
        for ip in ip_addresses:
            packets.append(u.gen_packet(self.server.msg_size, src_ipv4=ip))

        self.resize_matcher(action_templates, actions_data,
                            [[packets[0], packets[1]], [packets[2], packets[3]]],
                            [packets[0], packets[3]], [0x1111, 0x4444])

    def test_mlx5dv_hws_complex_matcher_resize_chain(self):
        """
        Create a small matcher(size 2) inserts 2 complex (2 STEs) rules with action tag, push vlan,
        counter and modify.
        Resize 2->4->8->16.
        On the way move rules, remove old rules, and add new rules.
        Remove unused matchers.
        Verify rules with traffic.
        """
        from tests.mlx5_prm_structs import SetActionIn
        tag_a = a.HwsActionTag(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_a = a.HwsActionPushVlan(self.server.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        counter_a, resource_counter = self.server.create_counter()
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=0)
        argument = HwsResource(self.server.hws_ctx, resource_attr)
        action1 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                              data=0x88888888, length=ModifyFieldsLen.MAC_47_16)
        action2 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_15_0,
                              data=0x8888, length=ModifyFieldsLen.MAC_15_0)
        mdf_actions = [action1, action2]
        mdf_a = a.HwsActionModify(self.server.hws_ctx, mdf_actions, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
                                  [argument], 1)

        actions = [tag_a, push_a, counter_a, mdf_a]
        action_templates = [a.HwsActionTemplate(actions, len(actions))]

        actions_data = []
        tag_values = [0x1111, 0x2222, 0x3333, 0x4444]
        for tag in tag_values:
            actions_data.append([a.HwsActionData(value=tag),
                                 a.HwsActionData(vlan_hdr=create_vlan_hdr()),
                                 a.HwsActionData(resource_offset=0),
                                 a.HwsActionData(modify_data=mdf_actions, resource_offset=0)])

        exp_src_mac = bytes.fromhex("88:88:88:88:88:88".replace(':', ''))
        packets = []
        exp_packets = []
        ip_addresses = ['1.1.1.1', '2.2.2.2', '3.3.3.3', '4.4.4.4']
        for ip in ip_addresses:
            packets.append(u.gen_packet(self.client.msg_size - u.PacketConsts.VLAN_HEADER_SIZE,
                                        src_ipv4=ip, with_vlan=False))
            exp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4=ip, src_mac=exp_src_mac,
                                            with_vlan=True))

        self.resize_matcher(action_templates, actions_data,
                            [[packets[0], packets[1]], [packets[2], packets[3]]],
                            [exp_packets[0], exp_packets[3]], [0x1111, 0x4444])
        HwsActionTest.verify_counter(self, self.server, resource_counter, self.server.num_msgs * 2)

    def create_5tuple_items_mts_diff_masks(self, sips, dips, sports, dports):
        """
        Create four 5tuple ipv4/udp items and matcher templates, each list must contain 4 values.
        Masks for each item:
        sip - '255.0.0.0'       dip - '255.0.0.0'       sport - 0xf000  dport - 0xf000
        sip - '255.255.0.0'     dip - '255.255.0.0'     sport - 0xff00  dport - 0xff00
        sip - '255.255.255.0'   dip - '255.255.255.0'   sport - 0xfff0  dport - 0xfff0
        sip - '255.255.255.255' dip - '255.255.255.255' sport - 0xffff  dport - 0xffff
        :param sips: List of source ips
        :param dips: List of destination ips
        :param sports: List of source ports
        :param dports: List of destination ports
        """
        items = []
        matcher_templates = []
        ip_masks = ['255.0.0.0', '255.255.0.0', '255.255.255.0', '255.255.255.255']
        port_masks = [0xf000, 0xff00, 0xfff0, 0xffff]
        for i in range(4):
            fields_list = [('ipv4_src', e.MLX5DV_HWS_HEADER_OUTER),
                           ('ipv4_dst', e.MLX5DV_HWS_HEADER_OUTER),
                           ('ip_udp_sport', e.MLX5DV_HWS_HEADER_OUTER),
                           ('ip_udp_dport', e.MLX5DV_HWS_HEADER_OUTER)]
            data_list = [(sips[i], ip_masks[i]),
                         (dips[i], ip_masks[i]),
                         (sports[i], port_masks[i]),
                         (dports[i], port_masks[i])]
            items.append(prepare_item(fields_list, data_list))
            matcher_templates.append(HwsMatchTemplate(self.server.hws_ctx, [items[i]], 1))
        return items, matcher_templates

    def test_mlx5dv_hws_ext_matcher_5tuple(self):
        """
        Extended matcher - Match on 5 tuple ipv4/udp header attributes:
        Create empty root matcher and forward all traffic to non root table
        with a matcher that matches on 4 different matcher templates:
        sip - '255.0.0.0'       dip - '255.0.0.0'       sport - 0xf000  dport - 0xf000
        sip - '255.255.0.0'     dip - '255.255.0.0'     sport - 0xff00  dport - 0xff00
        sip - '255.255.255.0'   dip - '255.255.255.0'   sport - 0xfff0  dport - 0xfff0
        sip - '255.255.255.255' dip - '255.255.255.255' sport - 0xffff  dport - 0xffff
        Create 4 rules with tir action on each MT, verify rules with the correct packets.
        Note - in the extended matcher the hashing is over the smallest shared prefix,
               and the matching is done on the exact match defined in the item.
        Hash fields:
                src_ip_o: 0xff000000
                src_port_o: 0xf000
                dst_port_o: 0xf000
                dst_ip_o: 0xff000000
        """
        # Root
        self.server.root_table = self.server.create_table(level=0,
                                                          table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Create root rule.
        self.server.root_matcher, self.server.root_dest_tbl_rule, self.server.table = \
            self.server.create_root_fwd_rule([], table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Non-root
        # the hashing depends on the most significant byte of ipv4 and most significant
        # 4 bits of the port, and the matching grows in accordance with the masks' values as
        # described above.
        # So we use ip values that differ in the first byte, and port values that differ in
        # the first 4 bits, then fill the other values according to the mask length.
        src_ips = ['1.0.0.0', '2.1.0.0', '3.1.1.0', '4.1.1.1']
        dst_ips = ['5.0.0.0', '6.5.0.0', '7.5.5.0', '8.5.5.5']
        sports = [0x8000, 0x7800, 0x6880, 0x5888]
        dports = [0x1000, 0x2100, 0x3110, 0x4111]
        ext_matcher_items, ext_matcher_templates = \
                            self.create_5tuple_items_mts_diff_masks(src_ips, dst_ips, sports, dports)
        action_templates = [a.HwsActionTemplate([], 0)]

        m_attr = HwsMatcherAttr(priority=0)
        ext_matcher = HwsMatcher(self.server.table, ext_matcher_templates,
                                 len(ext_matcher_templates), action_templates,
                                 len(action_templates), m_attr)

        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rules = []
        exp_packets = []
        unexp_packets = []
        unexp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4='2.0.0.0',
                                          dst_ipv4='5.0.0.0', src_port=0x8000, dst_port=0x1000))
        unexp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4='2.1.0.0',
                                          dst_ipv4='6.6.0.0', src_port=0x7800, dst_port=0x2100))
        unexp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4='3.1.1.0',
                                          dst_ipv4='7.5.5.0', src_port=0x6870, dst_port=0x3110))
        unexp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4='4.1.1.1',
                                          dst_ipv4='8.5.5.5', src_port=0x5888, dst_port=0x4112))

        for i in range(len(ext_matcher_items)):
            tir_rules.append(HwsRule(ext_matcher, i, [ext_matcher_items[i]], 0, [], tir_a,
                                     HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx))
            exp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4=src_ips[i],
                                            dst_ipv4=dst_ips[i], src_port=sports[i],
                                            dst_port=dports[i]))

        for i in range(len(ext_matcher_items)):
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs,
                              [exp_packets[i]] + unexp_packets)

    def create_diff_items(self, mac_vals, ip_vals, port_vals):
        """
        Create six diff mac/ipv4/udp items with full masks.
        Values for each item:
        dmac - '01:11:11:11:11:11'
        dmac - '01:22:22:22:22:22' smac - '01:11:11:11:11:11'
        dmac - '01:33:33:33:33:33' smac - '01:22:22:22:22:22' sip - '1.1.1.1'
        dmac - '01:44:44:44:44:44' smac - '01:33:33:33:33:33' sip - '2.2.2.2' dip - '1.1.1.1'
        dmac - '01:55:55:55:55:55' smac - '01:44:44:44:44:44' sip - '3.3.3.3' dip - '2.2.2.2'
                sport - 0x1111
        dmac - '01:66:66:66:66:66' smac - '01:55:55:55:55:55' sip - '4.4.4.4' dip - '3.3.3.3'
                sport - 0x2222 dport - 0x1111
        :param mac_vals: List of all mac values.
        :param ip_vals: List of all ip values.
        :param port_vals: List of all port values.
        """
        items = []
        header_outer = e.MLX5DV_HWS_HEADER_OUTER
        full_mac_mask = u.PacketConsts.MAC_MASK
        full_ip_mask = '255.255.255.255'
        full_port_mask = 0xffff
        item_configs = [
            {'fields': ['smac'], 'data': [mac_vals[0]]},
            {'fields': ['smac', 'dmac'], 'data': [mac_vals[1], mac_vals[0]]},
            {'fields': ['smac', 'dmac', 'ipv4_src'],
             'data': [mac_vals[2], mac_vals[1], ip_vals[0]]},
            {'fields': ['smac', 'dmac', 'ipv4_src', 'ipv4_dst'],
             'data': [mac_vals[3], mac_vals[2], ip_vals[1], ip_vals[0]]},
            {'fields': ['smac', 'dmac', 'ipv4_src', 'ipv4_dst', 'ip_udp_sport'],
             'data': [mac_vals[4], mac_vals[3], ip_vals[2], ip_vals[1], port_vals[0]]},
            {'fields': ['smac', 'dmac', 'ipv4_src', 'ipv4_dst', 'ip_udp_sport', 'ip_udp_dport'],
             'data': [mac_vals[5], mac_vals[4], ip_vals[3], ip_vals[2], port_vals[1], port_vals[0]]}
        ]
        for config in item_configs:
            fields_list = [(field, header_outer) for field in config['fields']]
            data_list = [(value, full_mac_mask if field in ['smac', 'dmac'] else \
                          full_ip_mask if field in ['ipv4_src', 'ipv4_dst'] else \
                          full_port_mask) for field, value in zip(config['fields'], config['data'])]
            items.append(prepare_item(fields_list, data_list))
        return items

    def create_diff_packets(self, mac_vals, ip_vals, port_vals):
        """
        Create six expected mac/ipv4/udp diff packets corresponding with the diff items.
        dmac - '01:11:11:11:11:11'
        dmac - '01:22:22:22:22:22' smac - '01:11:11:11:11:11'
        dmac - '01:33:33:33:33:33' smac - '01:22:22:22:22:22' sip - '1.1.1.1'
        dmac - '01:44:44:44:44:44' smac - '01:33:33:33:33:33' sip - '2.2.2.2' dip - '1.1.1.1'
        dmac - '01:55:55:55:55:55' smac - '01:44:44:44:44:44' sip - '3.3.3.3' dip - '2.2.2.2'
                sport - 0x1111
        dmac - '01:66:66:66:66:66' smac - '01:55:55:55:55:55' sip - '4.4.4.4' dip - '3.3.3.3'
                sport - 0x2222 dport - 0x1111
        :param mac_vals: List of all mac values.
        :param ip_vals: List of all ip values.
        :param port_vals: List of all port values.
        """
        exp_packets = []

        for i in range(6):
            packet_params = {'src_mac': bytes.fromhex(mac_vals[i].replace(':', ''))}
            if i > 0:
                packet_params['dst_mac'] = bytes.fromhex(mac_vals[i-1].replace(':', ''))
            if i >= 2:
                packet_params['src_ipv4'] = ip_vals[i-2]
            if i >= 3:
                packet_params['dst_ipv4'] = ip_vals[i-3]
            if i >= 4:
                packet_params['src_port'] = port_vals[i-4]
            if i == 5:
                packet_params['dst_port'] = port_vals[i-5]
            exp_packets.append(u.gen_packet(self.server.msg_size, **packet_params))
        return exp_packets

    def test_mlx5dv_hws_ext_matcher_diff(self):
        """
        Extended matcher - Match on smac/dmac/ipv4/udp.
        Create empty root matcher and forward all traffic to non root table
        with a matcher that matches on 6 different match templates:
        1. match on dmac
        2. match on dmac/smac
        3. match on dmac/smac/sip
        4. match on dmac/smac/sip/dip
        5. match on dmac/smac/sip/dip/sport
        6. match on dmac/smac/sip/dip/sport/dport
        Create 6 rules with tir action on each MT, verify rules with the correct packets.
        """
        # Root
        self.server.root_table = self.server.create_table(level=0,
                                                          table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Create root rule
        self.server.root_matcher, self.server.root_dest_tbl_rule, self.server.table = \
            self.server.create_root_fwd_rule([], table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        # Non-root
        # DST mac must be multicast
        mac_vals = ['01:11:11:11:11:11', '01:22:22:22:22:22', '01:33:33:33:33:33',
                   '01:44:44:44:44:44', '01:55:55:55:55:55', '01:66:66:66:66:66']
        ip_vals = ['1.1.1.1', '2.2.2.2', '3.3.3.3', '4.4.4.4']
        port_vals = [0x1111, 0x2222]
        matcher_templates = []
        ext_matcher_items = self.create_diff_items(mac_vals, ip_vals, port_vals)
        for item in ext_matcher_items:
            matcher_templates.append(HwsMatchTemplate(self.server.hws_ctx, [item], 1))
        action_templates = [a.HwsActionTemplate([], 0)]
        m_attr = HwsMatcherAttr(priority=0)
        ext_matcher = HwsMatcher(self.server.table, matcher_templates, len(matcher_templates),
                                 action_templates, len(action_templates), m_attr)
        # Create 6 rules with tir action
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rules = []
        exp_packets = self.create_diff_packets(mac_vals, ip_vals, port_vals)
        UNEXP_SRC_MAC = '01:77:77:77:77:77'
        unexp_packet = u.gen_packet(self.server.msg_size,
                                    src_mac=bytes.fromhex(UNEXP_SRC_MAC.replace(':', '')))
        for i in range(len(ext_matcher_items)):
            tir_rules.append(HwsRule(ext_matcher, i, [ext_matcher_items[i]], 0, [], tir_a,
                                     HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx))
        for i in range(len(ext_matcher_items)):
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs,
                              [exp_packets[i], unexp_packet])

    def test_mlx5dv_hws_isolated_matcher(self):
        """
        Verify isolated matcher creation and destination matcher action.
        Create standard root and non-root tables with a rule forwarding traffic
        from the root table to the non-root table.
        The non-root table has 3 matchers: matcher, isolated_matcher1, and isolated_matcher2.
        - matcher matches on IPv4 src ip 1.1.1.1, and his action is to jump to isolated_matcher1 to
          index DEST_MATCHER_IDX1.
        - Add rule to isolated_matcher1 in index DEST_MATCHER_IDX1 that matches on IPv4 src ip
          1.1.1.1, and his action is to jump to isolated_matcher2 to index DEST_MATCHER_IDX2.
        - Add rule to isolated_matcher2 in index DEST_MATCHER_IDX2 that matches on source port
          PacketConsts.SRC_PORT, and his action is to go to the tir.
        To check the structure, we use two packets:
        exp_packet with src ip 1.1.1.1 and PacketConsts.SRC_PORT source port, and un_exp_packet,
        with src ip 1.1.1.1 and UN_EXPECTED_SRC_PORT.
        We then send the traffic and verify that exp_packet went through and un_exp_packet did not.
        Test structure:
        [root_matcher] -> [matcher(src_ip=1.1.1.1) --> isolated_matcher1(src_ip=1.1.1.1) -->
                           isolated_matcher2(src_port=PacketConsts.SRC_PORT) --> tir]
        """
        items_sip = [create_sipv4_item()]
        fields_list = [('ip_sport', e.MLX5DV_HWS_HEADER_OUTER)]
        data_list = [(u.PacketConsts.SRC_PORT, 0xffff)]
        items_sport = [prepare_item(fields_list, data_list)]

        self.server.init_steering_resources(items=items_sip)

        m_attr = HwsMatcherAttr(priority=1, isolated=True, col_log=0,
                                insert_mode=e.MLX5DV_HWS_MATCHER_INSERT_BY_INDEX)
        isolated_matcher1 = HwsMatcher(self.server.table, self.server.matcher_templates,
                                       len(self.server.matcher_templates),
                                       self.server.action_templates,
                                       len(self.server.action_templates), m_attr)
        isolated2_mt = [HwsMatchTemplate(self.server.hws_ctx, items_sport, len(items_sport))]
        isolated_matcher2 = HwsMatcher(self.server.table, isolated2_mt, len(isolated2_mt),
                                       self.server.action_templates,
                                       len(self.server.action_templates), m_attr)
        self.rules = []
        # self.matcher --> self.isolated_matcher1
        dest_matcher_a1 = a.HwsActionDestMatcher(self.server.hws_ctx, isolated_matcher1,
                                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        DEST_MATCHER_IDX1 = 0
        self.rules.append(HwsRule(self.server.matcher, 0, items_sip, 0,
                                  [a.HwsActionData(index=DEST_MATCHER_IDX1)], dest_matcher_a1,
                                  HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx))
        # self.isolated_matcher1 --> self.isolated_matcher2
        dest_matcher_a2 = a.HwsActionDestMatcher(self.server.hws_ctx, isolated_matcher2,
                                                 e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        DEST_MATCHER_IDX2 = 1
        self.rules.append(HwsRule(isolated_matcher1, 0, items_sip, 0,
                                  [a.HwsActionData(index=DEST_MATCHER_IDX2)], dest_matcher_a2,
                                  HwsRuleAttr(user_data=bytes(8), rule_idx=DEST_MATCHER_IDX1),
                                  ctx=self.server.hws_ctx))
        # self.isolated_matcher2 -> tir
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        self.rules.append(HwsRule(isolated_matcher2, 0, items_sport, 0, [], tir_a,
                                  HwsRuleAttr(user_data=bytes(8), rule_idx=DEST_MATCHER_IDX2),
                                  ctx=self.server.hws_ctx))

        exp_packet = u.gen_packet(self.server.msg_size)
        UN_EXPECTED_SRC_PORT = 1235
        un_exp_packet = u.gen_packet(self.server.msg_size, src_port=UN_EXPECTED_SRC_PORT)
        packets = [exp_packet, un_exp_packet]
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, packets)


class HwsResourceManagementTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws resource management functionality class.
    """
    resources_list=[e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ASO_METER,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B,
                    e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B]

    def hws_resource_alloc_bulk(self, resource_type, log_size_list):
        for log_size in log_size_list:
            resource_attr = HwsResourceAttr(resource_type, log_size)
            resource = HwsResource(self.server.hws_ctx, resource_attr)
            # Need to free memory by closing the resource before next iteration.
            resource.close()

    def test_mlx5dv_hws_resource_alloc_bulks(self):
        """
        Allocate all types of resource bulks of different sizes.
        """
        hca_caps2 = u.get_hca_caps_layout_2(self)
        # COUNTER
        size_in_range = (MAX_COUNTER_BULK_LOG_SIZE - MIN_COUNTER_BULK_LOG_SIZE) / 2
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                                     [MIN_COUNTER_BULK_LOG_SIZE, size_in_range,
                                      MAX_COUNTER_BULK_LOG_SIZE])
        # FLOW HIT
        hca_caps = u.query_hca_caps(Mlx5Context(Mlx5DVContextAttr(), self.dev_name))
        max_log_size = hca_caps.log_flow_hit_aso_max_alloc
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_FLOW_HIT, [max_log_size])
        # METER
        qos_caps = u.get_qos_caps(Mlx5Context(Mlx5DVContextAttr(), self.dev_name))
        max_log_size = qos_caps.log_meter_aso_max_alloc
        min_log_size = 0
        size_list = [min_log_size, (min_log_size + max_log_size) / 2, max_log_size]
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_METER, size_list)
        # CT
        max_log_size = hca_caps2.log_conn_track_max_alloc
        min_log_size = 0
        size_list = [min_log_size, (min_log_size + max_log_size) / 2, max_log_size]
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK, size_list)
        # ARGS
        max_log_size = hca_caps.log_header_modify_argument_max_alloc
        min_log_size = 0
        size_list = [min_log_size, (min_log_size + max_log_size) / 2, max_log_size]
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B, size_list)

        max_log_size = max_log_size - 1
        size_list = [min_log_size, (min_log_size + max_log_size) / 2, max_log_size]
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_128B, size_list)

        max_log_size = max_log_size - 1
        size_list = [min_log_size, (min_log_size + max_log_size) / 2, max_log_size]
        self.hws_resource_alloc_bulk(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_256B, size_list)

    def create_resource_get_all_offsets(self, resource_type, log_size, qid):
        size =  2 ** log_size
        resource_attr = HwsResourceAttr(resource_type, log_size)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        offsets = []
        for i in range(size):
            offsets += [resource.get_offset(qid)]
        return resource, offsets

    def hws_resource_alloc_get_put(self, resource_type, log_size):
        qid = 0
        resource, offsets = self.create_resource_get_all_offsets(resource_type, log_size, qid)
        resource.put_offset(offsets[0], qid)
        resource.put_offset(offsets[1], qid)
        offsets += [resource.get_offset(qid)]
        offsets += [resource.get_offset(qid)]
        self.assertTrue(set(offsets) == set(range(2 ** log_size)))

    def test_mlx5dv_hws_resource_alloc_get_put(self):
        """
        Allocate aso counter resource bulk of log size 2 (min bulk size),
        Verify get and put offset functions.
        """
        self.hws_resource_alloc_get_put(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                                        MIN_COUNTER_BULK_LOG_SIZE)

    def hws_resource_alloc_get_bad_flow(self, resource_type, log_size):
        qid = 0
        resource, _ = self.create_resource_get_all_offsets(resource_type, log_size, qid)
        with self.assertRaises(PyverbsError) as ex:
            offset = resource.get_offset(qid)

    def test_mlx5dv_hws_resource_alloc_get_bad_flow(self):
        """
        Bad flow - allocate two aso counter resource bulk of log size 2 (min bulk size) and
        23 (max bulk size).
        Try to get the next offset and expect failure.
        """
        self.hws_resource_alloc_get_bad_flow(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                                             MIN_COUNTER_BULK_LOG_SIZE)
        self.hws_resource_alloc_get_bad_flow(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER,
                                             MAX_COUNTER_BULK_LOG_SIZE)

    def test_mlx5dv_hws_resource_get_put_multi_qid(self):
        """
        Allocate aso counter resource bulk of log size 3.
        Verify get and put offset functions using multiple qids.
        - Verify get in qid0 and put on qid1.
        - Verify get offset0 and offset1 using qid0 and qid1,
             and put offset0 and offset1 using qid0 and qid1.
        We can't use builk log size 2 for this test because it results in a single
        internal chunk which means two queues can't allocate at the same time.
        """
        log_size = 3
        qid0 = NUM_OF_QUEUES
        qid1 = 1
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER, log_size)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        offset0 = resource.get_offset(qid0)
        resource.put_offset(offset0, qid1)

        offset0 = resource.get_offset(qid0)
        offset1 = resource.get_offset(qid1)
        resource.put_offset(offset0, qid0)
        resource.put_offset(offset1, qid1)

        offset0 = resource.get_offset(qid0)
        offset1 = resource.get_offset(qid1)
        offset2 = resource.get_offset(qid0)
        offset3 = resource.get_offset(qid1)
        offset4 = resource.get_offset(qid0)
        offset5 = resource.get_offset(qid1)
        offset6 = resource.get_offset(qid0)
        offset7 = resource.get_offset(qid1)

        offsets = [offset0, offset1, offset2, offset3, offset4, offset5, offset6, offset7]
        self.assertTrue(set(offsets) == set(range(2 ** log_size)))

    def uniform_get_put(self, resource_log_sz, block_log_sz, qid):
        """
        Allocate all available offsets in chunks of the given block_log_sz, on a single qid.
        """
        num_offsets = 2 ** (resource_log_sz - block_log_sz)
        offsets = []
        for i in range(num_offsets):
            offsets.append(self.resource.get_offset_block(qid, block_log_sz))

        for offset in offsets:
            self.resource.put_offset_block(offset, qid, block_log_sz)

    def test_mlx5dv_hws_resource_block_uniform_get_put(self):
        """
        Allocate argument-64B resource bulk of log size 20,
        Verify get and put blocks of different log sizes from 0 to max block size available.
        """
        qid = 1
        RESOURCE_LOG_SZ = 20
        resource_type = e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B
        resource_attr = HwsResourceAttr(resource_type, RESOURCE_LOG_SZ, block_alloc_support=1)
        self.resource = HwsResource(self.server.hws_ctx, resource_attr)
        for block_log_sz in range(self.resource.max_block_size):
            self.uniform_get_put(RESOURCE_LOG_SZ, block_log_sz, qid)

    def hws_resource_devx_basic_bad_flow(self, resource_type, log_size=0):
        resource_attr = HwsResourceAttr(resource_type, log_size)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        with self.assertRaises(PyverbsError) as ex:
            resource.devx_obj_id

    def test_mlx5dv_hws_resource_devx_basic_bad_flow(self):
        """
        Bad flow - allocate all resource types.
        Call resource devx_id without calling get_devx_obj function first.
        """
        for res in HwsResourceManagementTest.resources_list:
            min_bulk_log_size = 0
            if res == e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
                min_bulk_log_size = MIN_COUNTER_BULK_LOG_SIZE
            self.hws_resource_devx_basic_bad_flow(res, min_bulk_log_size)

    def hws_resource_devx_basic(self, resource_type, log_size=0):
        resource_attr = HwsResourceAttr(resource_type, log_size)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        resource.get_devx_obj()
        if resource.devx_obj_id < 0:
            raise PyverbsError('Failed to get devx obj id.')

    def test_mlx5dv_hws_resource_devx_basic(self):
        """
        Allocate all resource types and get devx object and id.
        Verify that resource devx is not null and has non-negative devx obj id.
        """
        for res in HwsResourceManagementTest.resources_list:
            min_bulk_log_size = 0
            if res == e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER:
                min_bulk_log_size = MIN_COUNTER_BULK_LOG_SIZE
            self.hws_resource_devx_basic(res, min_bulk_log_size)

    def check_buffer(self, n=1, offset=0, exp_value=0):
        """
        Check if the first n bytes starting from offset in the MR buffer are equal to the expected
        value.
        """
        buf = self.server.mr.read(self.server.msg_size, 0)
        for i in range(offset, offset + n):
            if (buf[i] != exp_value):
                raise PyverbsError(f'Incorrect MR buffer values in index {i} got {buf[i]} \
                                   instead of expected {exp_value}.')

    def init_buffer(self, value=0):
        """
        Initiate and check that all bytes in the MR buffer are equal the value.
        :param value: A hexadecimal value representing the initial MR bytes value.
        """
        self.server.mr.write(bytes([value]) * self.server.msg_size, self.server.msg_size, 0)
        self.check_buffer(self.server.msg_size, offset=0, exp_value=value)

    def test_mlx5dv_hws_resource_query_config_using_aso_wqe(self):
        """
        Resource ASO WQE enqueue test.
        Init buffer to zero.
        Create two ASO WQEs with always-hit conditions and a non-zero data,
        Enqueue the 1st WQE to perform a query config operation, read the buffer(should be zero).
        Enqueue the 2nd WQE to perform a query config operation, read the buffer(should be 64 bytes
        of 0xff).
        """
        resource_type = e.MLX5DV_HWS_RESOURCE_TYPE_ASO_CONNTRACK
        ct_bulk_log_size = 2
        resource_attr = HwsResourceAttr(resource_type, ct_bulk_log_size)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        qid = 0
        offset = resource.get_offset(qid)
        self.init_buffer()
        aso_attr = HwsResourceEnqueueAsoAttr(burst=False, user_data=bytes(8), offset=offset)
        # Enqueue ASO WQE
        ct_aso_wqe, wqe_len = create_generic_aso_wqe(self.server, b"\xff" * 64)
        resource.enqueue_aso_wqe(qid, ct_aso_wqe, wqe_len, aso_attr)
        self.check_buffer(n=64, offset=0, exp_value=0)
        # Enqueue 2nd ASO WQE
        ct_aso_wqe2, wqe_len2 = create_generic_aso_wqe(self.server, b"\x11" * 64)
        resource.enqueue_aso_wqe(qid, ct_aso_wqe2, wqe_len2, aso_attr)
        self.check_buffer(n=64, offset=0, exp_value=0xff)

    def enqueue_aso_wqe_without_polling(self, test_type):
        """
        Resource ASO WQE enqueue with burst flag on test.
        Allocate two counter resources, enqueue with burst flag on two ASO WQEs to config both
        resources using the same qid.
        Drain the queue or enqueue with burst flag off a read-only ASO WQE.
        Read old resource context.
        Enqueue with burst flag off two more read-only ASO WQEs to read both resources'
        new contexts.
        Validate that both resources changed.
        :param test_type: Can be either "burst" or "drain".
        """
        resource_type = e.MLX5DV_HWS_RESOURCE_TYPE_ASO_COUNTER
        resource_attr = HwsResourceAttr(resource_type, 2)
        counter0 = HwsResource(self.server.hws_ctx, resource_attr)
        counter1 = HwsResource(self.server.hws_ctx, resource_attr)
        offset = 0
        qid = 0
        self.init_buffer(value=0xff)
        # Enqueue ASO WQE to counter0.
        aso_attr = HwsResourceEnqueueAsoAttr(burst=True, user_data=bytes(8), offset=offset)
        aso_wqe, wqe_len = create_generic_aso_wqe(self.server, b"\x11" * 64)
        counter0.enqueue_aso_wqe(qid, aso_wqe, wqe_len, aso_attr)
        # Enqueue ASO WQE to counter1.
        aso_attr = HwsResourceEnqueueAsoAttr(burst=True, user_data=bytes(8), offset=offset)
        aso_wqe, wqe_len = create_generic_aso_wqe(self.server, b"\x22" * 64)
        counter1.enqueue_aso_wqe(qid, aso_wqe, wqe_len, aso_attr)
        # Check nothing happened.
        self.check_buffer(n=64, offset=0, exp_value=0xff)
        if test_type == "burst":
            # Enqueue read only ASO WQE with burst=False.
            aso_attr = HwsResourceEnqueueAsoAttr(user_data=bytes(8), offset=offset)
            aso_wqe, wqe_len = create_generic_aso_wqe(self.server)
            counter0.enqueue_aso_wqe(qid, aso_wqe, wqe_len, aso_attr)
        else:
            self.server.hws_ctx.queue_execute_op(qid, e.MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC)

        self.check_buffer(n=64, offset=0, exp_value=0)
        # Enqueue read only ASO WQEs to counter0.
        aso_attr = HwsResourceEnqueueAsoAttr(user_data=bytes(8), offset=offset)
        aso_wqe, wqe_len = create_generic_aso_wqe(self.server)
        counter0.enqueue_aso_wqe(qid, aso_wqe, wqe_len, aso_attr)
        self.check_buffer(n=64, offset=0, exp_value=0x11)
        # Enqueue read only ASO WQEs to counter1.
        aso_attr = HwsResourceEnqueueAsoAttr(user_data=bytes(8), offset=offset)
        aso_wqe, wqe_len = create_generic_aso_wqe(self.server)
        counter1.enqueue_aso_wqe(qid, aso_wqe, wqe_len, aso_attr)
        self.check_buffer(n=64, offset=0, exp_value=0x22)

    def test_mlx5dv_hws_resource_burst(self):
        """
        Resource ASO WQE enqueue with burst flag on test.
        Allocate two counter resources, enqueue with burst flag on two ASO WQEs to config both
        resources using the same qid.
        Enqueue with burst flag off three more read-only ASO WQEs to read both resources'
        old and new contexts.
        Validate that both resources changed.
        """
        self.enqueue_aso_wqe_without_polling(test_type="burst")

    def test_mlx5dv_hws_resource_drain(self):
        """
        Resource ASO WQE enqueue with burst flag on test.
        Allocate two counter resources, enqueue with burst flag on two ASO WQEs to config both
        resources using the same qid.
        Drain the queue.
        Read old resource context.
        Enqueue with burst flag off two more read-only ASO WQEs to read both resources'
        new contexts.
        Validate that both resources changed.
        """
        self.enqueue_aso_wqe_without_polling(test_type="drain")


class HwsSetDefaultMissTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws default miss table functionality class.
    """

    @staticmethod
    def disconnect_miss_table(miss_table_array):
        for tbl in miss_table_array:
            tbl.set_default_miss()

    def send_and_verify(self, agr_ob, resource_counter, expected_num_packets, expectd_bytes=None):
        u.sampler_traffic(self.client, self.server, self.server.num_msgs)
        rec_packets, _ = agr_ob.query_counter(resource_counter, resource_counter.devx_obj_id)
        self.assertEqual(rec_packets, expected_num_packets,
                         f'Incorrect counter received packet {expected_num_packets} != \
                         {rec_packets}')

    def setup_root(agr_obj, root_items, table_type):
        """
        Creates a root table, and a matcher with root_items on the agr_obj.
        :param agr_obj: Mlx5DvHwsResources object.
        :param root_items: List of HwsItem.
        :param table_type: mlx5dv_hws_table_type enum.
        """
        agr_obj.root_table = agr_obj.create_table(level=0, table_type=table_type)
        root_matcher_template = [HwsMatchTemplate(agr_obj.hws_ctx, root_items, len(root_items))]
        action_temp = [a.HwsActionTemplate([], 0)]
        matcher_attr = HwsMatcherAttr(priority=1,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE)
        agr_obj.root_matcher = HwsMatcher(agr_obj.root_table, root_matcher_template,
                                          len(root_matcher_template), action_temp, len(action_temp),
                                          matcher_attr)

    def set_def_miss(self, agr_obj, table_type):
        """
        Setup:
        root -> tbl1(no matcher) *-> tbl2 *-> tbl3 (counter)
        *-> means "set default miss table"
        Send 10 packets and check counter = 10
        add matcher prio 3 to table2 - send 10 packets and check counter = 20
        add matcher prio 2 to table2 - send 10 packets and check counter = 30
        add matcher prio 4 to table2 - send 10 packets and check counter = 40
        del matcher prio 2 from table2 - send 10 packets and check counter = 50
        del matcher prio 4 from table2 - send 10 packets and check counter = 60
        """
        fields_list = [('eth_type', e.MLX5DV_HWS_HEADER_OUTER)]
        data_list = [(u.PacketConsts.ETHER_TYPE_IPV4, 0xffff)]
        items = [prepare_item(fields_list, data_list)]
        agr_obj.tbl1 = agr_obj.create_table(table_type=table_type)
        action_flag = agr_obj.table_type_to_action_type(table_type)
        root_action_flag = agr_obj.table_type_to_action_type(table_type, root=True)
        fwd_tbl_a = a.HwsActionDestTable(agr_obj.hws_ctx, agr_obj.tbl1, root_action_flag)
        HwsSetDefaultMissTest.setup_root(agr_obj, items, table_type)
        self.root_rule = HwsRule(agr_obj.root_matcher, 0, items, 0, [], fwd_tbl_a,
                                 HwsRuleAttr(user_data=bytes(8)), ctx=agr_obj.hws_ctx)
        agr_obj.tbl2 = agr_obj.create_table(level=2, table_type=table_type)
        agr_obj.tbl1.set_default_miss(agr_obj.tbl2)
        agr_obj.tbl3 = agr_obj.create_table(level=3, table_type=table_type)
        agr_obj.tbl2.set_default_miss(agr_obj.tbl3)
        # Create counter on tbl3
        counter_a, resource_counter = agr_obj.create_counter(action_flag)
        counter_at = [a.HwsActionTemplate([counter_a], 1)]
        self.matcher_tmpl = [HwsMatchTemplate(agr_obj.hws_ctx, items, len(items))]
        tbl3_matcher = agr_obj.create_matcher(agr_obj.tbl3, self.matcher_tmpl, counter_at, prio=0)
        dest_def_miss = a.HwsActionDestDefaultMiss(agr_obj.hws_ctx, action_flag)
        self.rule_counter = HwsRule(tbl3_matcher, 0, items, 0, [a.HwsActionData(resource_offset=0)],
                                    dest_def_miss, HwsRuleAttr(user_data=bytes(8)),
                                    ctx=agr_obj.hws_ctx)
        at = [a.HwsActionTemplate([], 0 )]
        try:
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs)
            agr_obj.matcher23 = agr_obj.create_matcher(agr_obj.tbl2, self.matcher_tmpl, at, prio=3)
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs * 2)
            agr_obj.matcher22 = agr_obj.create_matcher(agr_obj.tbl2, self.matcher_tmpl, at, prio=2)
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs * 3)
            agr_obj.matcher24 = agr_obj.create_matcher(agr_obj.tbl2, self.matcher_tmpl, at, prio=4)
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs * 4)
            agr_obj.matcher22.close()
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs * 5)
            agr_obj.matcher24.close()
            self.send_and_verify(agr_obj, resource_counter, self.server.num_msgs * 6)
        except Exception as ex:
            raise ex
        finally:
            self.disconnect_miss_table([agr_obj.tbl1, agr_obj.tbl2])

    def test_mlx5dv_hws_set_def_miss(self):
        """
        Verify set default miss table with different matchers' priority addition and removal.
        """
        self.set_def_miss(self.server, e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        self.set_def_miss(self.client, e.MLX5DV_HWS_TABLE_TYPE_NIC_TX)


class HwsFDBTest(Mlx5HwsTestCase):
    """
    Test FDB functionality class.
    """
    def setUp(self):
        self.vf_name = self.config['vf'][0] if self.config['vf'] else None
        if not self.vf_name:
            raise unittest.SkipTest('VF RDMA device must be provided')
        super().setUp()
        self.vf = Mlx5DvHwsResources(self.vf_name, ib_port=1, server=True,
                                     wc_flags=pe.IBV_WC_EX_WITH_FLOW_TAG)

    def tearDown(self):
        if self.vf:
            self.vf.hws_ctx.close()
            self.vf.ctx.close()
        super().tearDown()

    def action_vport(self, table_type):
        """
        Create rule with action goto Vport on FBD table on client.
        Create rule on VF with matching on src QP with action TIR.
        Sent traffic and verify on TIR on VF.
        """
        items = [create_sipv4_item()]
        at = [a.HwsActionTemplate([], 0)]
        self.client.init_steering_resources(items=items, at=at, table_type=table_type)
        port_num = self.client.vf_to_ib_port(self.vf_name)
        vport_a = \
            a.HwsActionDestVport(self.client.hws_ctx, port_num,
                                 self.client.table_type_to_action_type(table_type))
        vport_rule = HwsRule(self.client.matcher, 0, items, 0, [], vport_a,
                             HwsRuleAttr(user_data=bytes(8)), ctx=self.client.hws_ctx)
        # Tir rule on VF matching on src QP
        src_qp_fields_list = [('src_qp', e.MLX5DV_HWS_HEADER_NONE)]
        src_qp_data_list = [(self.client.qp.qp_num, 0xffffff)]
        items_src_qp = [prepare_item(src_qp_fields_list, src_qp_data_list)]
        self.vf.init_steering_resources(items=items_src_qp)
        tir_a = a.HwsActionDestTir(self.vf.hws_ctx, self.vf.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.vf.matcher, 0, items_src_qp, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.vf.hws_ctx)
        packet = u.gen_packet(self.client.msg_size)
        u.raw_traffic(self.client, self.vf, self.client.num_msgs, packet_to_send=packet)

    def test_mlx5dv_hws_action_vport(self):
        """
        Test action destination vport on FDB and FDB TX and test matching on
        source QP number.
        Note: Unable to test FDB RX since we can't create real FDB RX traffic
        on loop setup.
        """
        for table_type in [e.MLX5DV_HWS_TABLE_TYPE_FDB, e.MLX5DV_HWS_TABLE_TYPE_FDB_TX]:
            self.action_vport(table_type=table_type)

    def create_tmp_rule_and_verify(self, items, at_idx, action_data, dst_a, packets, exp_packet):
        """
        Creates a rule with provided items, action data and action template index with destination
        action vport.
        Sends and verifies packets.
        """
        tmp_rule = HwsRule(self.client.matcher, 0, items, at_idx, action_data, dst_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.client.hws_ctx)
        u.hws_raw_traffic(self.client, self.vf, self.client.num_msgs, packets,
                          expected_packet=exp_packet)
        tmp_rule.close()

    def hws_actions_fdb(self, table_type):
        """
        Verify actions on fdb by creating different rules on the same non root matcher:
        Drop
        Counter + Vport
        Modify header + Vport
        Dest table (+ on the next FT goto Vport)
        Verify traffic on Vport using action TIR.
        """
        from tests.mlx5_prm_structs import SetActionIn
        ib_port = self.client.vf_to_ib_port(self.vf_name)
        action_type = self.client.table_type_to_action_type(table_type)
        vport_a = a.HwsActionDestVport(self.client.hws_ctx, ib_port, action_type)

        sipv4_items = [create_sipv4_item()]
        src_qp_fields_list = [('src_qp', e.MLX5DV_HWS_HEADER_NONE)]
        src_qp_data_list = [(self.client.qp.qp_num, 0xffffff)]
        items_src_qp = [prepare_item(src_qp_fields_list, src_qp_data_list)]
        items = sipv4_items + items_src_qp
        dipv4_items = [create_dipv4_item()]

        counter_a, resource_counter = self.client.create_counter(action_type)

        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B, bulk_log_size=0)
        argument = HwsResource(self.client.hws_ctx, resource_attr)
        action1 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                              data=0x88888888, length=ModifyFieldsLen.MAC_47_16)
        action2 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_15_0,
                              data=0x8888, length=ModifyFieldsLen.MAC_15_0)
        mdf_actions = [action1, action2]
        mdf_a = a.HwsActionModify(self.client.hws_ctx, mdf_actions, action_type,
                                  [argument], 1)

        action_templates = [a.HwsActionTemplate([], 0),
                            a.HwsActionTemplate([counter_a], 1),
                            a.HwsActionTemplate([mdf_a], 1)]
        self.client.init_steering_resources(items=items, at=action_templates,
                                            table_type=table_type)

        # Create TIR rule on VF matching on dst IPv4
        self.vf.init_steering_resources(items=dipv4_items)
        tir_a = a.HwsActionDestTir(self.vf.hws_ctx, self.vf.tir, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.vf.matcher, 0, dipv4_items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.vf.hws_ctx)

        # Drop
        DROP_IP = '6.6.6.6'
        drop_items = [create_sipv4_item(DROP_IP)] + items_src_qp
        drop_a = a.HwsActionDestDrop(self.client.hws_ctx, action_type)
        drop_rule = HwsRule(self.client.matcher, 0, drop_items, 0, [], drop_a,
                            HwsRuleAttr(user_data=bytes(8)), self.client.hws_ctx)
        drop_packet = u.gen_packet(self.client.msg_size, src_ipv4=DROP_IP)

        # Counter
        # Verify that packet_drop is dropped and the other packet is passed through counter
        # and vport actions and reached TIR on VF.
        packet = u.gen_packet(self.client.msg_size)
        self.create_tmp_rule_and_verify(items, 1, [a.HwsActionData(resource_offset=0)], vport_a,
                                        [packet, drop_packet], packet)
        HwsActionTest.verify_counter(self, self.client, resource_counter, self.client.num_msgs)

        # Modify
        exp_src_mac = bytes.fromhex("88:88:88:88:88:88".replace(':', ''))
        exp_smac_packet = u.gen_packet(self.client.msg_size, src_mac=exp_src_mac)
        mdf_action_data = [a.HwsActionData(modify_data=mdf_actions, resource_offset=0)]
        self.create_tmp_rule_and_verify(items, 2, mdf_action_data, vport_a, [packet, drop_packet],
                                        exp_smac_packet)

        # Goto FT
        tbl2 = self.client.create_table(table_type=table_type, level=2)
        fwd_tbl_a = a.HwsActionDestTable(self.client.hws_ctx, tbl2, action_type)

        # Goto vport on table2
        mt = [HwsMatchTemplate(self.client.hws_ctx, items, len(items))]
        at = [a.HwsActionTemplate([], 0)]
        tbl2_matcher = HwsMatcher(tbl2, mt, len(mt), at, len(at), HwsMatcherAttr(priority=0))
        tbl2_rule = HwsRule(tbl2_matcher, 0, items, 0, [], vport_a,
                            HwsRuleAttr(user_data=bytes(8)), ctx=self.client.hws_ctx)
        self.create_tmp_rule_and_verify(items, 0, [], fwd_tbl_a, [packet, drop_packet], packet)

    def test_mlx5dv_hws_actions_fdb(self):
        """
        Test drop, counter, vport, modify header and dest table actions on FDB and FDB TX and test
        matching on sipv4 and source QP number.
        Note: Unable to test FDB RX since we can't create real FDB RX traffic
        on loop setup.
        """
        for table_type in [e.MLX5DV_HWS_TABLE_TYPE_FDB, e.MLX5DV_HWS_TABLE_TYPE_FDB_TX]:
            self.hws_actions_fdb(table_type=table_type)


class HwsRulesTest(Mlx5HwsTestCase):
    """
    Test various features of rule creation.
    """
    def create_burst_rules_without_polling(self, test_type):
        """
        Function creates a rule with burst flag on. Verifies with traffic that
        the rule wasn't written. Forces the rule to be written by creating
        another rule with bust flag off or by drain action flag using send
        queue action API. Verifies with traffic that the rule was written.
        :param test_type: Can be either "burst" or "drain".
        """
        items = [create_sipv4_item()]
        packet = u.gen_packet(self.server.msg_size)
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        rules = [HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                         HwsRuleAttr(user_data=bytes(8), burst=1))]
        with self.assertRaises(PyverbsError):
            u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet],
                              timeout=1)
        queue_id = 0
        if test_type == "burst":
            rules.append(HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                                 HwsRuleAttr(user_data=bytes(8), burst=0),
                                 ctx=self.server.hws_ctx))
        else:
            self.server.hws_ctx.queue_execute_op(queue_id, e.MLX5DV_HWS_QUEUE_OP_DRAIN_SYNC)
        self.server.hws_ctx.check_op_completion(queue_id)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet])

    def test_mlx5dv_hws_create_rules_burst(self):
        """
        Check that a rule with burst off writes previous rules with burst on.
        """
        self.create_burst_rules_without_polling(test_type="burst")

    def test_mlx5dv_hws_create_rules_drain(self):
        """
        Check that a drain op writes previous rules with burst on.
        """
        self.create_burst_rules_without_polling(test_type="drain")


class HwsRuleUpdate(Mlx5HwsTestCase):
    """
    Test rule update functionality class.
    """
    def update_rules_one_ste(self, agr_obj):
        """
        Update rules which fit in one STE.
        Initiate RX root and non-root tables on agr_obj (server or vf) create
        matcher with rule_idx_optimization hint(required for rule update) and
        then insert a rule with drop action, update this rule 3 times with
        actions counter, tir and push vlan + tir, run traffic to validate rule
        each time the rule is updated with counter and TIR.
        :param agr_obj: Aggregation object
        """
        items = [create_sipv4_item()]
        drop_a = a.HwsActionDestDrop(agr_obj.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        counter_a, resource_counter = agr_obj.create_counter()
        tir_a = a.HwsActionDestTir(agr_obj.hws_ctx, agr_obj.tir, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_a = a.HwsActionPushVlan(agr_obj.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_data = a.HwsActionData(vlan_hdr=create_vlan_hdr())
        at = [a.HwsActionTemplate([], 0 ), a.HwsActionTemplate([counter_a], 1),
              a.HwsActionTemplate([push_a], 1)]
        matcher_attr = HwsMatcherAttr(priority=0, rule_idx_opt=True)
        agr_obj.init_steering_resources(items=items, at=at, matcher_attr=matcher_attr)
        rule_attr = HwsRuleAttr(user_data=bytes(8))
        rule = HwsRule(agr_obj.matcher, 0, items, 0, [], drop_a, rule_attr, agr_obj.hws_ctx)
        packet = u.gen_packet(agr_obj.msg_size)
        u.hws_send_packets(self.client, [packet], iters=agr_obj.num_msgs)
        # Counter should be empty
        HwsActionTest.verify_counter(self, self.client, resource_counter, 0)
        rule.update(1, [a.HwsActionData(resource_offset=0)], drop_a, rule_attr, agr_obj.hws_ctx)
        u.hws_send_packets(self.client, [packet], iters=agr_obj.num_msgs)
        # All packets should go to counter
        HwsActionTest.verify_counter(self, self.client, resource_counter, agr_obj.num_msgs)
        rule.update(0, [], tir_a, rule_attr, agr_obj.hws_ctx)
        # All traffic should go to TIR and counter should stay the same
        u.hws_raw_traffic(self.client, agr_obj, agr_obj.num_msgs, [packet])
        HwsActionTest.verify_counter(self, self.client, resource_counter, agr_obj.num_msgs)
        rule.update(2, [push_data], tir_a, rule_attr, agr_obj.hws_ctx)
        packet = u.gen_packet(agr_obj.msg_size - u.PacketConsts.VLAN_HEADER_SIZE, with_vlan=False)
        expected_vlan_pkt = u.gen_packet(agr_obj.msg_size, with_vlan=True)
        # All traffic should be modifed and forwarded to TIR
        u.hws_raw_traffic(self.client, agr_obj, agr_obj.num_msgs, [packet], expected_vlan_pkt)
        HwsActionTest.verify_counter(self, self.client, resource_counter, agr_obj.num_msgs)

    def test_mlx5dv_hws_update_rules_one_ste(self):
        """
        Update rules which fit in one STE NIC RX.
        Call update_rules_one_ste method to test update API on PF.
        """
        self.update_rules_one_ste(self.server)

    def update_rules_two_stes(self, agr_obj):
        """
        Update rule which fits in two STEs.
        Initiate RX root and non-root tables on agr_obj (server or vf), create
        matcher with rule_idx_optimization hint and then insert a rule
        with modify + push vlan + tir, update this rule 8 times with different modify
        action values, run traffic to validate rule each time the rule is
        updated with TIR.
        :param agr_obj: Aggregation object
        """
        from tests.mlx5_prm_structs import SetActionIn
        items = [create_sipv4_item()]
        tir_a = a.HwsActionDestTir(agr_obj.hws_ctx, agr_obj.tir, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=1)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        action1 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                              data=0, length=ModifyFieldsLen.MAC_47_16)
        action2 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_15_0,
                              data=0, length=ModifyFieldsLen.MAC_15_0)
        actions = [action1, action2]
        modify_a = a.HwsActionModify(self.server.hws_ctx, actions, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
                                     [resource], 1)
        push_a = a.HwsActionPushVlan(agr_obj.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_data = a.HwsActionData(vlan_hdr=create_vlan_hdr())
        at = [a.HwsActionTemplate([modify_a, push_a], 2)]
        matcher_attr = HwsMatcherAttr(priority=0, rule_idx_opt=True)
        agr_obj.init_steering_resources(items=items, at=at, matcher_attr=matcher_attr)
        rule_attr = HwsRuleAttr(user_data=bytes(8))
        packet = u.gen_packet(agr_obj.msg_size - u.PacketConsts.VLAN_HEADER_SIZE, with_vlan=False)
        num_of_updates = 8
        for i in range(num_of_updates + 1):
            smac_47_16 = int('0x' + f'{str(i) * 8}', 16)
            smac_15_0 = int('0x' + f'{str(i) * 4}', 16)
            actions = [SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                                   data=smac_47_16, length=ModifyFieldsLen.MAC_47_16)]
            actions.append(SetActionIn(action_type=SET_ACTION,
                                       field=ModifyFields.OUT_SMAC_15_0, data=smac_15_0,
                                       length=ModifyFieldsLen.MAC_15_0))
            mdf_data = a.HwsActionData(modify_data=actions, resource_offset=0)
            if i == 0:
                rule = HwsRule(agr_obj.matcher, 0, items, 0, [mdf_data, push_data], tir_a,
                               rule_attr, agr_obj.hws_ctx)
            else:
                rule.update(0, [mdf_data, push_data], tir_a, rule_attr, agr_obj.hws_ctx)
            exp_src_mac = bytes.fromhex(f'{str(i) * 12}')
            exp_packet = u.gen_packet(self.server.msg_size, src_mac=exp_src_mac, with_vlan=True)
            u.raw_traffic(self.client, agr_obj, self.client.num_msgs, packet_to_send=packet,
                          expected_packet=exp_packet)

    def test_mlx5dv_hws_update_rules_two_stes(self):
        """
        Update rules which fit in two STEs NIC RX.
        Call update_rules_two_stes method to test update API on PF.
        """
        self.update_rules_two_stes(self.server)


class HwsFlexParserTest(Mlx5HwsTestCase):
    """
    Test mlx5dv_hws flex parser functionality class.
    """
    # The sampler currently supports only 32-bit granularity.
    SAMPLER_GRANULARITY = 32

    def setUp(self):
        """
        Verifies that the user has passed a PCI device name to work with.
        """
        self.pci_dev = self.config['pci_dev']
        if not self.pci_dev:
            raise unittest.SkipTest('PCI device must be passed by the user')
        super().setUp()

    def enum2flex_parser_name(self, parser_name):
        """
        Return the string representation of the given parser field enum value.
        """
        enum_to_str_map = {
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_0: 'flex_parser_0',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_1: 'flex_parser_1',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_2: 'flex_parser_2',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_3: 'flex_parser_3',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_4: 'flex_parser_4',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_5: 'flex_parser_5',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_6: 'flex_parser_6',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_7: 'flex_parser_7',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_0_OK: 'flex_parser_0_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_1_OK: 'flex_parser_1_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_2_OK: 'flex_parser_2_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_3_OK: 'flex_parser_3_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_4_OK: 'flex_parser_4_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_5_OK: 'flex_parser_5_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_6_OK: 'flex_parser_6_ok',
            e.MLX5DV_HWS_FNAME_FLEX_PARSER_7_OK: 'flex_parser_7_ok'
        }
        return enum_to_str_map.get(parser_name, "Unknown enum value")

    def create_sampler_item(self, parser_name, field_header, parser_value, parser_mask):
        """
        Given a parser name and header create an item with match value.
        """
        match = MatchParams()
        mask = MatchParams()
        field_name = self.enum2flex_parser_name(parser_name)
        setattr(match.misc_parameters, field_name, parser_value)
        setattr(mask.misc_parameters, field_name, parser_mask)
        data = HwsData(match, mask)
        smp_field = HwsField(parser_name, field_header,
                             e.MLX5DV_HWS_FIELD_MATCH_OP_EQ, e.MLX5DV_HWS_FIELD_MATCH_TYPE_TO_VALUE,
                             match.get_field_offset(parser_name, field_header))
        item = HwsItem([smp_field], 1, data)
        return item

    def match_with_traffic(self, items, packet, unexp_packets):
        """
        Given item list, create all necessary steering resources,
        Create TIR rule, send an expected and unexpected packets, verify using TIR.
        """
        self.server.init_steering_resources(items=items)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        tir_rule = HwsRule(self.server.matcher, 0, items, 0, [], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        u.hws_raw_traffic(self.client, self.server, self.server.num_msgs, [packet] + unexp_packets)

    def modify_match_with_traffic(self, item, modify_header_field, new_val, field_bit_len,
                                  packet_to_send, expected_packet):
        """
        Create an action template that includes a modify action to update the modify_header_field
        with the value new_val.
        Create all necessary steering resources using the given item and this action template.
        Create TIR rule with corresponding modify data, send an expected and unexpected packets,
        Verify using TIR.
        :param item: HwsItem for matching on the sampled field.
        :param modify_header_field: Modify header field id that refers to this samplers' value.
        :param new_val: New value to use in modify data.
        :param field_bit_len: Modify header field length in bits.
        :param packet_to_send: Packet to send.
        :param expected_packet: Expected packet to receive.
        """
        from tests.mlx5_prm_structs import SetActionIn
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=0)
        resource = HwsResource(self.server.hws_ctx, resource_attr)
        action = SetActionIn(action_type=SET_ACTION, field=modify_header_field,
                             data=new_val, length=field_bit_len)
        mdf_a = a.HwsActionModify(self.server.hws_ctx, [action], e.MLX5DV_HWS_ACTION_FLAG_NIC_RX,
                                  [resource], 1)
        at = [a.HwsActionTemplate([mdf_a], 1)]
        self.server.init_steering_resources(items=[item], at=at)
        tir_a = a.HwsActionDestTir(self.server.hws_ctx, self.server.tir,
                                   e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        mdf_data = a.HwsActionData(modify_data=[action], resource_offset=0)
        tir_rule = HwsRule(self.server.matcher, 0, [item], 0, [mdf_data], tir_a,
                           HwsRuleAttr(user_data=bytes(8)), ctx=self.server.hws_ctx)
        u.raw_traffic(self.client, self.server, self.server.num_msgs, packet_to_send=packet_to_send,
                      expected_packet=expected_packet)

    def insert_my_ip_node_to_graph(self):
        """
        Create a new node labeled my_ip at layer 3.
        Create an arc to connect the MAC node at layer 2 to the new node.
        Additionally, create two more arcs to connect the new node to the UDP/TCP nodes at layer 4.
        The parse graph will look like this:
        * MAC -> IPv4, IPv6, MY_IP
        * IPv4 -> TCP, UDP
        * IPv6 -> TCP, UDP
        * MY_IP -> TCP, UDP

        The MY_IP protocol header is structured as follows:
        * First 32 bits:
            payload_length (6 bits)
            reserved (10 bits)
            next_header (8 bits)
            magic (8 bits)
        * Next 32 bits:
            source_address (32 bits)
        * Next 32 bits:
            destination_address (32 bits)
        """
        # MY_IP Node
        payload_cfg = p.HwsParserHeaderFieldConfig(field_bit_offset=0, field_bit_length=6)
        next_hdr_cfg = p.HwsParserHeaderFieldConfig(field_bit_offset=16, field_bit_length=8)
        my_ip_hdr_len_cfg = p.HwsParserHeaderLenConfig(mode=e.MLX5DV_HWS_PARSER_HDR_LEN_MODE_FIELD,
                                                       field_config=payload_cfg,
                                                       multiplier=1,
                                                       addition=0)
        my_ip_cfg = p.HwsParserNodeConfig(my_ip_hdr_len_cfg, has_selection_field=True,
                                          selection_field=next_hdr_cfg)
        self.my_ip = p.HwsParserNode(self.graph, my_ip_cfg)
        # MAC -> MY_IP Arc
        src_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE,
                                      native_node=e.MLX5DV_HWS_PARSER_NATIVE_NODE_MAC)
        dst_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_FLEX,
                                      flex_node=self.my_ip)
        mac_arc_cfg = p.HwsParserArcConfig(src_node, dst_node,
                                           selection_field_data=u.PacketConsts.ETHER_TYPE_MY_IP)
        self.arc_mac_my_ip = p.HwsParserArc(self.graph, mac_arc_cfg)
        # MY_IP -> TCP Arc
        src_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_FLEX,
                                      flex_node=self.my_ip)
        dst_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE,
                                      native_node=e.MLX5DV_HWS_PARSER_NATIVE_NODE_TCP)
        tcp_arc_cfg = p.HwsParserArcConfig(src_node, dst_node,
                                           selection_field_data=socket.IPPROTO_TCP)
        self.arc_my_ip_tcp = p.HwsParserArc(self.graph, tcp_arc_cfg)
        # MY_IP -> UDP Arc
        src_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_FLEX,
                                      flex_node=self.my_ip)
        dst_node = p.HwsParserArcNode(e.MLX5DV_HWS_PARSER_NODE_TYPE_NATIVE,
                                      native_node=e.MLX5DV_HWS_PARSER_NATIVE_NODE_UDP)
        udp_arc_cfg = p.HwsParserArcConfig(src_node, dst_node,
                                           selection_field_data=socket.IPPROTO_UDP)
        self.arc_my_ip_udp = p.HwsParserArc(self.graph, udp_arc_cfg)

        return self.my_ip

    @requires_nvconfig('FLEX_PARSER_PROFILE_ENABLE', 4)
    @requires_nvconfig('PROG_PARSE_GRAPH', 1)
    def test_mlx5dv_hws_parser_my_ip(self):
        """
        Add another header (MY_IP) that will follow the MAC header with eth_type 0x799,
        and will proceed to UDP/TCP according the next header value.
        Create the parse graph and match on the source_address.

        Note - The sampler currently supports only 32-bit granularity.
        Hence, we will match 32 bits with an offset of 32 of MY_IP header.

        Create a new graph that will include to a new node labeled my_ip at layer 3.
        Create a sampler to sample the source_address field from my_ip node.
        Bind the graph.
        Query the sampler.
        Create all necessary steering resources, including TIR rule,
        Send traffic and match on flex parser sampled field (the source_address) and on
        the flex parser ok field,
        Modify the flex parser field, Send traffic and verify.
        Unbind the graph.
        """
        self.graph = p.HwsParserGraph(self.server.hws_ctx)
        self.my_ip = self.insert_my_ip_node_to_graph()
        sampler_field_cfg = \
            p.HwsParserHeaderFieldConfig(field_bit_offset=HwsFlexParserTest.SAMPLER_GRANULARITY,
                                         field_bit_length=HwsFlexParserTest.SAMPLER_GRANULARITY)
        sampler_cfg = p.HwsParserSamplerConfig(0, sampler_field_cfg)
        sampler = p.HwsParserSampler(sampler_cfg, self.my_ip)

        self.graph.bind()
        parser_fname, parser_ok_bit_fname, field_header, modify_header_field = sampler.query()

        ip = '1.2.3.4'
        bad_ip = '1.2.3.5'
        mdf_ip = '1.2.3.6'
        item = self.create_sampler_item(parser_fname, field_header, int(ipaddress.ip_address(ip)),
                                        0xffffffff)
        ok_item = self.create_sampler_item(parser_ok_bit_fname, field_header, 0b1, 0b1)

        packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER, src_ipv4=ip)
        unexp_packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,
                                    src_ipv4=bad_ip)
        ok_unexp_packet = u.gen_packet(self.server.msg_size, src_ipv4=ip)
        mdf_packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,src_ipv4=mdf_ip)

        self.match_with_traffic([item, ok_item], packet, [unexp_packet, ok_unexp_packet])
        self.modify_match_with_traffic(item, modify_header_field, int(ipaddress.ip_address(mdf_ip)),
                                       ModifyFieldsLen.IPV4, packet, mdf_packet)
        self.graph.unbind()

    @requires_nvconfig('FLEX_PARSER_PROFILE_ENABLE', 4)
    @requires_nvconfig('PROG_PARSE_GRAPH', 1)
    def test_mlx5dv_hws_parser_my_ip_multi_samplers(self):
        """
        Add another header (MY_IP) that will follow the MAC header with eth_type 0x799,
        and will proceed to UDP/TCP according the next header value.
        Create the parse graph and match on the magic number and source_address.

        Note - The sampler currently supports only 32-bit granularity.
        Hence, for source address we will match 32 bits with an offset of 32 of MY_IP header,
               and for the magic we will match using a partial mask on the 8 LSB starting from an
               offset of 0.

        Create a new graph that will include to a new node labeled my_ip at layer 3.
        Create two samplers to sample the magic field and source_address field from my_ip node.
        Bind the graph.
        Query the samplers.
        Create all necessary steering resources, including TIR rule,
        Send traffic and match on flex parser sampled fields and on the flex parser ok bits,
        Modify the flex parser fields, Send traffic and verify.
        Unbind the graph.
        """
        self.graph = p.HwsParserGraph(self.server.hws_ctx)
        self.my_ip = self.insert_my_ip_node_to_graph()

        sampler_field_cfg = \
            p.HwsParserHeaderFieldConfig(field_bit_offset=HwsFlexParserTest.SAMPLER_GRANULARITY,
                                         field_bit_length=HwsFlexParserTest.SAMPLER_GRANULARITY)
        src_addr_sampler_cfg = p.HwsParserSamplerConfig(0, sampler_field_cfg)
        src_addr_sampler = p.HwsParserSampler(src_addr_sampler_cfg, self.my_ip)

        sampler_field_cfg = \
            p.HwsParserHeaderFieldConfig(field_bit_offset=0,
                                         field_bit_length=HwsFlexParserTest.SAMPLER_GRANULARITY)
        magic_sampler_cfg = p.HwsParserSamplerConfig(0, sampler_field_cfg)
        magic_sampler = p.HwsParserSampler(magic_sampler_cfg, self.my_ip)

        self.graph.bind()
        src_parser_fname, src_parser_ok_bit_fname, src_field_header, src_modify_header_field = \
                                                                            src_addr_sampler.query()
        mgc_parser_fname, mgc_parser_ok_bit_fname, mgc_field_header, mgc_modify_header_field = \
                                                                            magic_sampler.query()

        IP = '1.2.3.4'
        BAD_IP = '1.2.3.5'
        src_addr_item = self.create_sampler_item(src_parser_fname, src_field_header,
                                                 int(ipaddress.ip_address(IP)), 0xffffffff)
        src_addr_ok_item = self.create_sampler_item(src_parser_ok_bit_fname, src_field_header, 0b1,
                                                    0b1)

        MAGIC = 0x7
        BAD_MAGIC = 0x8
        magic_item = self.create_sampler_item(mgc_parser_fname, mgc_field_header, MAGIC, 0xff)
        magic_ok_item = self.create_sampler_item(mgc_parser_ok_bit_fname, mgc_field_header, 0b1,
                                                 0b1)

        # Traffic
        packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER, src_ipv4=IP,
                              magic=MAGIC)
        unexp_packets = []
        unexp_packets.append(u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,
                                          src_ipv4=BAD_IP, magic=MAGIC))
        unexp_packets.append(u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,
                                          src_ipv4=IP, magic=BAD_MAGIC))
        unexp_packets.append(u.gen_packet(self.server.msg_size, src_ipv4=IP))
        self.match_with_traffic([src_addr_item, src_addr_ok_item, magic_item, magic_ok_item],
                                packet, unexp_packets)

        # Modify source ip and verify
        packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER, src_ipv4=IP)
        mdf_packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,
                                  src_ipv4=BAD_IP)
        self.modify_match_with_traffic(src_addr_item, src_modify_header_field,
                                       int(ipaddress.ip_address(BAD_IP)), ModifyFieldsLen.IPV4,
                                       packet, mdf_packet)
        # Modify magic and verify
        MAGIC_LEN = 8
        packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER, magic=MAGIC)
        mdf_packet = u.gen_packet(self.server.msg_size, l3=u.PacketConsts.MY_IP_VER,
                                  magic=BAD_MAGIC)
        self.modify_match_with_traffic(magic_item, mgc_modify_header_field, BAD_MAGIC, MAGIC_LEN,
                                       packet, mdf_packet)
        self.graph.unbind()

    @requires_nvconfig('FLEX_PARSER_PROFILE_ENABLE', [0, 8])
    @requires_geneve_tlv_sample
    @requires_geneve_tlv_option_offset
    def test_mlx5dv_hws_parser_geneve_options_sampler(self):
        """
        Create a sampler to sample the GENEVE TLV options.
        Query the sampler.
        Create all necessary steering resources, including TIR rule,
        Send traffic and match on flex parser sampled field (Option Data Len) and on
        the flex parser ok field.
        NOTE - option_len <= HCA_CAP.max_geneve_tlv_options
        NOTE - option_data_len <= HCA_CAP.max_geneve_tlv_option_data_len
        NOTE - cfg.offset_valid = 1, if FLEX_PARSE_PROFILE=8
                                  0, if FLEX_PARSE_PROFILE=0
        NOTE - geneve_tlv_sample supported from FW version 28.43.1014
        Geneve Header:
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |Ver|  Opt Len  |O|C|    Rsvd.  |          Protocol Type        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |        Virtual Network Identifier (VNI)       |    Reserved   |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |       Option Class       | Option Type |R|R|R| Option Data Len|
            ~                                                               ~Variable-Length Options
            |                      Option Data                              |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        """
        OPTION_DATA_LEN = 31
        OPTION_DATA_OFFSET = 0
        geneve_sampler_cfg = p.HwsParserGeneveOptionConfig(u.PacketConsts.GENEVE_OPTION_CLASS,
                                                           u.PacketConsts.GENEVE_OPTION_TYPE,
                                                           option_data_len=OPTION_DATA_LEN,
                                                           sample_offset=OPTION_DATA_OFFSET)
        geneve_sampler = p.HwsParserSampler(geneve_sampler_cfg, context=self.server.hws_ctx)
        parser_fname, parser_ok_bit_fname, field_header, modify_header_field = \
                                                                            geneve_sampler.query()
        # Match on Option Data Len
        OPTION_DATA = b'Hello'
        BAD_OPTION_DATA = b'Hi'
        GENEVE_VAL = len(OPTION_DATA)
        GENEVE_MASK = 0x1f

        item = self.create_sampler_item(parser_fname, field_header, GENEVE_VAL, GENEVE_MASK)
        ok_item = self.create_sampler_item(parser_ok_bit_fname, field_header, 0b1, 0b1)

        geneve_header = u.gen_geneve_header(opt_len=2, # There are 2 DWs in the options
                                            options=[(u.PacketConsts.GENEVE_OPTION_CLASS,
                                                      u.PacketConsts.GENEVE_OPTION_TYPE,
                                                      OPTION_DATA)])
        exp_packet = u.gen_packet(self.server.msg_size, tunnel=u.TunnelType.GENEVEoUDP,
                                  geneve_header=geneve_header)
        geneve_bad_header = \
            u.gen_geneve_header(opt_len=2, options=[(u.PacketConsts.GENEVE_OPTION_CLASS,
                                                     u.PacketConsts.GENEVE_OPTION_TYPE,
                                                     BAD_OPTION_DATA)])
        unexp_packet = u.gen_packet(self.server.msg_size, tunnel=u.TunnelType.GENEVEoUDP,
                                    geneve_header=geneve_bad_header)
        ok_unexp_packet = u.gen_packet(self.server.msg_size)

        self.match_with_traffic([item, ok_item], exp_packet, [unexp_packet, ok_unexp_packet])


class HwsMetricsTest(Mlx5RDMATestCase):
    def setUp(self):
        super().setUp()
        skip_if_hws_not_supported(Mlx5Context(Mlx5DVContextAttr(), self.dev_name))
        self.hws_res = Mlx5DvHwsResources(**self.dev_info, server=True)

    def tearDown(self):
        if self.hws_res:
            self.hws_res.hws_ctx.close()
            self.hws_res.ctx.close()

    @requires_adv_rdma_support
    def test_mlx5dv_hws_table_metric(self):
        """
        Creates tables of all supported types and verifies their metrics: FT type and level.
        Comparing values used in creation and vaues retrieved from metrics.
        """
        infos = []
        tables = []
        table_types = range(7)
        expected_table_types = [e.MLX5DV_HWS_TABLE_TYPE_NIC_RX,
                                e.MLX5DV_HWS_TABLE_TYPE_NIC_TX,
                                e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX,
                                e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX] + \
                               [e.MLX5DV_HWS_TABLE_TYPE_FDB] * 3 # FDB TX/RX are FDB type for FW
        table_levels = range(1, 8) # Root is not supported
        for t_type, t_level in zip(table_types, table_levels):
            if t_type in [e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_RX,
                          e.MLX5DV_HWS_TABLE_TYPE_RDMA_TRANSPORT_TX]:
                tables.append(self.hws_res.create_table(t_level, t_type, self.ib_port))
            else:
                tables.append(self.hws_res.create_table(t_level, t_type))
            infos.append(tables[-1].get_info())
        for i, t, l in zip(infos, expected_table_types, table_levels):
            self.assertEqual(i.ft_fw_type, t, 'Created table type is different from required.')
            self.assertEqual(i.level, l, 'Created table level is different from required.')

    def verify_metrics_resource_info(self, metrics, expected_values):
        error_messages = {
            'match_stes': 'Wrong number of match STEs from metric.',
            'action_stes': 'Wrong number of action STEs from metric.',
            'counters': 'Wrong number of counter actions from metric.',
            'aso': 'Wrong number of aso actions from metric.',
            'args': 'Wrong number of argument actions from metric.',
            'crypto': 'Wrong number of crypto actions from metric.'
        }
        for attr, expected_value in expected_values.items():
            self.assertEqual(getattr(metrics, attr), expected_value, error_messages[attr])

    def test_mlx5dv_hws_mt_metric(self):
        """
        Creates a matcher with no actions and complex AT (that includes tag,
        push VLAN, counter and modify actions) verifies MTs' metrics resource info.
        Compare costs between simple and complex ATs.
        Creates a matcher with complex MT (TCP source port range item) verifies
        MT's metrics info.
        Compare costs between simple AT (which also have simple MT
        (matching on sip)) and complex MT.
        """
        from tests.mlx5_prm_structs import SetActionIn
        simple_items = [create_sipv4_item()]
        simple_at = a.HwsActionTemplate([], 0 )
        tag_a = a.HwsActionTag(self.hws_res.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        push_a = a.HwsActionPushVlan(self.hws_res.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        counter_a, _ = self.hws_res.create_counter()
        resource_attr = HwsResourceAttr(e.MLX5DV_HWS_RESOURCE_TYPE_ARGUMENT_64B,
                                        bulk_log_size=0)
        argument = HwsResource(self.hws_res.hws_ctx, resource_attr)
        action1 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_47_16,
                              data=0x88888888, length=ModifyFieldsLen.MAC_47_16)
        action2 = SetActionIn(action_type=SET_ACTION, field=ModifyFields.OUT_SMAC_15_0,
                              data=0x8888, length=ModifyFieldsLen.MAC_15_0)
        mdf_actions = [action1, action2]
        mdf_a = a.HwsActionModify(self.hws_res.hws_ctx, mdf_actions,
                                  e.MLX5DV_HWS_ACTION_FLAG_NIC_RX, [argument], 1)
        actions = [tag_a, push_a, counter_a, mdf_a]
        complex_at = a.HwsActionTemplate(actions, len(actions))
        table = self.hws_res.create_table(level=3, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        mt = [HwsMatchTemplate(self.hws_res.hws_ctx, simple_items, len(simple_items))]
        matcher_attr = HwsMatcherAttr(priority=0)
        at = [simple_at, complex_at]
        matcher = HwsMatcher(table, mt, len(mt), at, len(at), matcher_attr)
        flags = e.MLX5DV_HWS_METRIC_FLAG_RESOURCE | e.MLX5DV_HWS_METRIC_FLAG_COST
        attr = HwsMetricMatcherTemplateAttr(flags, None, [], 0, 0)
        simple_at_metric = matcher.get_mt_metric(attr)
        # Verify simple AT
        expected_values = {
            'match_stes': 1,
            'action_stes': 0,
            'counters': 0,
            'aso': 0,
            'args': 0,
            'crypto': 0
        }
        self.verify_metrics_resource_info(simple_at_metric, expected_values)
        # Verify complex AT
        attr = HwsMetricMatcherTemplateAttr(flags, None, [], 0, 1)
        complex_at_metric = matcher.get_mt_metric(attr)
        expected_values = {
            'match_stes': 1,
            'action_stes': 1,
            'counters': 1,
            'aso': 0,
            'args': 1,
            'crypto': 0
        }
        self.verify_metrics_resource_info(complex_at_metric, expected_values)
        # Compare cost
        self.assertGreater(complex_at_metric.memory_cost, simple_at_metric.memory_cost,
                           'Memory cost metric mismatch.')
        self.assertGreater(complex_at_metric.pkt_processing_cost,
                           simple_at_metric.pkt_processing_cost,
                           'Packet processing cost metric mismatch.')
        self.assertGreater(complex_at_metric.insertion_cost, simple_at_metric.insertion_cost,
                           'Rule insertion cost metric mismatch.')
        complex_items = [create_outer_tcp_sport_range_item()]
        mt = [HwsMatchTemplate(self.hws_res.hws_ctx, complex_items, len(complex_items))]
        matcher_attr = HwsMatcherAttr(priority=0)
        at = [simple_at]
        matcher = HwsMatcher(table, mt, len(mt), at, len(at), matcher_attr)
        attr = HwsMetricMatcherTemplateAttr(flags, None, [], 0, 0)
        complex_mt_metric = matcher.get_mt_metric(attr)
        # Verify complex MT
        expected_values = {
            'match_stes': 2,
            'action_stes': 0,
            'counters': 0,
            'aso': 0,
            'args': 0,
            'crypto': 0
        }
        self.verify_metrics_resource_info(complex_mt_metric, expected_values)
        # Compare cost
        self.assertGreater(complex_mt_metric.memory_cost, simple_at_metric.memory_cost,
                           'Memory cost metric mismatch.')
        self.assertGreater(complex_mt_metric.pkt_processing_cost,
                           simple_at_metric.pkt_processing_cost,
                           'Packet processing cost metric mismatch.')
        self.assertGreater(complex_mt_metric.insertion_cost, simple_at_metric.insertion_cost,
                           'Rule insertion cost metric mismatch.')

    def test_mlx5dv_hws_matcher_metric(self):
        """
        Creates a matcher and verifies its metrics.
        """
        items = [create_sipv4_item()]
        push_a = a.HwsActionPushVlan(self.hws_res.hws_ctx, e.MLX5DV_HWS_ACTION_FLAG_NIC_RX)
        action_templates = [a.HwsActionTemplate([push_a, push_a], 2)]
        table = self.hws_res.create_table(level=3, table_type=e.MLX5DV_HWS_TABLE_TYPE_NIC_RX)
        matcher_templates = [HwsMatchTemplate(self.hws_res.hws_ctx, items, len(items))]
        prio = 3
        rule_log = 11 # to make sure we have collision matcher
        matcher_attr = HwsMatcherAttr(priority=prio,
                                      resource_mode=e.MLX5DV_HWS_MATCHER_RESOURCE_MODE_RULE,
                                      rule_log=rule_log)
        matcher = HwsMatcher(table, matcher_templates, len(matcher_templates), action_templates,
                             len(action_templates), matcher_attr)
        flags = e.MLX5DV_HWS_METRIC_FLAG_INFO | e.MLX5DV_HWS_METRIC_FLAG_COST
        m_metric = HwsMatcherMetric(matcher, flags)
        self.assertGreater(m_metric.matcher_tables_mem_sz, 0,
                           'Tables\' memory must be greater than 0.')
        self.assertGreater(m_metric.matcher_id, 0, 'Matcher\'s ID must be greater than 0.')
        self.assertEqual(m_metric.priority, prio, 'Matcher\'s metric priority doesn\'t match.')
        self.assertEqual(m_metric.log_num_of_rows, rule_log,
                         'Matcher\'s metric number of rows doesn\'t match.')
        self.assertGreater(m_metric.log_num_of_columns, 0,
                           'Matcher\'s metric number of cols must be greater than 0.')
        self.assertTrue(m_metric.col_matcher_exist,
                        'Matcher\'s metric collision matcher doesn\'t match.')
        self.assertGreater(m_metric.col_matcher_log_num_of_rows, 0,
                           'Matcher\'s metric collision matcher doesn\'t match.')
        self.assertGreater(m_metric.col_matcher_log_num_of_columns, 0,
                           'Matcher\'s metric collision matcher doesn\'t match.')
        self.assertEqual(m_metric.num_of_match_stes, 1,
                         'Matcher\'s metric number of match STEs doesn\'t match.')
        self.assertEqual(m_metric.num_of_action_stes, 1,
                         'Matcher\'s metric number of action STEs doesn\'t match.')
        self.assertFalse(m_metric.is_jumbo, 'Matcher\'s metric is jumbo doesn\'t match.')
        self.assertGreater(m_metric.memory_cost, 0,
                           'Matcher\'s metric memory cost must be greater than 0.')
        self.assertGreater(m_metric.miss_pkt_processing, 0,
                           'Matcher\'s metric miss packet processing cost must be greater than 0.')
